"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
import nord.pasion.progbtc.ecc as ecc
import nord.pasion.progbtc.helper as helper
import nord.pasion.progbtc.tx as tx
import nord.pasion.progbtc.script as script
from nord.wrigan.visibles.utils import show_error
from nord.sigurd.utils.log import note
import json
from io import BytesIO
import requests
import datetime


_user_pk = None
_user_address = None
_blockchain = None
_known_addresses = {}
_known_transactions = {}


def get_pk(filename='/tmp/nord.pkey', secret='This is the input secret key'):
    """
    If a wallet does not exist, create it, otherwise
    return the private key
    """
    try:
        with open(filename, 'r') as pkf:
            payload = pkf.read().strip().encode()
    except FileNotFoundError:
        payload = secret.encode()
        with open(filename, 'wb') as pkf:
            pkf.write(payload)

    secret = helper.little_endian_to_int(helper.hash256(payload))
    pk = ecc.PrivateKey(secret)
    return pk


@funlog
def create_address(*args, **kwargs):
    """
    Not implemented, the idea is to
    create an new address based upon
    an (as yet, unimplimented)
    Hierarchical Deterministic wallet.
    """
    pk = get_pk()
    address = pk.point.address(testnet=True)
    show_error(address)


@funlog
def establish_address(filename, secret):
    global _user_pk, _user_address
    _user_pk = get_pk(filename=filename, secret=secret)
    _user_address = _user_pk.point.address(testnet=True)


@funlog
def select_address(*args, **kwargs):
    global _user_pk, _user_address
    _user_pk = get_pk()
    _user_address = _user_pk.point.address(testnet=True)


@funlog
def set_pkfile(filename):
    global _user_pk, _user_address
    _user_pk = get_pk(filename)
    _user_address = _user_pk.point.address(testnet=True)


@funlog
def get_address():
    """
    Return the configured address to use for billing
    """
    global _user_address
    return _user_address


@funlog
def balance(*args, **kwargs):
    pk = get_pk()
    address = pk.point.address(testnet=True)
    resp = requests.get(f'https://api.blockcypher.com/v1/btc/test3/addrs/{address}/full?limit=50')
    if resp.status_code != 200:
        raise Exception(f"Bitcoin API failed: {resp.text}")
    vals = resp.json()
    show_error(f"Balance: {vals['balance']} satoshi")


@funlog
def pay_for_map(map_dict, payee, prev_tx, prev_idx):
    """
    Generate the signed transaction based upon the
    map's estimated run cost. This cost is
    estimated as the number of nodes in the
    map.
    """

    target_amount = len(map_dict["map"]["nodes"]) * 100
    return create_transaction(target_amount, payee, prev_tx, prev_idx)


def create_transaction(amount, payee, prev_tx=None, prev_idx=None):
    if prev_tx:
        if type(prev_tx) == str:
            prev_tx = bytes.fromhex(prev_tx)
        # the input transactions are
        # just the last transaction for this
        # wallet
        tx_ins = [tx.TxIn(prev_tx, prev_idx)]
    else:
        tx_ins = [tx.TxIn(helper.hash256(b'mw4FfpZygzuDQqi5sRyTYZuzXCc4WLnLq4'), 0)]

    # set up the pre-payment to the server
    tx_outs = []
    h160 = helper.decode_base58(payee)
    script_pubkey = script.p2pkh_script(h160)
    tx_outs.append(tx.TxOut(amount, script_pubkey))

    tx_obj = tx.Tx(1, tx_ins, tx_outs, 0, testnet=True)
    if prev_tx:
        tx_obj.sign_input(0, _user_pk)
    tx_obj.verify()
    return tx_obj.serialize().hex(), tx_obj.id()


@funlog
def establish_session_from_tx(tx_hex, payer, payee):
    """
    This is called from the Map on the runtime side
    When the map is loaded by wrigan. This extracts
    the wallet addresses, creates the economic session,
    and writes the prepayment (passed tx) to the chain
    """
    parsed = tx.Tx.parse(BytesIO(bytes.fromhex(tx_hex)))
    write_to_blockchain(tx_hex, payer, payee)
    import nord.pasion.managers.sessionmanager as sm
    from nord.pasion.providers import (hw, sw, stor, net)
    sm.create_session(payer, sw, hw, stor, net, parsed.tx_outs[0].amount)


@funlog
def settle_session(map_dict):
    import nord.pasion.managers.sessionmanager as sm
    esess = sm.get_current_session()
    spent_amt = esess.get_start_balance() - esess.get_balance()
    refund_amt = esess.get_balance()
    note(f"Completed graph cost {spent_amt} satoshi to run. Issuing refund of {refund_amt} satoshi")
    txid, idx = get_last_txid(get_address())
    trans, ntxid = create_transaction(int(refund_amt), sm.get_requestor(), txid, idx)
    write_to_blockchain(trans, get_address(), sm.get_requestor())


@funlog
def write_blockchain():
    global _blockchain
    with open('/tmp/wrigan.blockchain', 'w') as bck:
        json.dump(_blockchain, bck)


@funlog
def open_or_init_blockchain(filename='/tmp/wrigan.blockchain'):
    global _blockchain
    try:
        with open(filename, 'r') as bck:
            _blockchain = json.load(bck)
    except FileNotFoundError:
        _blockchain = []
        write_blockchain()


@funlog
def write_to_blockchain(transaction, payer, payee):
    if _blockchain is None:
        open_or_init_blockchain()

    parsed = tx.Tx.parse(BytesIO(bytes.fromhex(transaction)))

    block = {
        'txdata': transaction,
        'sender': payer,
        'recipient': payee,
        'timestamp': datetime.datetime.now().isoformat(),
        'hash': parsed.id(),
        'amount': parsed.tx_outs[0].amount
    }
    addr = helper.decode_base58(parsed.tx_outs[0].script_pubkey.address())
    _blockchain.append(block)
    write_blockchain()
    return parsed


# @funlog
# def grant_funds(payee, amount):
#     txblk, prev_id = create_transaction(amount, payee)
#     write_to_blockchain(txblk, "0000", payee)


@funlog
def load_blockchain(filename='/tmp/wrigan.blockchain'):
    open_or_init_blockchain()
    for block in _blockchain:
        parsed = tx.Tx.parse(BytesIO(bytes.fromhex(block["txdata"])))
        # store the latest values for each address...
        _known_addresses[block["recipient"]] = block
        _known_addresses[block["sender"]] = block
        _known_transactions[bytes.fromhex(parsed.id())] = block
        _known_transactions[parsed.id()] = block



@funlog
def get_last_txid(address):
    """
    Given an address, find the last transaction ID
    and index within the transaction for this address
    """
    load_blockchain()
    addr_b = helper.decode_base58(address)
    if address in _known_addresses:
        block = _known_addresses[address]
        parsed = tx.Tx.parse(BytesIO(bytes.fromhex(block["txdata"])))
        txid = parsed.id()
        cnt = 0
        for t in parsed.tx_outs:
            if helper.decode_base58(t.script_pubkey.address()) == addr_b:
                # The address in the output matches the address
                # of the requested wallet
                index = cnt
                break
            cnt += 1

        return txid, cnt
    else:
        return None, None


@funlog
def get_tx(txid):
    if txid in _known_transactions:
        return _known_transactions[txid]["txdata"]
    else:
        return ""
