"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The session manager provides the runtime with the ability to
determine if it is funded to continue.

Sessions have a balance, a set of resource providers and a resource requestor.
"""

import nord.pasion.providers.hw as _hw
import nord.pasion.providers.sw as _sw
import nord.pasion.providers.net as _net
import nord.pasion.providers.stor as _stor
import math


_current_session = None


class EconomicSession(object):
    def __init__(self,
                 requestor=None,
                 swprov=None,
                 hwprov=None,
                 storprov=None,
                 netprov=None,
                 balance=float('NaN')):
        self.requestor = requestor
        self.sw = swprov
        self.hw = hwprov
        self.stor = storprov
        self.net = netprov
        self.balance = balance
        self.start_balance = balance

    def get_requestor(self):
        return self.requestor

    def consumesw(self, units):
        if self.sw is None:
            return True
        amount = self.sw.amount(units)
        return self.spend_amount(amount)

    def consumehw(self, units):
        if self.hw is None:
            return True
        amount = self.hw.amount(units)
        return self.spend_amount(amount)

    def consumestor(self, units):
        if self.stor is None:
            return True
        amount = self.stor.amount(units)
        return self.spend_amount(amount)

    def consumenet(self, units):
        if self.net is None:
            return True
        amount = self.net.amount(units)
        return self.spend_amount(amount)

    def spend_amount(self, amount):
        # if an address is set for the requestor
        # but no balance is set, this is an error
        # so, indicate spend not possible
        if self.requestor and math.isnan(self.balance):
            return False
        if math.isnan(amount):
            return False
        if amount > self.balance and not math.isnan(self.balance):
            return False
        elif math.isnan(self.balance):
            return True
        else:
            self.balance -= amount
            return True

    def add_credit(self, amount):
        if not math.isnan(amount):
            if math.isnan(self.balance):
                self.balance = 0.0
            self.balance += amount

    def get_balance(self):
        return self.balance

    def get_start_balance(self):
        return self.start_balance


def get_current_session():
    """
    Returns the current session to
    the requestor. Uninitialized sessions
    have infinite balances (thus permitting
    free use of one's own resources, and, execution
    of the runtime in tests)
    """
    global _current_session
    return _current_session


def clear_current_session():
    """
    None'ifies _current_session, primarily for
    testing purposes.
    """
    global _current_session
    del _current_session
    _current_session = None  # noqa: F841


def create_session(requestor, sw, hw, net, stor, balance=float('NaN')):
    """
    The inputs are the addresses of the recipient wallets
    and the requestor object (with the path to their key within)
    """
    global _current_session
    _sw.set_rx_address(sw)
    _hw.set_rx_address(hw)
    _stor.set_rx_address(stor)
    _net.set_rx_address(net)
    _current_session = EconomicSession(requestor, _sw, _hw, _stor, _net, balance)


def consumehw(units):
    global _current_session
    if _current_session:
        return _current_session.consumehw(units)
    return True


def consumesw(units):
    global _current_session
    if _current_session:
        return _current_session.consumesw(units)
    return True


def consumenet(units):
    global _current_session
    if _current_session:
        return _current_session.consumenet(units)
    return True


def consumestor(units):
    global _current_session
    if _current_session:
        return _current_session.consumestor(units)
    return True


def get_balance():
    global _current_session
    if _current_session:
        return _current_session.get_balance()
    else:
        return float('NaN')


def get_start_balance():
    global _current_session
    if _current_session:
        return _current_session.get_start_balance()
    else:
        return float('NaN')


def get_requestor():
    global _current_session
    if _current_session:
        return _current_session.get_requestor()
    else:
        return None
