"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Set up which keyboard events Panda3d accepts.
"""

from nord.sigurd.utils.log import notice
from nord.sigurd.utils.verboser import funlog  # noqa: F401
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.event as event
from panda3d.core import KeyboardButton


_target = None
_ctrl_pressed = False
_shift_pressed = False
_up_pressed = False
_down_pressed = False
_left_pressed = False
_right_pressed = False
_alt_pressed = False


@funlog
def set_target(target):
    """Set the text input target."""
    global _target
    _target = target


@funlog
def get_target():
    """Return the current text input target"""
    global _target
    return _target


@funlog
def send_char_to_target(char):
    """Send the input key events to the target."""
    global _target

    @event.callback
    @funlog
    def send_char_to_target(*args, **kwargs):
        """The actual key event router to the target."""
        notice("Trying to send '{}' to text input".format(char))
        if _target is not None:
            notice("Sending '{}' to text input: {}".format(char, _target.uid))
            _target.handle(char)
    return send_char_to_target


@funlog
def toggle_target_edit():
    """Enable or disable the target's edit mode."""
    global _target

    @event.callback
    @funlog
    def toggle_target_edit(*args, **kwargs):
        if _target is not None:
            _target.toggleEdit()
    return toggle_target_edit


@funlog
def handler(name):
    """Return a function which uses the statemanager."""
    def rv(*args, **kwargs):
        """Execute the statemanager's handler for the passed event."""
        f = getattr(em, name)
        return f(*args, **kwargs)
    return rv


@funlog
def load(app):
    """Setup and load all keybord event names."""
    k = dict()
    lk = dict()
    k['arrow_up'] = 'key-arrow_up'
    k['arrow_down'] = 'key-arrow_down'
    k['arrow_left'] = 'key-arrow_left'
    k['arrow_right'] = 'key-arrow_right'
    k['arrow_up-up'] = 'key-arrow_up-release'
    k['arrow_down-up'] = 'key-arrow_down-release'
    k['arrow_left-up'] = 'key-arrow_left-release'
    k['arrow_right-up'] = 'key-arrow_right-release'
    k['shift-arrow_up'] = 'key-shift-arrow_up'
    k['shift-arrow_down'] = 'key-shift-arrow_down'
    k['shift-arrow_up-up'] = 'key-shift-arrow_up-release'
    k['shift-arrow_down-up'] = 'key-shift-arrow_down-release'
    k['control-arrow_down'] = 'key-control-arrow_down'
    k['control-arrow_up'] = 'key-control-arrow_up'
    k['control-arrow_left'] = 'key-control-arrow_left'
    k['control-arrow_right'] = 'key-control-arrow_right'
    k['control-arrow_down-up'] = 'key-control-arrow_down-release'
    k['control-arrow_up-up'] = 'key-control-arrow_up-release'
    k['control-arrow_left-up'] = 'key-control-arrow_left-release'
    k['control-arrow_right-up'] = 'key-control-arrow_right-release'
    k['shift-control-arrow_right'] = 'key-shift-control-arrow_right'
    k['shift-control-arrow_left'] = 'key-shift-control-arrow_left'
    k['shift-control-delete'] = 'key-shift-control-delete'
    k['alt-arrow_down'] = 'key-alt-arrow_down'
    k['alt-arrow_up'] = 'key-alt-arrow_up'
    k['alt-arrow_left'] = 'key-alt-arrow_left'
    k['alt-arrow_right'] = 'key-alt-arrow_right'
    k['alt-arrow_down-up'] = 'key-alt-arrow_down-release'
    k['alt-arrow_up-up'] = 'key-alt-arrow_up-release'   
    k['alt-arrow_left-up'] = 'key-alt-arrow_left-release'
    k['alt-arrow_right-up'] = 'key-alt-arrow_right-release'

    k['a'] = 'key-a'
    k['b'] = 'key-b'
    k['c'] = 'key-c'
    k['d'] = 'key-d'
    k['e'] = 'key-e'
    k['f'] = 'key-f'
    k['g'] = 'key-g'
    k['h'] = 'key-h'
    k['i'] = 'key-i'
    k['j'] = 'key-j'
    k['k'] = 'key-k'
    k['l'] = 'key-l'
    k['m'] = 'key-m'
    k['n'] = 'key-n'
    k['o'] = 'key-o'
    k['p'] = 'key-p'
    k['q'] = 'key-q'
    k['r'] = 'key-r'
    k['s'] = 'key-s'
    k['t'] = 'key-t'
    k['u'] = 'key-u'
    k['v'] = 'key-v'
    k['w'] = 'key-w'
    k['x'] = 'key-x'
    k['y'] = 'key-y'
    k['z'] = 'key-z'

    lk['~'] = 'key-tilde'
    k['`'] = 'key-backtick'
    lk['!'] = 'key-bang'
    lk['@'] = 'key-at'
    lk['#'] = 'key-octothorpe'
    lk['$'] = 'key-dollar'
    lk['%'] = 'key-percent'
    lk['^'] = 'key-caret'
    lk['&'] = 'key-ampersand'
    lk['*'] = 'key-asterisk'
    lk['('] = 'key-leftparen'
    lk[')'] = 'key-rightparen'
    lk['{'] = 'key-leftcurly'
    lk['}'] = 'key-rightcurly'
    k['1'] = 'key-1'
    k['2'] = 'key-2'
    k['3'] = 'key-3'
    k['4'] = 'key-4'
    k['5'] = 'key-5'
    k['6'] = 'key-6'
    k['7'] = 'key-7'
    k['8'] = 'key-8'
    k['9'] = 'key-9'
    k['0'] = 'key-0'
    k['['] = 'key-leftsquare'
    k[']'] = 'key-rightsquare'
    lk['"'] = 'key-doublequote'
    lk['<'] = 'key-leftangle'
    lk['>'] = 'key-rightangle'
    k["'"] = 'key-singlequote'
    k[','] = 'key-comma'
    k['.'] = 'key-period'
    k['/'] = 'key-forwardslash'
    k['='] = 'key-equals'
    k['\\'] = 'key-backslash'
    lk['?'] = 'key-question'
    lk['+'] = 'key-plus'
    lk['|'] = 'key-pipe'
    k['-'] = 'key-dash'
    lk['_'] = 'key-underscore'
    k[';'] = 'key-semicolon'
    lk[':'] = 'key-colon'

    k['shift-a'] = 'key-shift-a'
    k['shift-b'] = 'key-shift-b'
    k['shift-c'] = 'key-shift-c'
    k['shift-d'] = 'key-shift-d'
    k['shift-e'] = 'key-shift-e'
    k['shift-f'] = 'key-shift-f'
    k['shift-g'] = 'key-shift-g'
    k['shift-h'] = 'key-shift-h'
    k['shift-i'] = 'key-shift-i'
    k['shift-j'] = 'key-shift-j'
    k['shift-k'] = 'key-shift-k'
    k['shift-l'] = 'key-shift-l'
    k['shift-m'] = 'key-shift-m'
    k['shift-n'] = 'key-shift-n'
    k['shift-o'] = 'key-shift-o'
    k['shift-p'] = 'key-shift-p'
    k['shift-q'] = 'key-shift-q'
    k['shift-r'] = 'key-shift-r'
    k['shift-s'] = 'key-shift-s'
    k['shift-t'] = 'key-shift-t'
    k['shift-u'] = 'key-shift-u'
    k['shift-v'] = 'key-shift-v'
    k['shift-w'] = 'key-shift-w'
    k['shift-x'] = 'key-shift-x'
    k['shift-y'] = 'key-shift-y'
    k['shift-z'] = 'key-shift-z'
    k['alt-u'] = 'key-alt-u'
    k['alt-d'] = 'key-alt-d'
    k['alt-l'] = 'key-alt-l'
    k['alt-r'] = 'key-alt-r'
    k['control-alt-r'] = 'key-control-alt-r'
    k['control-alt-l'] = 'key-control-alt-l'
    k['shift-alt-arrow_up'] = 'key-shift-alt-arrow_up'
    k['shift-alt-arrow_down'] = 'key-shift-alt-arrow_down'
    k['shift-alt-arrow_left'] = 'key-shift-alt-arrow_left'
    k['shift-alt-arrow_right'] = 'key-shift-alt-arrow_right'
    k['arrow_up-repeat'] = 'key-arrow_up-repeat'
    k['arrow_down-repeat'] = 'key-arrow_down-repeat'
    k['arrow_left-repeat'] = 'key-arrow_left-repeat'
    k['arrow_right-repeat'] = 'key-arrow_right-repeat'
    k['shift-arrow_up-repeat'] = 'key-shift-arrow_up-repeat'
    k['shift-arrow_down-repeat'] = 'key-shift-arrow_down-repeat'
    k['control-arrow_down-repeat'] = 'key-control-arrow_down-repeat'
    k['control-arrow_up-repeat'] = 'key-control-arrow_up-repeat'
    k['control-arrow_left-repeat'] = 'key-control-arrow_left-repeat'
    k['control-arrow_right-repeat'] = 'key-control-arrow_right-repeat'
    k['shift-control-arrow_right-repeat'] = 'key-shift-control-arrow_right-repeat'
    k['shift-control-arrow_left-repeat'] = 'key-shift-control-arrow_left-repeat'
    k['f-repeat'] = 'key-f-repeat'
    k['b-repeat'] = 'key-b-repeat'
    k['l-repeat'] = 'key-l-repeat'
    k['r-repeat'] = 'key-r-repeat'
    k['u-repeat'] = 'key-u-repeat'
    k['d-repeat'] = 'key-d-repeat'
    k['alt-u-repeat'] = 'key-alt-u-repeat'
    k['alt-d-repeat'] = 'key-alt-d-repeat'
    k['alt-l-repeat'] = 'key-alt-l-repeat'
    k['alt-r-repeat'] = 'key-alt-r-repeat'
    k['control-alt-r-repeat'] = 'key-control-alt-r-repeat'
    k['control-alt-l-repeat'] = 'key-control-alt-l-repeat'
    k['shift-alt-arrow_up-repeat'] = 'key-shift-alt-arrow_up-repeat'
    k['shift-alt-arrow_down-repeat'] = 'key-shift-alt-arrow_down-repeat'
    k['shift-alt-arrow_left-repeat'] = 'key-shift-alt-arrow_left-repeat'
    k['shift-alt-arrow_right-repeat'] = 'key-shift-alt-arrow_right-repeat'
    k['backspace'] = 'key-backspace'
    k['shift-control-c'] = 'key-shift-control-c'
    k['space'] = 'key-space'
    k['return'] = 'key-return'
    k['enter'] = 'key-return'
    k['escape'] = 'key-escape'
    k['control-d'] = 'key-control-d'
    k['delete'] = 'key-delete'
    k['tab'] = 'key-tab'
    k['control-tab'] = 'key-control-tab'
    k['shift-control-r'] = 'key-control-shift-r'

    for n in k.keys():
        evname = k[n]
        app.accept(n, handler(evname))
        # since buttonThrowers is None when we are testing headless
        # this condition allows the tests to proceed (without some shift required characters like :)
        if base.buttonThrowers:  # noqa: F821
            base.buttonThrowers[0].node().setKeystrokeEvent('keystroke')  # noqa: F821

        def gen_keyinhandler(mapping):
            def subhandler(keyname):
                evname = mapping.get(keyname, None)
                if evname:
                    handler(evname)()
            return subhandler

        app.accept('keystroke', gen_keyinhandler(lk))

# k['mouse1-up'] = _mouseHandler.screen_touch
# k['mouse1'] = _mouseHandler.start_drag
# k['escape'] = sys.exit
# k['q'] = sys.exit
# k['control-d'] = sys.exit


def detect_down_dir_keys():
    """Detect if directional keys are up or down."""
    # to facilitate automated testing.... when there is no display
    # the mouseWatcherNode will be None...
    if base.mouseWatcherNode is None:  # noqa: F821
        return
    is_pressed = base.mouseWatcherNode.is_button_down  # noqa: F821

    ctrl = KeyboardButton.control()
    alt = KeyboardButton.alt()
    shift = KeyboardButton.shift()
    up = KeyboardButton.up()
    down = KeyboardButton.down()
    left = KeyboardButton.left()
    right = KeyboardButton.right()

    # _ctrl_pressed = False
    # _shift_pressed = False
    # _up_pressed = False
    # _down_pressed = False
    # _left_pressed = False
    # _right_pressed = False

    if is_pressed(ctrl) and not is_pressed(shift):
        if is_pressed(up):
            em.fbyf_ctrl_up_pressed()
        if is_pressed(down):
            em.fbyf_ctrl_down_pressed()
        if is_pressed(left):
            em.fbyf_ctrl_left_pressed()
        if is_pressed(right):
            em.fbyf_ctrl_right_pressed()

    elif is_pressed(shift) and not is_pressed(ctrl):
        if is_pressed(up):
            em.fbyf_shift_up_pressed()
        if is_pressed(down):
            em.fbyf_shift_down_pressed()
        if is_pressed(left):
            em.fbyf_shift_left_pressed()
        if is_pressed(right):
            em.fbyf_shift_right_pressed()

    elif is_pressed(alt) and not is_pressed(ctrl):
        if is_pressed(up):
            em.fbyf_alt_up_pressed()
        if is_pressed(down):
            em.fbyf_alt_down_pressed()
        if is_pressed(left):
            em.fbyf_alt_left_pressed()
        if is_pressed(right):
            em.fbyf_alt_right_pressed()

    elif is_pressed(shift) and is_pressed(ctrl):
        if is_pressed(up):
            em.fbyf_ctrl_shift_up_pressed()
        if is_pressed(down):
            em.fbyf_ctrl_shift_down_pressed()
        if is_pressed(left):
            em.fbyf_ctrl_shift_left_pressed()
        if is_pressed(right):
            em.fbyf_ctrl_shift_right_pressed()

    else:
        if is_pressed(up):
            em.fbyf_up_pressed()
        if is_pressed(down):
            em.fbyf_down_pressed()
        if is_pressed(left):
            em.fbyf_left_pressed()
        if is_pressed(right):
            em.fbyf_right_pressed()
