"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Process raw mouse/pointer events and synthesize higher level events.
"""

from direct.task import Task
from panda3d.core import CollisionTraverser,\
    CollisionHandlerQueue,\
    CollisionNode,\
    CollisionRay,\
    GeomNode, VBase3
from nord.sigurd.config.globalVars import DRAGTHRESHOLD
from nord.sigurd.utils.log import info, notice, verb
from nord.sigurd.utils.verboser import funlog
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.objectmanager as objects
import nord.sigurd.managers.containermanager as cm


class MouseHandler(object):
    """Class which interprets pointer events."""

    @funlog
    def __init__(self, base, app):
        """Prepare the collision detection objects."""
        base.disableMouse()
        self.app = app

        self.dragging = False
        self.mouseDown = False
        self.last_mouse_x = None
        self.last_mouse_y = None
        self.over_node = None

        # set up the pointer to object "collision" detector
        self.pointerTraverser = CollisionTraverser("pointer traverser")
        self.pickerNode = CollisionNode('mouseRay')
        self.pointerCollisionHandler = CollisionHandlerQueue()
        picker_np = camera.attachNewNode(self.pickerNode)  # noqa: F821
        self.pickerNode.setFromCollideMask(GeomNode.getDefaultCollideMask())
        self.pickerRay = CollisionRay()
        self.pickerNode.addSolid(self.pickerRay)
        self.pointerTraverser.addCollider(picker_np,
                                          self.pointerCollisionHandler)

        self.lastOrigin = None

    @funlog
    def load(self):
        """Call the app dependent functions after app is set up."""
        self.app.accept('mouse1-up', self.screen_touch)
        self.app.accept('mouse1', self.start_drag)

    #  @  funlog # commented out due to crazy verbosity
    def mouse_task(self, task):
        """Deal with changes in mouse position."""
        # When testing (with offscreen window) there will be no
        # base.mouseWatcherNode
        if base.mouseWatcherNode is None:
            return Task.cont
        if not base.mouseWatcherNode.has_mouse() and self.over_node is None:  # noqa: F821
            return Task.cont
        elif self.over_node is not None and not base.mouseWatcherNode.has_mouse():
            mpos = self.last_pos
        else:
            mpos = base.mouseWatcherNode.getMouse()  # noqa: F821
            self.last_pos = mpos


        self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())  # noqa: F821
        self.pointerTraverser.traverse(cm.cwc())

        if self.pointerCollisionHandler.getNumEntries() > 0:
            self.pointerCollisionHandler.sortEntries()
            picked_obj = self.pointerCollisionHandler.getEntry(0)\
                .getIntoNodePath()
            picked_obj = picked_obj.findNetTag('hoverable')

            if not picked_obj.isEmpty():
                if not picked_obj.is_hidden():
                    if self.over_node is None:
                        self.over_node = objects.fetch(picked_obj.getTag('hoverable'))
                        verb("Got a mouseover: {}".format(self.over_node))
                        em.mouse_over(self.over_node, self.pickerRay.getOrigin())
                    elif self.over_node != objects.fetch(picked_obj.getTag('hoverable')):
                        em.mouse_was_over(self.over_node, self.over_node.getPos())
                        self.over_node = objects.fetch(picked_obj.getTag('hoverable'))
                        verb("Got a mouseover: {}".format(self.over_node))
                        em.mouse_over(self.over_node, self.pickerRay.getOrigin())
            else:
                if self.over_node is not None:
                    em.mouse_was_over(self.over_node, self.over_node.getPos())
                    self.over_node = None

        if not self.mouseDown:
            return Task.cont

        if self.last_mouse_x is None:
            self.last_mouse_x = mpos.getX()
            self.last_mouse_y = mpos.getY()

        dx = self.start_mouse_x - mpos.getX()
        dy = self.start_mouse_y - mpos.getY()
        rdx = self.last_mouse_x - mpos.getX()
        rdy = self.last_mouse_y - mpos.getY()

        self.last_mouse_x = mpos.getX()
        self.last_mouse_y = mpos.getY()

        if self.dragging:
            # Move the selected objects
            if abs(dx) > DRAGTHRESHOLD or abs(dy) > DRAGTHRESHOLD:
                em.mouse_selected_dragging(mpos)

        else:
            verb("Mouse delta since last: ({}, {})".format(rdx, rdy))
            # rotate the camera
            if rdx > DRAGTHRESHOLD:  # new position is left of old
                em.mouse_drag_left(dx)
            elif rdx < -1.0 * DRAGTHRESHOLD:
                em.mouse_drag_right(-1.0 * dx)

            if rdy > DRAGTHRESHOLD:  # the new position is below the old
                em.mouse_drag_down(dy)
            elif rdy < -1.0 * DRAGTHRESHOLD:
                em.mouse_drag_up(-1.0 * dy)

        return Task.cont

    @funlog
    def start_drag(self, tag='clickable'):
        """Initiate mouse drag processing."""
        verb("Pressed the mouse button")
        self.mouseDown = True
        mpos = base.mouseWatcherNode.getMouse()  # noqa: F821
        self.start_mouse_x = mpos.getX()
        self.start_mouse_y = mpos.getY()

        self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())  # noqa: F821
        self.pointerTraverser.traverse(cm.cwc())

        self.dragging = False

        if self.pointerCollisionHandler.getNumEntries() > 0:
            self.pointerCollisionHandler.sortEntries()
            picked_obj = self.pointerCollisionHandler.getEntry(0)\
                .getIntoNodePath()
            picked_obj = picked_obj.findNetTag(tag)
            if not picked_obj.isEmpty() and not picked_obj.is_hidden():
                print("Pressed the mouse button on an object")
                self.dragging = True

        return Task.cont

    @funlog
    def screen_touch(self, tag='clickable'):
        """Handle a screen touch or mouse up."""
        self.mouseDown = False

        mpos = base.mouseWatcherNode.getMouse()  # noqa: F821
        self.pickerRay.setFromLens(base.camNode, mpos.getX(), mpos.getY())  # noqa: F821
        self.pointerTraverser.traverse(cm.cwc())

        self.lastOrigin = self.pickerRay.getOrigin()

        if self.pointerCollisionHandler.getNumEntries() > 0:
            self.pointerCollisionHandler.sortEntries()
            picked_obj = self.pointerCollisionHandler.getEntry(0)\
                .getIntoNodePath()
            picked_obj = picked_obj.findNetTag(tag)
            if not picked_obj.is_hidden():
                if not picked_obj.isEmpty():
                    info("Selected Object: {}".format(picked_obj.getTag(tag)))
                    obj = objects.fetch(picked_obj.getTag(tag))
                    em.clickable_object_clicked(obj, self.pickerRay.getOrigin())

                else:
                    notice("Un-IDed object: {}"
                        .format(self.pointerCollisionHandler
                                .getEntry(0).getIntoNodePath()))
                    info("Selected a non-clickable object")
                    em.mouse_unidentified_object(self.pickerRay.getOrigin())
        elif self.last_mouse_x == mpos.getX()\
                and self.last_mouse_y == mpos.getY()\
                and not self.dragging:

            info("Touch did not collide with any objects")
            em.mouse_freespace_click(self.pickerRay.getOrigin(), self.app)

        if self.dragging:
            self.dragging = False

    @funlog
    def get_last_origin(self):
        """Return the last location of the mouse in 3space or the center."""
        if self.lastOrigin:
            return self.lastOrigin
        else:
            return VBase3(0.0, 0.0, 0.0)
