"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.utils.log import err, verb, warn
import nord.sigurd.managers.event as event
import nord.sigurd.managers.containermanager as cm
from nord.sigurd.managers.ui.decorationmanager import DecorationManager as dm
from nord.sigurd.shell.menuItem import MenuItem
import nord.nord.utils.SpaceManager as use
from nord.nord.exceptions import UseException as UseException
import nord.sigurd.utils.namegen as ng


"""
    The callback manager class is used to enable child classes
    to have a standardized way to allow consumers to register
    functions to be called when the child class triggers a given
    event

"""


def set_find_mapentry(vis):
    module = use.module(f'nord.{vis.space}.nodes.{vis.__class__.__name__}')
    if '__init__' in dir(module):
        modconstructor = module
    else:
        modconstructor = getattr(module, vis.__class__.__name__)

    import nord.sigurd.shell.vismap as vm
    vis.mapentry = modconstructor(vm.get_working_map())
    vis.mapentry.set_untracked_attr('visualization', vis)


class CallbackManager(object):
    @funlog
    def __init__(self, space=None, app=None):
        self.eventSubscribers = {}
        self.events = []
        self.links = {}
        self.delegates = {}
        self.decor_to_del_params = {}
        if app is not None:
            self.decorationmanager = dm(self, app, space=space)
        else:
            self.decorationmanager = None

        # if this cbm is used as a decoration for another object,
        # when it is removed, the decoration also needs to be
        # this relationship is tracked with this list
        self.representing_decorations = []

        self.onMappedFuncs = []
        self.onAllMappedFuncs = []
        self.onUnmappedFuncs = [self.cleanup_repr_decors]
        self.container = cm.get_current_name()

        self.filter_categories = [self.__class__.__name__]
        if space:
            self.filter_categories.append(space)
        # import nord.sigurd.managers.filtermanager as fm
        # fm.add_node(self)

    @funlog
    def cleanup_repr_decors(self):
        for decm, dec, rule, direction in self.representing_decorations:
            decm.rem_decoration(dec.split(':')[0], rule=rule, direction=direction)
        del self.representing_decorations[:]

    @funlog
    def track_decoratee(self, decm, dec, rule, direction):
        if (decm, dec) not in self.representing_decorations:
            self.representing_decorations.append((decm, dec, rule, direction))

    @funlog
    def get_filter_categories(self):
        return self.filter_categories

    @funlog
    def add_filter_category(self, category):
        if category not in self.filter_categories:
            self.filter_categories.append(category)
            import nord.sigurd.managers.filtermanager as fm
            fm.add_node(self)

    @funlog
    def addEvent(s, name):
        if name not in s.events:
            s.events.append(name)
            s.eventSubscribers[name] = []

    @funlog
    def addCallback(s, name, cbk):
        if name in s.events:
            if cbk not in s.eventSubscribers[name]:
                s.eventSubscribers[name].append(cbk)
        else:
            err("Class: {} does not trigger events named: {}".format(s.__class__.__name__, name))

    @funlog
    def triggerEvent(s, name, *args, **kwargs):
        if name in s.events:
            for cbk in s.eventSubscribers[name]:
                cbk(event.EventInstance(s.__class__.__name__, *args, **kwargs))
        else:
            err("Class: {} unable to trigger events named: {}".format(s.__class__.__name__, name))

    @funlog
    def link(self, obj):
        """Add retationship detween this object and another."""
        if hasattr(obj, 'uid'):
            verb('Linking {} to {}'.format(obj.uid, self.uid))
            self.links[obj.uid] = obj
        else:
            err("Cannot link {} to {}".format(str(obj), self.uid))

    @funlog
    def unlink(self, obj):
        """Remove any existing links between this and another object."""
        if hasattr(obj, 'uid'):
            if obj.uid in self.links:
                verb('Unlinking {} from {}'.format(obj.uid, self.uid))
                del self.links[obj.uid]
            else:
                err("{} not linked to to {}, cannot unlink"
                    .format(self.uid, obj.uid))
        else:
            err("Cannot unlink {} to {}".format(str(obj), self.uid))

    @funlog
    def is_delegate(self, ctype, direction):
        return False
        if direction in self.delegates and\
                ctype in self.delegates[direction]:
            return True
        else:
            return False

    @funlog
    def also_delegate(self, dec, rule_type, direction, extra=None):
        """Given a decoration, establish delegation lookups for provided connection type."""
        if direction not in self.delegates:
            self.delegates[direction] = {}
        if rule_type not in self.delegates[direction]:
            self.delegates[direction][rule_type] = []
        if dec not in self.delegates[direction][rule_type]:
            self.delegates[direction][rule_type].append(dec)
        if dec.uid not in self.decor_to_del_params:
            self.decor_to_del_params[dec.uid] = {"direction": direction,
                                                 "rule": rule_type,
                                                 "decoration": dec,
                                                 "extra": extra}

    @funlog
    def get_delegate(self, uid, direction, rule_type):
        if direction in self.delegates and\
                rule_type in self.delegates[direction]:
            rv = None
            for delegate in self.delegates[direction][rule_type]:
                if delegate.uid.split(':')[0] == uid:
                    rv = delegate
                    break
            return rv

    @funlog
    def process_delegated_click(self, mapentry, payload):
        """
        Placeholder for descendents to override.
        """
        pass

    @funlog
    def handle_delegate_click(self, uid):
        if uid not in self.decor_to_del_params:
            warn("Implementation ERROR: Misdirected decoration click")
        else:
            entry = self.decor_to_del_params[uid]
            mapentry = entry["decoration"]
            self.process_delegated_click(mapentry, entry)

    @funlog
    def repoint_to_delegates(self):
        for direction in self.delegates:
            for rule_type in self.delegates[direction]:
                for dec in self.delegates[direction][rule_type]:
                    if hasattr(dec, '_edges'):
                        for e in dec._edges:
                            if direction == 'as_src':
                                e.move_source_to(dec)
                            if direction == 'as_tgt':
                                e.move_target_to(dec)

    @funlog
    def repoint_from_delegates(self):
        for direction in self.delegates:
            for rule_type in self.delegates[direction]:
                for dec in self.delegates[direction][rule_type]:
                    if hasattr(dec, '_edges'):
                        for e in dec._edges:
                            if direction == 'as_src':
                                e.move_source_to(self)
                            if direction == 'as_tgt':
                                e.move_target_to(self)

    @funlog
    def connect_allowed(self, other, ctype, direction, delegate=None):
        """Placeholder to allow child instances to prevent connections."""
        return True

    @funlog
    def getParent(self):  # noqa: N802
        """Return the parent node of the Panda3d NodePath."""
        return self.data.getParent()

    @funlog
    def reparentTo(self, node):  # noqa: N802
        """Set the Panda3d NodePath's parent."""
        if not self.data.is_empty():
            self.data.reparentTo(node)
        else:
            warn(f"{self} panda3d data is empty")
        self.container = cm.get_current_name()
        if self.menu is not None:
            self.menu.update_container()

        for lk in self.links:
            link = self.links[lk]
            link.reparentTo(node)

    @funlog
    def get_container(self):
        """Return the current container for the instance."""
        return self.container

    @funlog
    def dump_to_dict(self, out_dict):
        """
        Save the position, type and name of this model to a dictionary.

        Dump all the relevant visual information to the passed dictionary
        such that a dictionary representation of this model can be saved,
        restored, and rendered to its current state.
        """
        p = self.getPos()
        out_dict["position"] = (p[0], p[1], p[2])
        out_dict["type"] = self.__class__.__name__.lower()
        if "name" not in out_dict:
            out_dict["name"] = self.uid

    @funlog
    def move_cam_to(self):
        """Reposition the camera, and move to correct container to look at this model."""

        pos = self.getWorldPos()
        cm.enter(self.container)

        self.app.anim_move_cam_to(pos)

    @funlog
    def add_decoration(self, name, model, delegate=None, tiptext=None, space=None):
        if self.decorationmanager is None:
            err("Trying to use un-initialized decoration manager. Probably an implementation error...")
            return

        if not self.menu.has_item('menutoggledecorations'):
            @event.callback
            def show_decorations(*args, **kwargs):
                warn("==== Calling VisNode Toggle Decoration Visibility")
                if self.decorationmanager.toggle_all_visible():
                    self.repoint_to_delegates()
                else:
                    self.repoint_from_delegates()
            self.menu.add_item(MenuItem('menutoggledecorations',
                                        show_decorations,
                                        title="Toggle Decorations"))

        return self.decorationmanager.add_decoration(name, model, delegate=delegate, tiptext=tiptext, space=space)

    @funlog
    def clear_delegates(self):
        if self.decorationmanager is not None:
            self.decorationmanager.clear()
        self.delegates.clear()

    @funlog
    def add_delegate(self, name, model, rule_type, direction, extra=None, tiptext=None, space=None):
        """
        Create a new decoration and assign it as the delegate for a connection type.

        This creates a new delegate. If the same decoration is to be used for
        multiple delegations, use also delegate.
        """
        dec = self.add_decoration(name, model, delegate=self, tiptext=tiptext, space=space)

        self.also_delegate(dec, rule_type, direction, extra=extra)

        return dec

    @funlog
    def show_decoration(self, name):
        if self.decorationmanager is None:
            err("Trying to use un-initialized decoration manager. Probably an implementation error...")
            return

        self.decorationmanager.show(name)
        self.repoint_to_delegates()

    @funlog
    def hide_decoration(self, name):
        if self.decorationmanager is None:
            err("Trying to use un-initialized decoration manager. Probably an implementation error...")
            return

        self.decorationmanager.hide(name)
        self.repoint_from_delegates()

    @funlog
    def set_delegated_conn(self, delegate, conn):
        if not hasattr(delegate, '_edges'):
            delegate._edges = []
        if conn not in delegate._edges:
            delegate._edges.append(conn)

    @funlog
    def rem_delegated_conn(self, conn):
        if hasattr(self, '_edges') and conn in self._edges:
            self._edges.remove(conn)

    @funlog
    def get_delegated_conns(self):
        if hasattr(self, '_edges'):
            return self._edges
        else:
            return []

    @funlog
    def add_to_map(self):
        """
        LOOKS DEPRECATED!!

        Add associated logical construct to the map.

        The code below is how this visual artifact is linked to
        the logical node in the graph (as stored in the Map)

        mapentry is the reference variable in the visual construct to the
        logical data structure from walujapi
        """
        raise Exception("callbackmanager.add_to_map should not have been called...")
        import nord.sigurd.shell.vismap as vm
        try:
            set_find_mapentry(self)
            vm.addNode(self)

            for func in self.onMappedFuncs:
                func()
        except UseException:
            # These setattr's allow flow edges to connect to
            # the file* visual nodes...
            setattr(self, 'mapentry', use.module('nord.nord.Node')(vm.get_working_map()))
            setattr(self.mapentry, 'id', ng.id())
            err("No corresponding logical element found for visual\
                 artifact of type: {}".format(self.__class__.__name__))

    @funlog
    def handle_mapentry_data(self, payload):
        """
        This function should be overridden by
        implementing extensions.
        """
        verb(f"{self.mapentry} sent back '{payload}'', but no handler implemented")

    @funlog
    def mapentry_comm(self, payload):
        """
        Send the provided payload to the backend.
        Provide the current map if it has changed.
        Return the response to the caller (probably a
        descendent instance)
        """
        # Have to import vmap here so as to prevent
        # circular import of this module via model
        import nord.sigurd.shell.vismap as vmap
        from nord.wrigan.visibles.utils import communicate, show_error
        map_id = vmap.synchronize()
        if map_id is None:
            show_error("Unable to Synchronize Map!!!")
            return

        if not hasattr(self, 'mapentry'):
            show_error("IMPLEMENTATION ERROR: Node Missing Map Entry")
        data = communicate(map_id, self.mapentry.id, payload)
        if data is not None:
            self.handle_mapentry_data(data)

    @funlog
    def handle_map_node_update(self, update):
        """
        To be overridden by elements which need to additionally process
        backend map (mapentry) changes
        """
        pass

    @funlog
    def handle_map_edge_src_create(self, edge):
        """
        To be overridden by elements which need to additionally process
        new edges leaving this node as provided by backend
        """
        pass

    @funlog
    def handle_map_edge_tgt_create(self, edge):
        """
        To be overridden by elements which need to additionally process
        new edges targetting this node as provided by backend
        """
        pass

    @funlog
    def handle_map_edge_src_delete(self, edge):
        """
        To be overridden by elements which need to additionally process
        deleted edges leaving this node as indicated by backend
        """
        pass

    @funlog
    def handle_map_edge_tgt_delete(self, edge):
        """
        To be overridden by elements which need to additionally process
        deleted edges targetting this node as provided by backend
        """
        pass

    @funlog
    def demap(self):
        """
        Remove the connection between the map entity
        and the visible one, most likely for the purposes
        of removing nodes and edges from a graph
        """
        if hasattr(self, "mapentry"):
            if hasattr(self.mapentry, "visualization"):
                del self.mapentry.__dict__["visualization"]
            delattr(self, "mapentry")
