"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Manage all connections between nodes.

This module facilitates the creation and rendering of
edges between nodes.
"""


import sys
from nord.sigurd.utils.log import err, info
from nord.sigurd.utils.verboser import funlog
from nord.sigurd.shell.connection import Connection
import nord.sigurd.shell.application as app
import nord.sigurd.managers.event as event
import nord.sigurd.shell.vismap as vm
import nord.sigurd.managers.objectmanager as om


def _add_missing(dct, name, value):
    if name not in dct:
        dct[name] = value


class ConnectionManager(object):
    """
    This class keeps track of all connections (edges).

    It primarily facilitates the steps needed to
    draw the conneciton between any nodes.
    """

    @funlog
    def __init__(self):
        """The constructor for the connection manager class."""
        self.targets = {}
        self.sources = {}
        self.types = {}
        self.pendingSource = None
        self.pendingCtype = None
        self.__all__ = []  # needed to allow VisManager to
        #                    use SpaceManager as it looks for __all__

    @event.callback
    @funlog
    def start_conn(self, source, ctype):
        """
        Initiate the connection process.

        Called to set the state up wth the source
        of the edge, with the expectation that finish
        connection will be called next.
        """
        self.pendingSource = source
        self.pendingCtype = ctype
        """
            With the changes to submenu and menu.py, either adding:

            "
            , 'model_start_connection': {
                'new_intercept': 'state_data_start_connection'
            }"

            To the statemanager dictionary for 'state_main_menu_visible'

            or the line below, allows the complete_conn method to be called.

            It seems that state_main_menu_visible should not be in effect
            when start_conn is called by the submenu...

            And, that the line below would also be un-necessary

        """
    @event.callback
    @funlog
    def complete_conn(self, tgt, ignore_me):
        src = self.pendingSource
        ctype = self.pendingCtype
        if src is not None:
            if not vm.addConnection(src, tgt, ctype):
                err("Cannot connect {} to {} with {}"
                    .format(src.__class__.__name__,
                            tgt.__class__.__name__,
                            ctype))
        self.pendingCtype = None
        self.pendingSource = None

    @funlog
    def render_completed_conn(self, src, tgt, ctype, edge):
        """
        Create the connection (edge) between the nodes.

        This expects that start_conn was called. This
        uses vismap (and thereby, generally, the connection class)
        to make sure the requested connection can be made.
        """
        rv = None
        """
            here the visual mapping is requested to add the connection between
            the passed nodes if the mapentries in the src and tgt are elegible
            to be connected, vismap returns True
        """
        conn = Connection(ctype, src, tgt, app, edge)
        _add_missing(self.sources, src.uid, {})
        _add_missing(self.sources[src.uid], tgt.uid, {})
        _add_missing(self.sources[src.uid][tgt.uid], ctype, [])

        _add_missing(self.targets, tgt.uid, {})
        _add_missing(self.targets[tgt.uid], src.uid, {})
        _add_missing(self.targets[tgt.uid][src.uid], ctype, [])

        _add_missing(self.types, ctype, {})
        _add_missing(self.types[ctype], src.uid, {})
        _add_missing(self.types[ctype][src.uid], tgt.uid, [])

        self.sources[src.uid][tgt.uid][ctype].append(conn)
        self.targets[tgt.uid][src.uid][ctype].append(conn)
        self.types[ctype][src.uid][tgt.uid].append(conn)
        conn.calcPos()
        rv = conn

        conn.mapentry = edge
        edge.set_untracked_attr('visualization', conn)

        if hasattr(edge, 'delegate'):
            if edge.delegate["direction"] == "as_src":
                if src.delegates == {} and hasattr(src, 'establish_delegates'):
                    # Source may not yet have interrogated their contents
                    # to determine if they have any delegates before
                    # the connection to the container is being established
                    src.establish_delegates()
                for delegate in src.delegates['as_src'][edge.rule]:
                    if delegate.uid.startswith(edge.delegate["uid"]):
                        src.set_delegated_conn(delegate, conn)
                        conn.move_source_to(delegate)
            if edge.delegate["direction"] == "as_tgt":
                if tgt.delegates == {} and hasattr(tgt, 'establish_delegates'):
                    # container's may not yet have interrogated their contents
                    # to determine if they have any delegates before
                    # the connection to the container is being established
                    tgt.establish_delegates()
                for delegate in tgt.delegates['as_tgt'][edge.rule]:
                    if delegate.uid.startswith(edge.delegate["uid"]):
                        tgt.set_delegated_conn(delegate, conn)
                        conn.move_target_to(delegate)

        if hasattr(src, 'on_connect'):
            src.on_connect(conn)

        if hasattr(tgt, 'on_connect'):
            tgt.on_connect(conn)

        return rv

    @funlog
    def cancel_conn(self, *args, **kwargs):
        """Cancel the connection and reset state."""
        self.pendingSource = None
        self.pendingCtype = None

    @funlog
    def draw(self, src, tgt, ctype, edge):
        """Wrapper to be called on load."""
        info(f"Rendering connection of type {ctype} between {src} and {tgt}.")
        self.start_conn(event.EventInstance('connectionmanager', src, ctype))
        self.complete_conn(
            event.EventInstance('connectionmanager', tgt, None))
        return self.render_completed_conn(src, tgt, ctype, edge)

    @funlog
    def remove(self):
        """Remove the passed edge's visualization attributes."""
        err("Should this method be doing something??? (connectionmanager.remove)")
        err("Should this method be doing something??? (connectionmanager.remove)")
        err("Should this method be doing something??? (connectionmanager.remove)")

    def __getattr__(self, name):
        """
        Allow draw_* named methods to be called.

        This will then extract the rule from the
        requested name and pass it as an argument
        to the draw method such that the correct
        visualization is applied.
        """
        if not name.startswith('draw_') and name in dir(sys.modules[__name__]):
            return getattr(sys.modules[__name__], name)
        elif name.startswith('draw_'):
            def wrapped_draw_conn(edge):
                ctype = '_'.join(name.split('_')[1:])
                src = om.fetch(edge.source.id)
                tgt = om.fetch(edge.target.id)
                return self.draw(src, tgt, ctype, edge)
            return wrapped_draw_conn
        else:
            raise AttributeError('No connection manager attribute named: {}'
                                 .format(name))


sys.modules[__name__] = ConnectionManager()
