"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Manage nested container'self camera + physics engine for each level.
"""


from nord.sigurd.utils.verboser import funlog
from panda3d.core import NodePath
from panda3d.bullet import BulletWorld
import nord.sigurd.managers.event as event
from nord.sigurd.shell.worldview import WorldView
import nord.sigurd.managers.objectmanager as om
from nord.sigurd.managers.placementmanager import gen_dumb_placer
from panda3d.core import PGTop


_current = None
_containers = {}
_current_name = None
_stack = None


@funlog
def clear():
    global _current, _current_name, _containers, _stack
    _current = None
    _containers = {}
    _current_name = None
    _stack = None


clear()


@funlog
def prepare():
    """
    Initialize everything.

    This cannot be called at module load time due to how other modules
    depend upon this one to be importable, and when the application(ShowBase)
    constructor is executed as render et al haven't yet been laoded to builtins
    """
    global _stack, _current, _current_name, _containers
    world = BulletWorld()
    placer = gen_dumb_placer()
    _containers['__top_of_the_world__'] = (render, world, render2d, None, placer)  # noqa: F821
    _stack = [('__top_of_the_world__', render, world, render2d, None, placer)]  # noqa: F821
    _current_name = '__top_of_the_world__'
    _current = (render, world, render2d, None, placer)  # noqa: F821


@funlog
def add_if_missing(container_name):  # noqa: N802
    """Add name to list of containers if it is not already present."""
    global _stack, _current, _current_name, _containers
    if container_name not in _containers:
        add(container_name)


@funlog
def add(container_name):
    """Create node and physics world and add under given name."""
    global _stack, _current, _current_name, _containers
    np = NodePath("__nested_render__{}".format(container_name))
    world = BulletWorld()

    a2d = NodePath(PGTop("__nested_aspect2d__{}".format(container_name)))
    a2d.setDepthTest(False)
    a2d.setDepthWrite(False)
    # a2d.setScale(1.0 / base.getAspectRatio(), 1.0, 1.0)
    # pgt = a2d.attachNewNode(PGTop("__nested_pgt__{}".format(container_name)))
    # pgt.setScale(1.0 / base.getAspectRatio(), 1.0, 1.0)

    # to facilitate headless testing, do not call the
    # following if the mouseWatcherNode is None
    if base.mouseWatcherNode is not None:
        a2d.node().setMouseWatcher(base.mouseWatcherNode)  # noqa: F821

    _containers[container_name] = (np, world, a2d, None, gen_dumb_placer())
    # Render the fog
    WorldView.set_fog(np, base)  # noqa: F821
    # draw the grid
    WorldView.draw_grid(np)

    # set up the lighting
    WorldView.set_lighting(np)


@funlog
def has(container_name):
    """Check if the named container already exists."""
    global _containers
    return container_name in _containers


@funlog
def current_name():
    """Return the name of the current container."""
    global _current_name
    return _current_name


@event.callback
@funlog
def enter(container_name):
    """Set everything up to be in the named container."""
    global _stack, _current, _current_name, _containers
    if _current is None:
        prepare()
    _stack.append((_current_name,
                   _current[0],
                   _current[1],
                   _current[2],
                   _current[3],
                   _current[4]))
    if container_name is None:
        container_name = _current_name
    _current = _containers[container_name]
    _current_name = container_name
    reparent_cam()

    if om.has(container_name):
        vis_container = om.fetch(container_name)
        vis_container.enter()


@event.callback
@funlog
def exit(*args, **kwargs):
    """Pop out to the outer container from the current one."""
    global _stack, _current, _current_name, _containers
    if om.has(_current_name):
        vis_container = om.fetch(_current_name)
        vis_container.exit()

    if len(_stack) > 1:
        _current_name, current_render, current_world, current_a2d, current_pgt, placer =\
            _stack.pop()
        _current = (current_render, current_world, current_a2d, current_pgt, placer)
    else:
        _current = _containers['__top_of_the_world__']
        _current_name = '__top_of_the_world__'
    reparent_cam()


@funlog
def is_top():
    """Return true if this is thou outermost container."""
    global _stack, _current, _current_name, _containers
    if _current is None:
        prepare()
    return _current == _containers['__top_of_the_world__']


@funlog
def next_loc():
    global _current
    return next(_current[4])


@funlog
def get_current_name():
    """Return the name of the current container."""
    global _stack, _current, _current_name, _containers
    return _current_name


@funlog
def reparent_cam():
    """Move the camera into the current container."""
    global _stack, _current, _current_name, _containers
    camera.reparentTo(_current[0])  # noqa: F821
    base.camera2d.reparentTo(_current[2])  # noqa: F821
    current_pgt = _current[3]
    if current_pgt is not None:
        current_pgt.node().setMouseWatcher(base.mouseWatcherNode)  # noqa: F821

    import nord.sigurd.shell.mainmenu as mainmenu
    mainmenu.update_container()


@funlog
def add_to_current(model):
    """Add passed model to the current container."""
    global _stack, _current, _current_name, _containers
    p = model.getPos()
    if  p[0] == 0.0 and p[1] == 0.0 and p[2] == 0.0:
        p = next_loc()
        model.mapentry.position = p
    model.reparentTo(_current[0])
    if hasattr(model, 'physics'):
        _current[1].attachRigidBody(model.physics)


# @  funlog
def cwc():
    """
    Return the current container's render method.

    Shorthand for Current Working Container.
    """
    # allows for run to have been called first...
    global _stack, _current, _current_name, _containers
    if _current is None:
        prepare()
    return _current[0]


# @  funlog
def get_by_name(name):
    """Return the "render" node for the named container."""
    global _containers
    rv = _containers.get(name, None)
    if rv is not None:
        return rv[0]
    else:
        return None


# @   funlog
def cur_world():
    """Return the current container's physics world."""
    # allows for run to have been called first...
    global _stack, _current, _current_name, _containers
    if _current is None:
        prepare()
    return _current[1]


# @  funlog
def c2d():
    """Return the current container's aspect2d."""
    global _stack, _current, _current_name, _containers
    if _current is None:
        prepare()
    return _current[2]


"""
    The idea here is to manage container entry and exit, but it may be that the
    best place for that code is within the container object itself...

    we'd still need a global means of accessing the current working container
    though...




"""
