"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""


from nord.sigurd.utils.log import verb, note, err, warn, info
from nord.sigurd.utils.verboser import funlog

_event_count = 0


class EventInstance(object):
    @funlog
    def __init__(s, callername, *args, **kwargs):
        global _event_count
        _event_count += 1
        s.event_id = _event_count
        s.caller = callername
        verb("   ******      Created new event with id {} for <{}>   ******".format(s.event_id, s.caller))
        s.setArgs(*args, **kwargs)
        s.injects = {}

    @funlog
    def __eq__(s, other):
        return s.event_id == other.event_id

    @funlog
    def setArgs(s, *args, **kwargs):
        s.args = args
        s.kwargs = kwargs

    @funlog
    def apply(s, function, *args, **kwargs):
        if len(args) >= 1 and isinstance(args[0], EventInstance):
            if args[0].event_id == s.event_id:
                note("Event self called, but not yet unwrapped: id: {}".format(s.event_id))
                return args[0].apply(function, *s.args, **s.kwargs)
            else:
                note("Event ID {} called event: {}".format(s.event_id, args[0].event_id))
                return args[0].apply(function)
        try:
            rv = function(*s.args, **s.kwargs)
            verb("    ^^^^^^^  Event ID: {} for <{}> applied args ({}, {}) and executed callback which returned: {}   ^^^^^^".format(s.event_id, s.caller, s.args, s.kwargs, rv))
            return rv
        except TypeError as e:  # noqa: F841
            warn("Failed to apply event instance arguments ({}, {}) to {} named <{}>".format(s.args, s.kwargs, function, s.caller))
            rv = function(*args, **kwargs)
            verb("    ^*^*^*^  Event ID: {} for <{}> passed through args ({}, {}) and executed callback which returned: {}   ^*^*^*^".format(s.event_id, s.caller, args, kwargs, rv))
            # Comment out the raise if you'd like to have a bitch of a time 
            # troubleshooting events....
            raise e
            # return rv

    @funlog
    def inject(s, name):
        s.injects[name] = True

    @funlog
    def alreadyInjected(s, name):
        return name in s.injects


def callback(func):
    """
        This is a decorator used to pull the argument data out of the event instance passed to the
        func (the implemented callback function)
    """
    @funlog
    def wrapper(*args, **kwargs):
        if len(args) > 2:
            for arg in args:
                if isinstance(arg, EventInstance):
                    err("Too many nested callback's detected, raising exception")
                    raise Exception("ImplementationError: ({}) received too many parameters: {} {}".format(func, args, kwargs))

            warn("-v-v-v-  Function ({}) flagged as event.callback called with non-EventInstance parameters: {} {} -v-v-v-".format(func, args, kwargs))
            rv = func(*args, **kwargs)
            verb("    ^_^_^_^  Unidentified Event passed args ({}, {}) and executed callback which returned: {}   ^_^_^_^".format(args, kwargs, rv))
            return rv

            # err("Too many nested callback's detected, raising exception")
            # raise Exception("ImplementationError: ({}) received too many parameters: {} {}".format(func, args, kwargs))

        if len(args) > 1:
            ei = args[len(args) - 1]
        elif len(args) > 0:
            ei = args[0]
        else:
            ei = None
        if not isinstance(ei, EventInstance):
            warn("-v-v-v-  Function ({}) flagged as event.callback called with non-EventInstance parameters: {} {} -v-v-v-".format(func, args, kwargs))
            rv = func(*args, **kwargs)
            verb("    ^_^_^_^  Unidentified Event passed args ({}, {}) and executed callback which returned: {}   ^_^_^_^".format(args, kwargs, rv))
            return rv
            #raise Exception("ImplementationError: Function ({}) flagged as event.callback called with non-EventInstance parameters: {} {}".format(func, args, kwargs))

        # get any earlier arguments (like self...) and insert them to the arguments...
        newArgs = ()
        for a in args[:-1]:
            newArgs+= (a,)
        for a in ei.args:
            newArgs+= (a,)
        ei.args = newArgs

        rv =  ei.apply(func, *args, **kwargs)

        # should the same event trigger multiple object member methods, the following is 
        # to keep them from accumulating in the eventinsntance's args field
        if len(args) > 1:
            info("Omitting {} for <{}> from the EventInstance's argument list".format(ei.caller, ei.args[0]))
            ei.args = ei.args[1:]

        return rv
    return wrapper


@funlog
def noop(*args, **kwargs):
    pass
