"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Abstraction layer for arbitrary function (or event" names.

This is the stateless event translation layer. This module
allows consuming modules to call an event by name via this module like:

    import nord.sigurd.managers.eventmap as em
    em.some_random_event

and this module will look for a configured mapping named
"some_random_event" in the configuration file, and, if
present, call the configured module+method.
"""


import nord.nord.utils.SpaceManager as use
from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import info, error, tb
import sys


_eventMap = {}
_modulesLoaded = {}


@funlog
def _load(fname, space="nord"):
    """
    Populate the _eventMap dictionary based upon the contents of fname.

    The file format is:

    <event_name>:<module>.<function>
    <event_name>:<module>.<function>
    <event_name>:<module>.<function>

    This is the main initialization function. It is called by Application.
    """
    global _eventMap, _modulesLoaded
    # for k in list(_eventMap.keys()):
    #     del _eventMap[k]
    clear()
    append(fname, space)


@funlog
def clear():
    """Clear all mappings."""
    global _eventMap, _modulesLoaded
    _eventMap.clear()
    _modulesLoaded.clear()


@funlog
def append(fname, space="nord"):
    """Load the passed file and add to the current maps."""
    with open(fname, 'r') as f:
        for l in f.readlines():
            line = l.strip()
            if ':' not in line or '.' not in line:
                continue
            proc_line(line, space)


def proc_line(line, space="nord"):
    event_name, funpath = line.split(":", 2)
    arr = funpath.split('.')
    funmod, funname = ".".join(arr[:-1]), arr[-1]
    try:
        info("Loading function: '{}' from module: '{}' for event: '{}'"
             .format(funname, funmod, event_name))
        if funmod not in _modulesLoaded:
            # _modulesLoaded[funmod] = importlib.import_module(funmod)
            use.space(space)
            _modulesLoaded[funmod] = use.module(funmod)

        func = getattr(sys.modules[funmod], funname)
        internal_load_event_handler(event_name, func)

    except Exception as e:
        error("Unable to map event '{}' to '{}' due to: {}"
              .format(event_name, funpath, str(e)))
        last_type, last_value, exc_traceback = sys.exc_info()
        tb(exc_traceback, last_type, last_value)


def internal_load_event_handler(event_name, func):
    # Prevent loading the same function twice into an event
    # handler. This situation can arise when a space is loaded
    # both by a map, and, the configuration.
    if event_name not in _eventMap:
        _eventMap[event_name] = []
    if func not in _eventMap[event_name]:
        _eventMap[event_name].append(func)


@funlog
def _placeholder(name):
    error("Unmapped event: {} requested".format(name))

    def func(*args, **kwargs):
        pass
    return func


@funlog
def wrap(funclist):
    """
    Wrap the list of event methods, calling them in the order they were added.

    if there is only one method to return non-None, then that
    value is returned. If more than one of the methods returns
    non-None, a list of the non-None return value is returned

    The reality is that this approach has some non-linked return
    values to a bunch of function calls. Yet, event calls probably
    should not return anything to begin with....
    """
    def rv(*args, **kwargs):
        ret = None
        for f in funclist:
            r = f(*args, **kwargs)
            if r is not None:
                if ret is not None and type(ret) == list:
                    ret.append(r)
                elif ret is not None:
                    t = []
                    t.append(ret)
                    t.append(r)
                else:
                    ret = r
        return ret
    return rv


def test_use_only_inject_handler(name, func):
    internal_load_event_handler(name, func)


#  @ funlog  # possibly another funlog call could be employed, but the default breaks things.
def __getattr__(name):  # noqa: N802
    """
    The 'secret sauce' of this module.

    This allows consuming modules to request arbitrary methods from
    this module, and this module will translate them accord to the
    configuration directives in the provided config file.
    """
    global _eventMap, _modulesLoaded
    if name == 'load':
        return _load
    elif name in ['append', 'clear'] or name.startswith('test_use_only'):
        return globals().get(name)
    elif name.startswith('__'):
        return globals().get(name)
    else:
        if name in _eventMap:
            return wrap(_eventMap[name])
        else:
            return _placeholder(name)
