"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The model manager deals with two things.

1) The mapping of node names from the nord module (nodes being the
   primary constructs of the language) to the visual representations.
2) The abstraction of visual representation name to the actual model file loaded
"""

from direct.actor.Actor import Actor
from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import err, warn, info, notice
import os
from nord.sigurd.config.globalVars import DEFAULTMODELDIR, FALLBACKMODEL
import nord.nord.utils.SpaceManager as use
from nord.nord.exceptions import UseException as UseException


_mapping = {}


@funlog
def clear():
    """Clear out any existing configuration."""
    global _mapping
    _mapping.clear()


@funlog
def load(fname, space="nord"):
    """Fresh populate the lookup dictionary with files that exist."""
    clear()
    append(fname, space=space)


@funlog
def append(fname, space="nord"):
    """Append to the lookup dictionary with files that exist in the named space."""
    with open(fname, 'r') as f:
        for l in f.readlines():
            line = l.strip()
            if ':' not in line or '!' not in line:
                err("Invalid model map line: {}".format(line))
                continue
            model_name, mdlpath = line.split(":", 2)
            use.space(space)
            try:
                mdlpath = use.path(mdlpath.replace('!', os.path.sep))
            except UseException:
                # Fallback and check in the default namespace for the model.
                # if it is not here, exceptions may follow
                use.space("nord")
                mdlpath = use.path(mdlpath.replace('!', os.path.sep))
                use.space(space)

            """
                Check if the requested model path exists
                if not, check in the default model directory (assuming the requested path is not
                the default)
                lastly, just use the fallback model
            """

            if not os.path.isfile(mdlpath):
                warn("Model Manager unable to find requested model file: {}".format(mdlpath))
                mdlfname = mdlpath.split(os.path.sep)[-1]
                mdlpath = "{}{}{}".format(DEFAULTMODELDIR, os.path.sep, mdlfname)

            if space not in _mapping:
                _mapping[space] = {}
            _mapping[space][model_name] = mdlpath
            info("Model Manager setting '{}' as the model file for '{}'"
                 .format(mdlpath, model_name))
        if space not in _mapping:
            _mapping[space] = {}


@funlog
def lookup(name, space):
    """In a space aware fashion, return the mapped name if it exists."""
    if space in _mapping and\
            name in _mapping[space]:
        return _mapping[space][name]
    elif "nord" in _mapping and name in _mapping["nord"]:
        return _mapping["nord"][name]
    else:
        notice("Using fallback model for: '{}.{}'".format(space, name))
        oldspace = use.current_space()
        use.space("nord")
        rv = use.path(FALLBACKMODEL)
        use.space(oldspace)
        return rv


@funlog
def exists(name, space="nord"):
    """Check to see if the named model has a mapping in the requested, or default, space."""
    return space in _mapping and\
        name in _mapping[space] or (
            "nord" in _mapping and
            name in _mapping["nord"]
        )


@funlog
def load_model(name, space):
    """
    Load the requested model.

    If the name doesn't exist in the passed space,
    try the default, "nord", and if it is not in
    either location, load and return the default
    model.
    """
    return Actor(lookup(name, space), {"animate": lookup(name, space)})
