"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import *
from nord.sigurd.config.globalVars import *
import sys
import nord.sigurd.managers.vismanager as vm

class Precedence(object):
    @funlog
    def __init__(s):
        s._precedents = {'__catchall__': s.allow}
        s._loaded = False

    @funlog
    def load(s, fname):
        s._precedents.clear()
        s._precedents = {'__catchall__': s.allow}

        with open(fname, 'r') as f:
            for l in f.readlines():
                line = l.strip()
                if ':' not in line:
                    continue
                e1name, e2name, rule,  = line.split(":", 3)
                if hasattr(s, rule):
                    s._precedents['{}::{}'.format(e1name, e2name)] = getattr(s, rule)
                else:
                    err('Unknown precedence rule: {} reqested for: {}::{} in file: {}'.format(rule, e1name, e2name, rule, fname))

    @funlog
    def allow(s, func):
        def precedence_allowed(*args, **kwargs):
            return func(*args, **kwargs)
        return precedence_allowed

    @funlog
    def block(s, func):
        def precedence_denied(*args, **kwargs):
            return False
        return precedence_denied

    @funlog
    def call(s, name, func):

        @funlog
        def precedence_wrap(edge):
            """
                Look at the source and target node's visualizations
                check to see if there are more than just the named 
                edges between the source and the target

                if so, look for any precedence rules governing any of the 
                edges.

                if there are any precedence rules, apply them all, and 
                mark those edges as applied
            """

            if hasattr(edge, '_visual_precedence_applied'):
                verb("precedence of {} already processed: {}".format(name, edge), depth=4)
                return None  # s.block(func)

            blocked = False
            related_edges = []

            #find all the rules/edges that this edge's source also point to the same target and have precedence over the named rule
            for rule in edge.get_source()._targetNodes[edge.get_target()]:
                verb("Evaluating precedence of: {}\n{}".format(rule, edge.get_source()._targetNodes[edge.get_target()][rule]))
                #if rule == name:
                #	verb("Precedence rule matching self: {} == {}".format(rule, name))
                #	continue
                for rel_edge in edge.get_source()._targetNodes[edge.get_target()][rule]:

                    if "{}::{}".format(rule, name) in s._precedents and s._precedents["{}::{}".format(rule, name)] == 'precedes':
                        note("Found preceding edge {}".format(rel_edge), depth=4)
                        related_edges.append((rule, rel_edge))
                    elif "{}::{}".format(rule, name) in s._precedents and s._precedents["{}::{}".format(rule, name)] == s.block:
                        blocked = True
                        note("Found blocking edge: {}".format(rel_edge), depth=4)
                    else:
                        note("Didn't do anything with: {} when evaluating: {}::{} not in {}".format(rel_edge, rule, name, s._precedents))


            # applying any precedes/succeeds precedents 
            # and mark all the edges with _visual_precedence_applied
            for r1, e1 in related_edges:

                f = getattr(vm, 'rawget_{}'.format(r1))
                f(e1)
                e1._visual_precedence_applied = True

            rv = None
            if not blocked:
                rv = func(edge)
            edge._visual_precedence_applied = True
            return rv


        return precedence_wrap

sys.modules[__name__] = Precedence()
