"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Intercept event requests and apply state manipulation to them.

This module pre-processes events through the intercepts dictionary.
"""

import nord.sigurd.managers.eventmap as em
from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.utils.log import info, debug, err, verb, warn
from nord.sigurd.managers.event import EventInstance

"""
    States are a pre-cursor to the event map. Input handlers will call the
    state manager, which, if it has a state represented for the named input,
    will operate upon that state accordingly, which in turn then calls the
    event mapper. Should there not be any state operations to be called, it
    will just pass through to the event mapper.

    The state manager operates by intercepting the triggered event name,
    and then either:
       1. replacing it with a new event name
       2. injecting an additional event to trigger
       3. just calling the requested event name

    An intercept can set a new next interceptor which is then applied to the
    next event recieved


"""
_bypass_state = {
    'mouse_over': None,
    'mouse_was_over': None
}

_intercepts = {
    'default': {
        'injects': None,
        'intercepts': {
            'mouse_freespace_click': {
                'replacement': 'show_main_menu',
                'new_intercept': 'state_main_menu_visible'
            },
            'data_show_menu': {
                'new_intercept': 'state_data_menu_visible'
            },
            'edge_show_menu': {
                'new_intercept': 'state_edge_menu_visible'
            },
            'show_sub_menu': {
                'new_intercept': 'state_sub_menu_visible'
            },
            'model_start_connection': {
                'new_intercept': 'state_data_start_connection'
            },
            'toggle_text_edit': {
                'new_intercept': 'state_text_edit_mode'
            },
            'key-control-d': {
                'replacement': 'program_exit'
            }
        }
    },
    'state_main_menu_visible': {
        'injects': 'hide_main_menu',
        'intercepts': {
            'mouse_freespace_click': {
                'replacement': 'hide_main_menu',
                'new_intercept': 'default'
            },
            'model_start_connection': {
                'new_intercept': 'state_data_start_connection'
            },
            'clickable_object_clicked': {  # This is needed to prevent menuitem clicks being
                                           # proc'd by__catchall__
                'replacement': 'clickable_object_clicked'
            },
            'show_sub_menu': {
                'new_intercept': 'state_sub_menu_visible'
            },
            '__catchall__': {
                'new_intercept': 'default',
                'replacement': 'noop'

            }
        }
    },
    'state_sub_menu_visible': {
        'injects': 'hide_sub_menu',
        'intercepts': {
            'mouse_freespace_click': {
                'replacement': 'hide_sub_menu',
                'new_intercept': 'default'
            },
            'clickable_object_clicked': {  # This is needed to prevent menuitem clicks being
                                           # proc'd by__catchall__
                'replacement': 'clickable_object_clicked'
            },
            'show_sub_menu': {
                'new_intercept': 'state_sub_menu_visible'
            },
            '__catchall__': {
                'new_intercept': 'state_sub_menu_visible',
                'replacement': 'hide_model_menu'

            }
        }
    },
    'state_data_menu_visible': {
        'injects': 'hide_model_menu',
        'intercepts': {
            'mouse_freespace_click': {
                'replacement': 'hide_model_menu',
                'new_intercept': 'default'
            },
            'toggle_text_edit': {
                'new_intercept': 'state_text_edit_mode'
            },
            'model_start_connection': {
                'new_intercept': 'state_data_start_connection'
            },
            'clickable_object_clicked': {
                'replacement': 'clickable_object_clicked'
            },
            '__catchall__': {
                'new_intercept': 'default',
                'replacement': 'noop'

            }
        }
    },
    'state_edge_menu_visible': {
        'injects': 'hide_model_menu',
        'intercepts': {
            'mouse_freespace_click': {
                'replacement': 'hide_model_menu',
                'new_intercept': 'default'
            },
            'clickable_object_clicked': {
                'replacement': 'clickable_object_clicked'
            },
            '__catchall__': {
                'new_intercept': 'default',
                'replacement': 'noop'

            }
        }
    },
    'state_data_start_connection': {
        'intercepts': {
            'clickable_object_clicked': {
                'replacement': 'model_finish_connection'
            },
            'hide_model_menu': {  # deal with the injected hide menu stuff
                'new_intercept': 'state_data_start_connection'
            },
            'hide_main_menu': {   # deal with the injected hide menu stuff
                'new_intercept': 'state_data_start_connection'
            },
            'mouse_freespace_click': {
                'replacement': 'model_cancel_connection',
                'new_intercept': 'default'  # catch freespace click
                                            # and resets the connection manager
            },
            'mouse_selected_dragging': {
                'replacement': 'hide_main_menu',
                'new_intercept': 'state_data_start_connection'  # this is
                #  needed because start_drag is called before screen_touch
            },
            '__catchall__': {
                'replacement': 'model_cancel_connection'
            }
        }
    },
    'state_text_edit_mode': {
        'intercepts': {
            'key-escape': {
                'replacement': 'toggle_text_edit',
                'new_intercept': 'default'
            },
            '__catchall__': {
                'replacement': 'toggle_text_edit',
                'new_intercept': 'default'
            },
            'mouse_freespace_click': {
                'replacement': 'toggle_text_edit',
                'new_intercept': 'default'
            },
            'clickable_object_clicked': {
                'replacement': 'toggle_text_edit',
                'new_intercept': 'default'
            },
            'toggle_text_edit': {
              'new_intercept': 'default'
            },
            'fbyf_left_pressed': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_left_released': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_right_pressed': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_right_released': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_up_pressed': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_up_released': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_down_pressed': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'fbyf_down_released': {
              'replacement': 'noop',
              'new_intercept': 'state_text_edit_mode'
            },
            'key-a': {'replacement': 'text-a',
                      'new_intercept': 'state_text_edit_mode'},
            'key-b': {'replacement': 'text-b',
                      'new_intercept': 'state_text_edit_mode'},
            'key-c': {'replacement': 'text-c',
                      'new_intercept': 'state_text_edit_mode'},
            'key-d': {'replacement': 'text-d',
                      'new_intercept': 'state_text_edit_mode'},
            'key-e': {'replacement': 'text-e',
                      'new_intercept': 'state_text_edit_mode'},
            'key-f': {'replacement': 'text-f',
                      'new_intercept': 'state_text_edit_mode'},
            'key-g': {'replacement': 'text-g',
                      'new_intercept': 'state_text_edit_mode'},
            'key-h': {'replacement': 'text-h',
                      'new_intercept': 'state_text_edit_mode'},
            'key-i': {'replacement': 'text-i',
                      'new_intercept': 'state_text_edit_mode'},
            'key-j': {'replacement': 'text-j',
                      'new_intercept': 'state_text_edit_mode'},
            'key-k': {'replacement': 'text-k',
                      'new_intercept': 'state_text_edit_mode'},
            'key-l': {'replacement': 'text-l',
                      'new_intercept': 'state_text_edit_mode'},
            'key-m': {'replacement': 'text-m',
                      'new_intercept': 'state_text_edit_mode'},
            'key-n': {'replacement': 'text-n',
                      'new_intercept': 'state_text_edit_mode'},
            'key-o': {'replacement': 'text-o',
                      'new_intercept': 'state_text_edit_mode'},
            'key-p': {'replacement': 'text-p',
                      'new_intercept': 'state_text_edit_mode'},
            'key-q': {'replacement': 'text-q',
                      'new_intercept': 'state_text_edit_mode'},
            'key-r': {'replacement': 'text-r',
                      'new_intercept': 'state_text_edit_mode'},
            'key-s': {'replacement': 'text-s',
                      'new_intercept': 'state_text_edit_mode'},
            'key-t': {'replacement': 'text-t',
                      'new_intercept': 'state_text_edit_mode'},
            'key-u': {'replacement': 'text-u',
                      'new_intercept': 'state_text_edit_mode'},
            'key-v': {'replacement': 'text-v',
                      'new_intercept': 'state_text_edit_mode'},
            'key-w': {'replacement': 'text-w',
                      'new_intercept': 'state_text_edit_mode'},
            'key-x': {'replacement': 'text-x',
                      'new_intercept': 'state_text_edit_mode'},
            'key-y': {'replacement': 'text-y',
                      'new_intercept': 'state_text_edit_mode'},
            'key-z': {'replacement': 'text-z',
                      'new_intercept': 'state_text_edit_mode'},
            'key-shift-a': {'replacement': 'text-shift-a',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-b': {'replacement': 'text-shift-b',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-c': {'replacement': 'text-shift-c',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-d': {'replacement': 'text-shift-d',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-e': {'replacement': 'text-shift-e',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-f': {'replacement': 'text-shift-f',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-g': {'replacement': 'text-shift-g',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-h': {'replacement': 'text-shift-h',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-i': {'replacement': 'text-shift-i',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-j': {'replacement': 'text-shift-j',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-k': {'replacement': 'text-shift-k',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-l': {'replacement': 'text-shift-l',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-m': {'replacement': 'text-shift-m',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-n': {'replacement': 'text-shift-n',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-o': {'replacement': 'text-shift-o',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-p': {'replacement': 'text-shift-p',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-q': {'replacement': 'text-shift-q',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-r': {'replacement': 'text-shift-r',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-s': {'replacement': 'text-shift-s',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-t': {'replacement': 'text-shift-t',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-u': {'replacement': 'text-shift-u',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-v': {'replacement': 'text-shift-v',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-w': {'replacement': 'text-shift-w',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-x': {'replacement': 'text-shift-x',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-y': {'replacement': 'text-shift-y',
                            'new_intercept': 'state_text_edit_mode'},
            'key-shift-z': {'replacement': 'text-shift-z',
                            'new_intercept': 'state_text_edit_mode'},
            'key-space': {'replacement': 'text-space',
                          'new_intercept': 'state_text_edit_mode'},
            'key-return': {'replacement': 'text-return',
                           'new_intercept': 'state_text_edit_mode'},
            'key-arrow_up': {'replacement': 'text-up',
                             'new_intercept': 'state_text_edit_mode'},
            'key-arrow_down': {'replacement': 'text-down',
                               'new_intercept': 'state_text_edit_mode'},
            'key-arrow_left': {'replacement': 'text-left',
                               'new_intercept': 'state_text_edit_mode'},
            'key-arrow_right': {'replacement': 'text-right',
                                'new_intercept': 'state_text_edit_mode'},
            'key-backspace': {'replacement': 'text-backspace',
                              'new_intercept': 'state_text_edit_mode'},
            'key-delete': {'replacement': 'text-delete',
                           'new_intercept': 'state_text_edit_mode'},
            'key-tilde': {'replacement': 'text-tilde',
                          'new_intercept': 'state_text_edit_mode'},
            'key-backtick': {'replacement': 'text-backtick',
                             'new_intercept': 'state_text_edit_mode'},
            'key-bang': {'replacement': 'text-bang',
                         'new_intercept': 'state_text_edit_mode'},
            'key-at': {'replacement': 'text-at',
                       'new_intercept': 'state_text_edit_mode'},
            'key-octothorpe': {'replacement': 'text-octothorpe',
                               'new_intercept': 'state_text_edit_mode'},
            'key-dollar': {'replacement': 'text-dollar',
                           'new_intercept': 'state_text_edit_mode'},
            'key-percent': {'replacement': 'text-percent',
                            'new_intercept': 'state_text_edit_mode'},
            'key-caret': {'replacement': 'text-caret',
                          'new_intercept': 'state_text_edit_mode'},
            'key-ampersand': {'replacement': 'text-ampersand',
                              'new_intercept': 'state_text_edit_mode'},
            'key-asterisk': {'replacement': 'text-asterisk',
                             'new_intercept': 'state_text_edit_mode'},
            'key-leftparen': {'replacement': 'text-leftparen',
                              'new_intercept': 'state_text_edit_mode'},
            'key-rightparen': {'replacement': 'text-rightparen',
                               'new_intercept': 'state_text_edit_mode'},
            'key-leftcurly': {'replacement': 'text-leftcurly',
                              'new_intercept': 'state_text_edit_mode'},
            'key-rightcurly': {'replacement': 'text-rightcurly',
                               'new_intercept': 'state_text_edit_mode'},
            'key-1': {'replacement': 'text-1',
                      'new_intercept': 'state_text_edit_mode'},
            'key-2': {'replacement': 'text-2',
                      'new_intercept': 'state_text_edit_mode'},
            'key-3': {'replacement': 'text-3',
                      'new_intercept': 'state_text_edit_mode'},
            'key-4': {'replacement': 'text-4',
                      'new_intercept': 'state_text_edit_mode'},
            'key-5': {'replacement': 'text-5',
                      'new_intercept': 'state_text_edit_mode'},
            'key-6': {'replacement': 'text-6',
                      'new_intercept': 'state_text_edit_mode'},
            'key-7': {'replacement': 'text-7',
                      'new_intercept': 'state_text_edit_mode'},
            'key-8': {'replacement': 'text-8',
                      'new_intercept': 'state_text_edit_mode'},
            'key-9': {'replacement': 'text-9',
                      'new_intercept': 'state_text_edit_mode'},
            'key-0': {'replacement': 'text-0',
                      'new_intercept': 'state_text_edit_mode'},
            'key-leftsquare': {'replacement': 'text-leftsquare',
                               'new_intercept': 'state_text_edit_mode'},
            'key-rightsquare': {'replacement': 'text-rightsquare',
                                'new_intercept': 'state_text_edit_mode'},
            'key-doublequote': {'replacement': 'text-doublequote',
                                'new_intercept': 'state_text_edit_mode'},
            'key-leftangle': {'replacement': 'text-leftangle',
                              'new_intercept': 'state_text_edit_mode'},
            'key-rightangle': {'replacement': 'text-rightangle',
                               'new_intercept': 'state_text_edit_mode'},
            'key-singlequote': {'replacement': 'text-singlequote',
                                'new_intercept': 'state_text_edit_mode'},
            'key-comma': {'replacement': 'text-comma',
                          'new_intercept': 'state_text_edit_mode'},
            'key-period': {'replacement': 'text-period',
                           'new_intercept': 'state_text_edit_mode'},
            'key-forwardslash': {'replacement': 'text-forwardslash',
                                 'new_intercept': 'state_text_edit_mode'},
            'key-equals': {'replacement': 'text-equals',
                           'new_intercept': 'state_text_edit_mode'},
            'key-backslash': {'replacement': 'text-backslash',
                              'new_intercept': 'state_text_edit_mode'},
            'key-question': {'replacement': 'text-question',
                             'new_intercept': 'state_text_edit_mode'},
            'key-plus': {'replacement': 'text-plus',
                         'new_intercept': 'state_text_edit_mode'},
            'key-pipe': {'replacement': 'text-pipe',
                         'new_intercept': 'state_text_edit_mode'},
            'key-dash': {'replacement': 'text-dash',
                         'new_intercept': 'state_text_edit_mode'},
            'key-underscore': {'replacement': 'text-underscore',
                               'new_intercept': 'state_text_edit_mode'},
            'key-semicolon': {'replacement': 'text-semicolon',
                              'new_intercept': 'state_text_edit_mode'},
            'key-colon': {'replacement': 'text-colon',
                          'new_intercept': 'state_text_edit_mode'}
        }
    }
}


def internal_initialize_statemanager():
    global _intercepts, _current, _states, _listeners, _bypass_state

    _states = {
        "current": None,
        "intercept": _intercepts['default'],
        "pendingParams": {},
        "bypassers": _bypass_state
    }

    _current = 'default'
    _listeners = {}


internal_initialize_statemanager()


@funlog
def listen(name, listener_id, callback):
    """Allow consumers register their callback to run on an event."""
    global _intercepts, _current, _states, _listeners
    if name not in _listeners:
        _listeners[name] = {}
    _listeners[name][listener_id] = callback


@funlog
def unlisten(name, listener_id):
    """Allow consumers to un-register their callback from an event."""
    global _intercepts, _current, _states, _listeners
    if name in _listeners and listener_id in _listeners[name]:
        del _listeners[name]

    if name in _listeners and len(_listeners[name].keys()) == 0:
        del _listeners[name]


@funlog
def call_listeners(name):
    """Execute the listener callbacks for the named event."""
    global _intercepts, _current, _states, _listeners
    if name in _listeners:
        for k in _listeners[name]:
            cbk = _listeners[name][k]
            debug("Calling listener {}:{}".format(name, k))
            cbk()


@funlog
def set_next_state(name):
    """Detect named event's interception and replace if should."""
    global _intercepts, _current, _states, _listeners
    if name in _intercepts:
        _states['intercept'] = _intercepts[name]
        _current = name


@funlog
def clear_pending(name):
    """Remove any pending parameters from the current state."""
    global _intercepts, _current, _states, _listeners
    if name in _states['pendingParams']:
        del _states['pendingParams'][name]


@funlog
def post_pending(name, *args, **kwargs):
    """Set the pending arguments for the state manager to apply."""
    global _intercepts, _current, _states, _listeners
    _states['pendingParams'][name] = (args, kwargs)


@funlog
def apply_pending(func, name):
    """Grab pending parameters and call the passed function with them."""
    global _intercepts, _current, _states, _listeners
    if name in _states['pendingParams']:
        args, kwargs = _states['pendingParams'][name]
        del _states['pendingParams'][name]

        def rv(*a, **k):
            return func(EventInstance(name, *args, **kwargs))
        return rv
    else:
        err("Impossibly got to apply_pending without any pendingParams" +
            " for event: {}".format(name))


@funlog
def inject(name, function):
    """
    Grab the function name from the event mapper.

    Thencall it before calling the passed 'function'
    """
    global _intercepts, _current, _states, _listeners
    debug("Event State Machine injecting event '{}' with wrapper"
          .format(name))

    def rv(ei):
        debug("Event State Machine wrapper calling: {}\
               before original function".format(name))
        if not ei.alreadyInjected(name):
            ei.inject(name)
            f1 = getattr(em, name)
            if name in _states['pendingParams']:
                f1 = apply_pending(f1, name)
            f1(ei)

        call_listeners(name)
        return function(ei)

    return rv


@funlog
def introduce_intercept(name, intercept):
    global _intercepts
    if name in _intercepts:
        warn(f"Not replacing intercept {name}")
    else:
        _intercepts[name] = intercept


@funlog
def set_state_intercept(name):
    global _intercepts, _states, _current
    if name not in _intercepts:
        err(f"No such intercept named {name}")
    _states['intercept'] = _intercepts[name]
    _current = name


@funlog
def get_intercept_named(name):
    global _intercepts
    if name not in _intercepts:
        err(f"Cannot fetch intercept, none named {name}")
        return {}
    return _intercepts[name].copy()


@funlog
def process_intercept(intercept, name):
    """
    Apply the mappings identified by the passed intercept.

    This is to set the new function() and next intercept.
    """
    global _intercepts, _current, _states, _listeners
    if 'replacement' in intercept:
        debug("Event State Machine Intercept replacing '{}' with '{}'"
              .format(name, intercept['replacement']))
        name = intercept['replacement']

    if 'new_intercept' in intercept and intercept['new_intercept'] in _intercepts:
        info("Event State Machine setting new intercept named: {}"
             .format(intercept['new_intercept']))
        _states['intercept'] = _intercepts[intercept['new_intercept']]
        _current = intercept['new_intercept']
    else:
        debug("Event State Machine resetting intercept to default")
        _states['intercept'] = _intercepts['default']
        _current = 'default'

    f = getattr(em, name)

    return f, name


@funlog
def process_state(name):
    """Employ the intercept dictionary to read+change state as needed."""
    global _intercepts, _current, _states, _listeners
    f = None
    orig_name = name

    intercept = None
    if 'intercept' in _states:
        intercept = _states['intercept']

    if intercept is None:
        warn("state['intercept': {}".format(_states['intercept']))
    else:
        warn("intercept: {}".format(intercept))

    if intercept is not None \
            and 'intercepts' in intercept \
            and name in intercept['intercepts']:
        """
            the current state has an intercept defined, apply it
        """
        debug(f"Event State Machine [{_current}] found intercept for '{name}'")
        f, name = process_intercept(intercept['intercepts'][name],
                                    name)

    elif intercept is not None \
            and 'intercepts' in intercept \
            and '__catchall__' in intercept['intercepts']:

        debug((f"Event State Machine [{_current}] found catchall intercept, "
               f"using it instead of: '{name}'"))
        f, name = process_intercept(intercept['intercepts']
                                    [
                                    '__catchall__'
                                    ], '__catchall__')

    elif intercept is None:
        """
            Apply the default intercept
        """
        debug((f"Event State Machine [{_current}] did not find intercept for '{name}' "
               ", applying default"))
        f, name = process_intercept(_intercepts['default'], name)
    else:

        verb((f"Event State Machine [{_current}] fell through for '{name}',"
              f" applying default {intercept}"))
        f = getattr(em, name)

    if name in _states['pendingParams']:
        f = apply_pending(f, name)

    if intercept is not None \
            and 'injects' in intercept \
            and intercept['injects'] is not None:
        f = inject(intercept['injects'], f)
        debug(f"Event State Machine [{_current}] injected event named: {intercept['injects']}")

    if name != orig_name:
        call_listeners(name)

    return f


def __repr__():  # noqa: N802
    """Simplify the string representation of the state manager."""
    # return "<StateManager(current: '{}', {} intercepts, {} listeners)>"\
    #     .format(_current,
    #             len(_states['intercept'].keys()),
    #             len(_listeners))
    global _intercepts, _current, _states, _listeners
    import pprint
    return "<StateManager(\n\tcurrent: '{}',\n\t{} listeners,\n\tintercepts:\n{})>"\
        .format(_current,
                len(_listeners),
                pprint.pformat(_states['intercept'], indent=8))


def load(arg):
    """Placeholder for load."""
    em.load(arg)


def test_use_only_get_internal_vars():
    global _intercepts, _current, _states, _listeners
    return _intercepts, _current, _states, _listeners


#  @  funlog
def __getattr__(name):  # noqa: N802
    """Allow runtime attribute requests to be evaluated via intercepts."""
    global _intercepts, _current, _states, _listeners
    if name in ('set_next_state',
                'listen',
                'unlisten',
                'introduce_intercept',
                'set_state_intercept',
                'get_intercept_named',
                'test_use_only_get_internal_vars',
                'internal_initialize_statemanager'):
        return globals().get(name, None)
    elif name == '__catchall__':
        pass
    elif name.startswith('__') and not name == '__getattr__':
        return globals().get(name, None)
    elif name in ['clear', 'append']:
        return getattr(em, name)
    elif name in _states['bypassers']:
        # bypassers jump directly to the eventmanager
        return getattr(em, name)

    f = process_state(name)

    if name in ('_load',):
        return f

    @funlog
    def state_mgr_wrapped(*args, **kwargs):
        call_listeners(name)

        if len(args) != 1 or not isinstance(args[0], EventInstance):
            return f(EventInstance(name, *args, **kwargs))
        else:
            return f(*args, **kwargs)

    return state_mgr_wrapped
