"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Manage placement and visibility of VisNode decorations.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from enum import Enum, auto


class Attachment(Enum):
    """
    Enumerate the possible points around a model where decorations may reside.

    Also, facilitate the conversion of the attachment points to world coordinates.
    """
    FRONT_TOP_LEFT = auto()
    FRONT_TOP_CENTER = auto()
    FRONT_TOP_RIGHT = auto()
    FRONT_MIDDLE_RIGHT = auto()
    FRONT_BOTTOM_RIGHT = auto()
    FRONT_BOTTOM_CENTER = auto()
    FRONT_BOTTOM_LEFT = auto()
    FRONT_MIDDLE_LEFT = auto()

    MIDDLE_TOP_LEFT = auto()
    MIDDLE_TOP_CENTER = auto()
    MIDDLE_TOP_RIGHT = auto()
    MIDDLE_MIDDLE_RIGHT = auto()
    MIDDLE_BOTTOM_RIGHT = auto()
    MIDDLE_BOTTOM_CENTER = auto()
    MIDDLE_BOTTOM_LEFT = auto()
    MIDDLE_MIDDLE_LEFT = auto()

    REAR_TOP_LEFT = auto()
    REAR_TOP_CENTER = auto()
    REAR_TOP_RIGHT = auto()
    REAR_MIDDLE_RIGHT = auto()
    REAR_BOTTOM_RIGHT = auto()
    REAR_BOTTOM_CENTER = auto()
    REAR_BOTTOM_LEFT = auto()
    REAR_MIDDLE_LEFT = auto()

    FRONT = auto()
    MIDDLE = auto()
    REAR = auto()
    TOP = auto()
    BOTTOM = auto()
    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()

    @staticmethod
    @funlog
    def to_components(att):
        A = Attachment
        return {
            A.FRONT_TOP_LEFT: (A.FRONT, A.TOP, A.LEFT),
            A.FRONT_TOP_CENTER: (A.FRONT, A.TOP, A.CENTER),
            A.FRONT_TOP_RIGHT: (A.FRONT, A.TOP, A.RIGHT),
            A.FRONT_MIDDLE_LEFT: (A.FRONT, A.MIDDLE, A.LEFT),
            A.FRONT_MIDDLE_RIGHT: (A.FRONT, A.MIDDLE, A.RIGHT),
            A.FRONT_BOTTOM_LEFT: (A.FRONT, A.BOTTOM, A.LEFT),
            A.FRONT_BOTTOM_CENTER: (A.FRONT, A.BOTTOM, A.CENTER),
            A.FRONT_BOTTOM_RIGHT: (A.FRONT, A.BOTTOM, A.RIGHT),
            A.MIDDLE_TOP_LEFT: (A.MIDDLE, A.TOP, A.LEFT),
            A.MIDDLE_TOP_CENTER: (A.MIDDLE, A.TOP, A.CENTER),
            A.MIDDLE_TOP_RIGHT: (A.MIDDLE, A.TOP, A.RIGHT),
            A.MIDDLE_MIDDLE_LEFT: (A.MIDDLE, A.MIDDLE, A.LEFT),
            A.MIDDLE_MIDDLE_RIGHT: (A.MIDDLE, A.MIDDLE, A.RIGHT),
            A.MIDDLE_BOTTOM_LEFT: (A.MIDDLE, A.BOTTOM, A.LEFT),
            A.MIDDLE_BOTTOM_CENTER: (A.MIDDLE, A.BOTTOM, A.CENTER),
            A.MIDDLE_BOTTOM_RIGHT: (A.MIDDLE, A.BOTTOM, A.RIGHT),
            A.REAR_TOP_LEFT: (A.REAR, A.TOP, A.LEFT),
            A.REAR_TOP_CENTER: (A.REAR, A.TOP, A.CENTER),
            A.REAR_TOP_RIGHT: (A.REAR, A.TOP, A.RIGHT),
            A.REAR_MIDDLE_LEFT: (A.REAR, A.MIDDLE, A.LEFT),
            A.REAR_MIDDLE_RIGHT: (A.REAR, A.MIDDLE, A.RIGHT),
            A.REAR_BOTTOM_LEFT: (A.REAR, A.BOTTOM, A.LEFT),
            A.REAR_BOTTOM_CENTER: (A.REAR, A.BOTTOM, A.CENTER),
            A.REAR_BOTTOM_RIGHT: (A.REAR, A.BOTTOM, A.RIGHT)
        }.get(att, (A.FRONT, A.TOP, A.LEFT))

    @staticmethod
    @funlog
    def to_pos(model, att, tier=1):
        A = Attachment

        zsel, ysel, xsel = A.to_components(att)

        ymlt = {A.FRONT: -1, A.MIDDLE: 0, A.REAR: 1}.get(zsel)
        zmlt = {A.TOP: 1, A.MIDDLE: 0, A.BOTTOM: -1}.get(ysel)
        xmlt = {A.LEFT: -1, A.CENTER: 0, A.RIGHT: 1}.get(xsel)

        mn, mx = model.getTightBounds()
        length = (mx - mn).length()/3.2 * tier

        center = model.get_nodep().get_bounds().get_center()

        x = center.getX() + xmlt * length
        y = center.getY() + ymlt * length
        z = center.getZ() + zmlt * length

        return x, y, z


class DecorationManager(object):
    """Manage the placement of decorations around a VisNode."""

    @funlog
    def __init__(self, decoratee, app, space="nord"):
        """Initialize the manager."""
        self.decoratee = decoratee
        self.app = app
        self.decorations = dict()
        self.space = space
        self.all_visible = False
        self.positions = {
            Attachment.FRONT_TOP_LEFT: [],
            Attachment.FRONT_TOP_CENTER: [],
            Attachment.FRONT_TOP_RIGHT: [],
            Attachment.FRONT_MIDDLE_LEFT: [],
            Attachment.FRONT_MIDDLE_RIGHT: [],
            Attachment.FRONT_BOTTOM_LEFT: [],
            Attachment.FRONT_BOTTOM_CENTER: [],
            Attachment.FRONT_BOTTOM_RIGHT: [],
            Attachment.MIDDLE_TOP_LEFT: [],
            Attachment.MIDDLE_TOP_CENTER: [],
            Attachment.MIDDLE_TOP_RIGHT: [],
            Attachment.MIDDLE_MIDDLE_LEFT: [],
            Attachment.MIDDLE_MIDDLE_RIGHT: [],
            Attachment.MIDDLE_BOTTOM_LEFT: [],
            Attachment.MIDDLE_BOTTOM_CENTER: [],
            Attachment.MIDDLE_BOTTOM_RIGHT: [],
            Attachment.REAR_TOP_LEFT: [],
            Attachment.REAR_TOP_CENTER: [],
            Attachment.REAR_TOP_RIGHT: [],
            Attachment.REAR_MIDDLE_LEFT: [],
            Attachment.REAR_MIDDLE_RIGHT: [],
            Attachment.REAR_BOTTOM_LEFT: [],
            Attachment.REAR_BOTTOM_CENTER: [],
            Attachment.REAR_BOTTOM_RIGHT: []
        }

    @funlog
    def size(self):
        """Return the number of decorations managed."""
        return len(self.decorations)

    @funlog
    def clear(self):
        """Remove all managed decorations."""
        remkeys = []
        for dk in self.decorations.keys():
            remkeys.append(dk)
        for rk in remkeys:
            self.rem_decoration(rk)
        self.decorations.clear()

    @funlog
    def get_next_open_position(self, tier=1):
        """Find the next open attachment position and level."""
        cnt = tier
        while True:
            for pos in Attachment:
                if pos == Attachment.FRONT:
                    break
                if len(self.positions[pos]) < cnt:
                    return pos, cnt
            cnt += 1

    @funlog
    def add_decoration(self, name, model, attachment=None, delegate=None, tiptext=None, space=None):
        """Add the named decoration, finding an attachment position as needed."""
        if attachment is None:
            attachment, tier = self.get_next_open_position()
        else:
            tier = len(self.positions[attachment])
            tier += 1

        wspace = self.space
        if space is not None:
            wspace = space

        from nord.sigurd.visibles.ui.decoration import Decoration

        dec = Decoration(name,
                         model,
                         attachment,
                         self.app,
                         self.decoratee,
                         tier=tier,
                         space=wspace,
                         delegate=delegate,
                         tiptext=tiptext)

        self.decorations[name] = dec
        self.positions[attachment].append(dec)
        self.decoratee.link(dec)
        return dec

    @funlog
    def rem_decoration(self, name, rule=None, direction=None):
        """Remove the named decoration."""
        dec = self.decorations.get(name, None)
        if dec is not None:
            dec.remove()
            del self.decorations[name]
            self.positions[dec.attachment].remove(dec)
            self.decoratee.unlink(dec)
        if rule and direction and direction in self.decoratee.delegates and \
                rule in self.decoratee.delegates[direction]:
            to_rem = []
            for itm in self.decoratee.delegates[direction][rule]:
                if itm.uid.startswith(name):
                    to_rem.append(itm)
            for itm in to_rem:
                self.decoratee.delegates[direction][rule].remove(itm)

    @funlog
    def toggle_all_visible(self):
        """Show/Hide all the decorations under management."""
        if not self.all_visible:
            for name in self.decorations:
                self.decorations[name].show()
            self.all_visible = True
        else:
            for name in self.decorations:
                self.decorations[name].hide()
            self.all_visible = False
        return self.all_visible

    @funlog
    def show(self, name):
        """Show the named decoration, if it exists."""
        dec = self.decorations.get(name, None)
        if dec is not None:
            dec.show()

    @funlog
    def hide(self, name):
        """Hide the named decoration, if it exists."""
        dec = self.decorations.get(name, None)
        if dec is not None:
            dec.hide()
