"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Meta-module which looks up requested visualizations from configured mappings.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import info, error, tb
import nord.sigurd.managers.precedence as precedence
import nord.nord.utils.SpaceManager as _use
import sys

_visualizationMap = {}
_modulesLoaded = {}


@funlog
def _load(fname, precedencefname):
    """
    Clear and populate the _visualizationMap dictionary based upon the contents of fname.

    The file format is:

    <edge_name>:<module>.<function>
    <edge_name>:<module>.<function>
    <edge_name>:<module>.<function>

    """
    clear()
    append(fname, precedencefname)


@funlog
def clear():
    """Clear out any existing configuration."""
    global _visualizationMap, _modulesLoaded
    _visualizationMap.clear()
    _modulesLoaded.clear()


@funlog
def append(fname, precedencefname):
    """Populate the _visualizationMap dictionary based upon the contents of fname."""
    global _visualizationMap, _modulesLoaded
    precedence.load(precedencefname)
    with open(fname, 'r') as f:
        for l in f.readlines():  # noqa: E741
            line = l.strip()
            if ':' not in line or '.' not in line:
                continue
            edge_name, funpath = line.split(":", 2)
            arr = funpath.split('.')
            funmod, funname = '.'.join(arr[:-1]), arr[-1]
            try:
                info("Loading function: '{}' from module: '{}' for visualization: '{}'"
                     .format(funname, funmod, edge_name))
                if funmod not in _modulesLoaded:
                    # _modulesLoaded[funmod] = importlib.import_module(funmod)
                    _use.space("nord")
                    _modulesLoaded[funmod] = _use.module(funmod)
                if edge_name not in _visualizationMap:
                    _visualizationMap[edge_name] = []
                if getattr(_modulesLoaded[funmod], funname) not in _visualizationMap[edge_name]:
                    _visualizationMap[edge_name].append(getattr(_modulesLoaded[funmod], funname))
                else:
                    error(f"Trying to duplicate load visualization function {funname} for edge {edge_name}")

            except Exception as e:
                error("Unable to map visualization '{}' to '{}' due to: {}"
                      .format(edge_name, funpath, str(e)))
                last_type, last_value, exc_traceback = sys.exc_info()
                tb(exc_traceback, last_type, last_value)


@funlog
def _placeholder(name):
    error("Unmapped visualization: {} requested".format(name))

    def func(*args, **kwargs):
        pass
    return func


@funlog
def _rawget(name):
    global _visualizationMap, _modulesLoaded

    def rv(edge):
        if name in _visualizationMap:
            return _visualizationMap[name]
        else:
            return _placeholder(name)


@funlog
def wrap(funclist, name):
    """
    Wrap the passed functions.

    This method wraps the list of visualization methods
    and calls each of them in the order they were added
    if there is only one method to return non-None, then that
    value is returned. If more than one of the methods returns
    non-None, a list of the non-None return value is returned

    The reality is that this approach has some non-linked return
    values to a bunch of function calls. Yet, visualization calls probably
    should not return anything to begin with....
    """
    global _visualizationMap, _modulesLoaded

    def rv(edge):
        ret = None
        for f in funclist:
            r = f(edge)
            if r is not None:
                if ret is not None and type(ret) == list:
                    ret.append(r)
                elif ret is not None:
                    t = []
                    t.append(ret)
                    t.append(r)
                else:
                    ret = r
        return ret
    return precedence.call(name, rv)


def __getattr__(name):
    global _visualizationMap, _modulesLoaded
    if name == 'load':
        return _load
    elif name.startswith('rawget_'):
        return _rawget(name.replace('rawget_', ''))
    elif name.startswith('__'):
        return globals().get(name)
    else:
        if name in _visualizationMap:
            return wrap(_visualizationMap[name], name)
        else:
            return _placeholder(name)
