"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
This is the main "entry point" for this program.
"""

from nord.sigurd.config.globalVars import MOVERATE,\
    DEFAULTCODEPATH,\
    DEFAULTPATHPATH,\
    DEFAULTSPACECONF,\
    ROTRATE,\
    ROTVRATE,\
    ROTHRATE,\
    DEFAULTFPS,\
    DEFAULTVISPRECEDENCEMAP,\
    PLACEDIST,\
    DEFAULTEVENTMAP,\
    DEFAULTMODELMAP,\
    DEFAULTVISMAP,\
    ROTINCR,\
    MOVEINCR,\
    DEFAULTSPACES,\
    DEFAULTSPACETITLES,\
    DEBUGCAMDISTANCE,\
    DEFAULTNUMSTEPS
import time

from math import pi, sin, cos
import sys
from direct.showbase.ShowBase import ShowBase
from direct.task import Task

from panda3d.core import CollisionTraverser,\
    CollisionHandlerQueue,\
    CollisionNode,\
    CollisionRay,\
    GeomNode,\
    Point3

from nord.sigurd.shell.worldview import WorldView
from nord.sigurd.handlers.mousehandler import MouseHandler
import nord.sigurd.handlers.keyboardhandler as keyboardhandler
import nord.sigurd.handlers.textinputhandler as textinputhandler
from nord.sigurd.utils.verboser import funlog, show_profile  # noqa: F401
from nord.sigurd.utils.log import debug, notice, err, info
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.selectionmanager as sm
import nord.sigurd.managers.event as event
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.vismanager as vm
from pathlib import Path
from nord.sigurd.shell.tooltip import expire_tooltips
from nord.sigurd.exceptions import UIException


# Set up the base paths for code and file/dir names
# so that subsequent calls to use can find the stuff they need
import nord.nord.utils.SpaceManager as use
for p in DEFAULTCODEPATH:
    use.add_location(p, 'code')
for p in DEFAULTPATHPATH:
    use.add_location(p, 'path')


class Application(ShowBase):
    """Eponymous class for this module."""

    @funlog
    def log_cam_vector(self):
        """Debug output function."""
        x1, y1, z1 = self.base.camera.getPos()
        x2, y2, z2 = self.base.camera.getPos() + self.base.camLens.getViewVector()
        debug("Camera pos: ({:.4f},{:.4f},{:.4f}) \
               looking at -> ({:.4f},{:.4f},{:.4f})"
              .format(x1, y1, z1, x2, y2, z2))

    @event.callback
    @funlog
    def accel_move_forward(self):
        """Signal decrease forward velocity."""
        if self.vel_forward <= MOVERATE:
            self.vel_forward += MOVEINCR
        self.go_foreback = True

    @event.callback
    @funlog
    def accel_move_backward(self):
        """Signal decrease backward velocity."""
        if self.vel_forward >= -1.0 * MOVERATE:
            self.vel_forward -= MOVEINCR
        self.go_foreback = True

    @event.callback
    @funlog
    def accel_move_down(self):
        """Increase downward velocity."""
        if self.vel_downward >= -1.0 * MOVERATE:
            self.vel_downward -= MOVEINCR
        self.go_updown = True

    @event.callback
    @funlog
    def accel_move_up(self):
        """Increase upward velocity."""
        if self.vel_downward <= MOVERATE:
            self.vel_downward += MOVEINCR
        self.go_updown = True

    @event.callback
    @funlog
    def accel_move_left(self):
        """Signal decrease leftward velocity."""
        if self.vel_rightward >= -1.0 * MOVERATE:
            self.vel_rightward -= MOVEINCR
        self.go_rightleft = True

    @event.callback
    @funlog
    def accel_move_right(self):
        """Signal decrease rightward velocity."""
        if self.vel_rightward <= MOVERATE:
            self.vel_rightward += MOVEINCR
        self.go_rightleft = True

    @event.callback
    @funlog
    def accel_rot_right(self):
        """Signal decrease rightward rotational velocity."""
        if self.vel_rotright <= ROTRATE:
            self.vel_rotright += ROTINCR
        self.turn_rightleft = True

    @event.callback
    @funlog
    def accel_rot_left(self):
        """Signal decrease leftward rotational velocity."""
        if self.vel_rotright >= -1.0 * ROTRATE:
            self.vel_rotright -= ROTINCR
        self.turn_rightleft = True

    @event.callback
    @funlog
    def accel_rot_up(self):
        """Signal decrease upward rotational velocity."""
        if self.vel_rotup <= ROTRATE:
            self.vel_rotup += ROTINCR
        self.turn_updown = True

    @event.callback
    @funlog
    def accel_rot_down(self):
        """Signal decrease downward rotational velocity."""
        if self.vel_rotup >= -1.0 * ROTRATE:
            self.vel_rotup -= ROTINCR
        self.turn_updown = True

    @event.callback
    @funlog
    def decel_rot_up(self):
        """Signal decrease upward rotational velocity."""
        self.turn_updown = False

    @event.callback
    @funlog
    def decel_rot_down(self):
        """Signal decrease downward rotational velocity."""
        self.turn_updown = False

    @event.callback
    @funlog
    def decel_rot_right(self):
        """Signal decrease rightward rotational velocity."""
        self.turn_rightleft = False

    @event.callback
    @funlog
    def decel_rot_left(self):
        """Signal decrease leftward rotational velocity."""
        self.turn_rightleft = False

    @event.callback
    @funlog
    def decel_move_forward(self):
        """Signal decrease forward velocity."""
        self.go_foreback = False

    @event.callback
    @funlog
    def decel_move_backward(self):
        """Signal decrease backward velocity."""
        self.go_foreback = False

    @event.callback
    @funlog
    def decel_move_up(self):
        """Signal decrease upward velocity."""
        self.go_updown = False

    @event.callback
    @funlog
    def decel_move_down(self):
        """Signal decrease downward velocity."""
        self.go_updown = False

    @event.callback
    @funlog
    def decel_move_left(self):
        """Signal decrease leftward velocity."""
        self.go_rightleft = False

    @event.callback
    @funlog
    def decel_move_right(self):
        """Signal decrease rightward velocity."""
        self.go_rightleft = False

    @event.callback
    @funlog
    def move_forward_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (0, MOVERATE, 0))
        self.base.camera.setPos(pos)
        self.log_cam_vector()

    @event.callback
    @funlog
    def move_back_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (0, -1.0 * MOVERATE, 0))
        self.base.camera.setPos(pos)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def move_left_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (-1.0 * MOVERATE, 0, 0))
        self.base.camera.setPos(pos)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def move_right_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (MOVERATE, 0, 0))
        self.base.camera.setPos(pos)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def move_down_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (0, 0, -1.0 * MOVERATE))
        self.base.camera.setPos(pos)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def move_up_camera(self):
        """Event execution method."""
        pos = cm.cwc().getRelativePoint(self.base.camera, (0, 0, MOVERATE))
        self.base.camera.setPos(pos)
        return

    @funlog
    def anim_move_cam_to(self, pos):
        """
        Calculate the number of steps to move to look at the passed
        postion, and add the appropriate iteration function and arguments list
        """
        point_list = list()
        a = base.camera.getPos()  # noqa: F821
        a = Point3(a[0], a[1] + DEBUGCAMDISTANCE, a[2])
        b = pos

        step_length = (a - b).length() / DEFAULTNUMSTEPS

        norm = (b - a).normalized()
        np = a
        for x in range(0, int(DEFAULTNUMSTEPS)):
            np += norm * step_length
            point_list.append(Point3(np))

        def do_move(points):
            if len(points) > 0:
                pt = points.pop(0)
                base.camera.setPos((pt[0], pt[1] - DEBUGCAMDISTANCE, pt[2]))  # noqa: F821
                base.camera.lookAt(pt)  # noqa: F821
            else:
                self.del_iter_event('move_cam_to')

        self.add_iter_event('move_cam_to', do_move, (list(point_list),))

    @event.callback
    @funlog
    def rotate_up_camera(self, rate=ROTVRATE):
        """Event execution method."""
        angle = self.base.camera.getP() + rate
        self.base.camera.setP(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def rotate_down_camera(self, rate=ROTVRATE):
        """Event execution method."""
        angle = self.base.camera.getP() - rate
        self.base.camera.setP(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def rotate_left_camera(self, rate=ROTHRATE):
        """Event execution method."""
        angle = self.base.camera.getH() + rate
        self.base.camera.setH(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def rotate_right_camera(self, rate=ROTHRATE):
        """Event execution method."""
        angle = self.base.camera.getH() - rate
        self.base.camera.setH(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def rotate_clockwise_camera(self):
        """Event execution method."""
        angle = self.base.camera.getR() + ROTRATE
        self.base.camera.setR(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def rotate_counterclock_camera(self):
        """Event execution method."""
        # While this rotation works, the other rotations are not affected by
        # changes in this rotation
        angle = self.base.camera.getR() - ROTRATE
        self.base.camera.setR(angle)
        self.log_cam_vector()
        return

    @event.callback
    @funlog
    def move_forward_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def move_back_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def move_left_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def move_right_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def move_up_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def move_down_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_up_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_down_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_left_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_right_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_clockwise_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @event.callback
    @funlog
    def rotate_counterclock_selected(self):
        raise UIException("NOT IMPLEMENTED")

    @funlog
    def get_onscreen_pos(self):
        """Return a set of coordinates that will be on camera."""
        return self.mouseHandler.get_last_origin()

    #  @  funlog
    def move_camera(self):
        """Apply the various velocities, as needed to the camera."""
        if abs(self.vel_rotright) > 0.0:
            self.rotate_right_camera()
        if abs(self.vel_rotup) > 0.0:
            self.rotate_up_camera()
        if abs(self.vel_forward) > 0.0:
            self.move_forward_camera()
        if abs(self.vel_rightward) > 0.0:
            self.move_right_camera()
        if abs(self.vel_downward) > 0.0:
            self.move_down_camera()

        if not self.go_rightleft and abs(self.vel_rightward) > 0.0:
            if self.vel_rightward > MOVEINCR:
                self.vel_rightward -= MOVEINCR
            if self.vel_rightward < MOVEINCR:
                self.vel_rightward += MOVEINCR
            if abs(self.vel_rightward) <= MOVEINCR * 1.9:
                self.vel_rightward = 0.0

        if not self.go_updown and abs(self.vel_downward) > 0.0:
            if self.vel_downward > MOVEINCR:
                self.vel_downward -= MOVEINCR
            if self.vel_downward < MOVEINCR:
                self.vel_downward += MOVEINCR
            if abs(self.vel_downward) <= MOVEINCR * 1.9:
                self.vel_downward = 0.0

        if not self.go_foreback and abs(self.vel_forward) > 0.0:
            if self.vel_forward > MOVEINCR:
                self.vel_forward -= MOVEINCR
            if self.vel_forward < MOVEINCR:
                self.vel_forward += MOVEINCR
            if abs(self.vel_forward) <= MOVEINCR * 1.9:
                self.vel_forward = 0.0

        if not self.turn_rightleft and abs(self.vel_rotright) > 0.0:
            if self.vel_rotright > ROTINCR:
                self.vel_rotright -= ROTINCR
            if self.vel_rotright < ROTINCR:
                self.vel_rotright += ROTINCR
            if abs(self.vel_rotright) <= ROTINCR * 1.9:
                self.vel_rotright = 0.0

        if not self.turn_updown and abs(self.vel_rotup) > 0.0:
            if self.vel_rotup > ROTINCR:
                self.vel_rotup -= ROTINCR
            if self.vel_rotup < ROTINCR:
                self.vel_rotup += ROTINCR
            if abs(self.vel_rotup) <= ROTINCR * 1.9:
                self.vel_rotup = 0.0

    @event.callback
    @funlog
    def stop_camera_abrupt(self):
        """Immediately cease all camera movement."""
        self.vel_forward = 0.0
        self.vel_downward = 0.0
        self.vel_rightward = 0.0
        self.vel_rotright = 0.0
        self.vel_rotup = 0.0
        self.turn_updown = False
        self.turn_rightleft = False
        self.go_foreback = False
        self.go_updown = False
        self.go_rightleft = False

    @event.callback
    @funlog
    def move_relative_selected(self, mpos):
        """Event execution method."""
        # moves the selected objects relative to the
        # camera'self viewing plane....
        sm.apply(lambda x: x.camPlaneMove(self.base.camera, mpos))

    @event.callback
    @funlog
    def orbit_up_camera(self):
        """Event execution method."""
        pass

    @event.callback
    @funlog
    def orbit_down_camera(self):
        """Event execution method."""
        pass

    @event.callback
    @funlog
    def orbit_left_camera(self):
        """Event execution method."""
        angle_deg = 6.0
        angle_rad = angle_deg * (pi / 180.0)
        self.base.camera.setPos(-20 * sin(angle_rad), 20.0 * cos(angle_rad), 3)
        self.base.camera.setHpr(angle_deg, 0, 0)

    @event.callback
    @funlog
    def orbit_right_camera(self):
        """Event execution method."""
        angle_deg = 6.0
        angle_rad = angle_deg * (pi / 180.0)
        self.base.camera.setPos(20 * sin(angle_rad), -20.0 * cos(angle_rad), 3)
        self.base.camera.setHpr(angle_deg, 0, 0)

    @event.callback
    @funlog
    def re_center(self):
        """Event execution method."""
        self.base.camera.setPos(0, -1.0 * PLACEDIST, 0)
        self.base.camera.lookAt(0, 0, 0)
        self.log_cam_vector()

    @event.callback
    @funlog
    def delete_selected(self):
        """Delete the currently selected objects."""
        def remove(x):
            sm.deselect(x)
            x.remove()

        sm.apply(remove)

    @event.callback
    @funlog
    def load_space_from_browser(self, *args, **kwargs):
        """Open the file browser, and load the selected space."""
        if not hasattr(self, 'filebrowser'):
            use.space("fregata")
            mod = use.module('nord.fregata.visibles.filebrowser')
            fb = mod.FileBrowser(".", self)
            self.filebrowser = fb
        else:
            fb = self.filebrowser

        @funlog
        def on_browser_close(itms):
            """Given the selection in the browser."""
            before_spaces = list(use.get_spaces())

            for itm in itms:
                err(f"Trying to load space: {itm}")
                if Path(itm).is_dir() and Path(itm).joinpath('space.conf').is_file():
                    info(f"Found Space config file in {itm}, adding it's config.")
                    use.add_config(Path(itm).joinpath('space.conf').resolve(),
                                   prefix=itm, parent_too=True)

            for spc in use.get_spaces():
                if spc not in before_spaces:
                    notice(f"Loading space: {spc} from path: {itm}")
                    self.load_space_config(spc)
                    import nord.sigurd.shell.mainmenu as mm
                    mm.add_space_menu(spc, mm)
                else:
                    err(f"Skipping already loaded space: {spc}")

        fb.set_exit_callback(on_browser_close)
        fb.show()

    @funlog
    def load_space_config(self, space):
        """Load the config entries from the requested space."""
        import nord.sigurd.managers.modelmanager as mgr
        use.space(space)
        if not use.space_loaded(space):
            if use.file_exists(DEFAULTMODELMAP):
                mgr.append(use.path(DEFAULTMODELMAP), space=space)
            if use.file_exists(DEFAULTEVENTMAP):
                em.append(use.path(DEFAULTEVENTMAP), space=space)
            if use.file_exists(DEFAULTVISMAP) and use.file_exists(DEFAULTVISPRECEDENCEMAP):
                vm.append(use.path(DEFAULTVISMAP), use.path(DEFAULTVISPRECEDENCEMAP))

    @funlog
    def mapevents(self):
        """
        Call the major load and initialization functions.

        These modules are initialized in this fashion to
        avoid import loop ond similar issues. Very chicken vs egg.
        """
        import nord.sigurd.managers.modelmanager as mgr
        import nord.sigurd.shell.vismap as Map
        use.add_config(use.path(str(Path.joinpath(Path('nord/sigurd'), Path(DEFAULTSPACECONF)))))
        mgr.clear()
        em.clear()
        vm.clear()
        for space in use.get_spaces():
            self.load_space_config(space)

        use.space("nord")
        keyboardhandler.load(self)
        self.mouseHandler.load()
        import nord.sigurd.shell.mainmenu as mm
        mm.load(DEFAULTSPACETITLES)
        if not hasattr(self, "origin"):
            self.origin = mgr.load_model("origin", "nord")
            self.origin.reparentTo(cm.cwc())
            # self.origin.loop("animate")
        em.listen('program_exit', 'Base Application', Map.save)

    @funlog
    def set_keyboard_text_target(self, target):
        """Redirect the keyboard events to a different handler."""
        keyboardhandler.set_target(target)

    @funlog
    def clear_keyboard_text_target(self):
        """Restore the keyboard events to default handler."""
        keyboardhandler.set_target(None)

    @funlog
    def load_map(self, mapfile=None, bypass_cdc=True):
        """
        Load the graph in the passed file into the Map memory representation.

        Then calls the methods needed to render the visualization of the Map
        """
        import nord.sigurd.shell.vismap as Map
        if mapfile and Path(mapfile).exists():
            use.prepend_location(f'dir://{str(Path(mapfile).parent)}', 'path')
            use.prepend_location(f'dir://{str(Path(mapfile).parent)}', 'files')
        Map.load(mapfile, self, bypass_cdc=bypass_cdc)

    @event.callback
    @funlog
    def show_browser(self, *args, **kwargs):
        """Show the file browser."""
        if not hasattr(self, 'filebrowser'):
            use.space("fregata")
            mod = use.module('nord.fregata.visibles.filebrowser')
            fb = mod.FileBrowser(".", self)
            self.filebrowser = fb
        else:
            fb = self.filebrowser
        fb.show()

    @funlog
    def get_browser(self):
        """Return the instance of the filebrowser for interrogation."""
        if hasattr(self, 'filebrowser'):
            return self.filebrowser
        else:
            return None

    @funlog
    def __init__(self):
        """The application constructor."""
        self.vel_forward = 0.0
        self.vel_downward = 0.0
        self.vel_rightward = 0.0
        self.vel_rotright = 0.0
        self.vel_rotup = 0.0
        self.turn_updown = False
        self.turn_rightleft = False
        self.go_foreback = False
        self.go_updown = False
        self.go_rightleft = False

    def attach_default_spaces(self):
        for space in DEFAULTSPACES:
            use.space(space)
        use.space("nord")

    def prep_to_run(self):
        # to facilitate running in a test fixture at the function scope
        # since this method will be called repeatedly, for each test
        # in the module, we'll return so as to prevent trying to
        # re-initialize..... there may be side effects from this...
        if hasattr(self, 'base'):
            return
        self.base = ShowBase()

        self.mouseHandler = MouseHandler(base, self)  # noqa: F821
        # self.keyboardhandler as keyboardhandler = KeyboardHandler(self)
        textinputhandler.bootstrap(self)

        # Establish the fog
        WorldView.set_fog(cm.cwc(), base)  # noqa: F821

        # load the data egg
        self.re_center(event.EventInstance('Application'))

        # this is our means of abstracting specific inputs to specific
        # behaviors, so we could map the up direction key to strafe up,
        # or rotate the camera up, or to move forward
        self.key2action = dict()

        # set up the pointer to object "collision" detector
        self.pointerTraverser = CollisionTraverser("pointer traverser")
        self.picker_node = CollisionNode('mouseRay')
        self.pointerCollisionHandler = CollisionHandlerQueue()
        picker_np = self.base.camera.attachNewNode(self.picker_node)
        self.picker_node.setFromCollideMask(GeomNode.getDefaultCollideMask())
        self.pickerRay = CollisionRay()
        self.picker_node.addSolid(self.pickerRay)
        self.pointerTraverser.addCollider(picker_np,
                                          self.pointerCollisionHandler)

        # draw the grid
        WorldView.draw_grid(cm.cwc())

        # set up the lighting
        WorldView.set_lighting(cm.cwc())

        self.base.taskMgr.add(self.mouseHandler.mouse_task, "MouseMoveDetector")

    # Define a procedure to move the camera.
    @funlog
    def spin_cam_task(self, task):
        """Sample camera task based animation."""
        angle_deg = task.time * 6.0
        angle_rad = angle_deg * (pi / 180.0)
        self.base.camera.setPos(20 * sin(angle_rad), -20.0 * cos(angle_rad), 3)
        self.base.camera.setHpr(angle_deg, 0, 0)
        return Task.cont

    @event.callback
    @funlog
    def exit_program(self):
        """Quit the program."""
        notice("Exiting visual runtime...")
        show_profile()
        sys.exit(0)

    @event.callback
    @funlog
    def abort_program(self):
        """Exit the program with an error code."""
        err("Exiting visual runtime...")
        show_profile()
        sys.exit(-1)

    def add_iter_event(self, name, func, args):
        """Add named event + handler to the dictionary of events called per main loop iteration."""
        if not hasattr(self, '_iter_events'):
            self._iter_events = {}
        self._iter_events[name] = (func, args)

    def del_iter_event(self, name):
        """Delete the named event from the per app loop iteration set of functions."""
        if hasattr(self, '_iter_events') and name in self._iter_events:
            del self._iter_events[name]

    def iteration_events(self):
        """Call functions which have been requested to be called once per main loop iteration."""
        if hasattr(self, '_iter_events'):
            for k in self._iter_events.copy():
                func, args = self._iter_events[k]
                func(*args)

    def one_loop_iteration(self):
        dt = globalClock.getDt()  # noqa: F821
        cm.cur_world().doPhysics(dt)
        expire_tooltips()
        self.iteration_events()
        keyboardhandler.detect_down_dir_keys()
        # self.move_camera()

        self.base.taskMgr.step()  # noqa: F821

    @funlog
    def run(self, mapfile=None):
        """Prep to and run the main event loop."""
        self.attach_default_spaces()
        self.prep_to_run()
        self.mapevents()
        if mapfile:
            self.load_map(mapfile, bypass_cdc=False)
        else:
            self.show_browser('.')

        while True:
            self.one_loop_iteration()
            time.sleep(1.0 / DEFAULTFPS)

    @funlog
    def run_as_executable(self):

        if len(sys.argv) > 1:
            runmap = sys.argv[1]
        else:
            runmap = None

        self.run(mapfile=runmap)

    def _cleanup_instance(self):
        """
        This should only be called in testing.

        Thus, the leading _.
        """
        del self.base
        del sys.modules[__name__]
        sys.modules[__name__] = Application()


sys.modules[__name__] = Application()
