"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The underlying class providing base edge functionality.

There probably will not be any subclasses of this class.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.shell.connsrc import ConnSrc
from nord.sigurd.shell.connedge import ConnEdge
from nord.sigurd.shell.conntgt import ConnTgt
from nord.sigurd.shell.menuItem import MenuItem
import nord.nord.Receiver as Receiver
import nord.sigurd.utils.namegen as namegen
import nord.sigurd.managers.modelmanager as mm
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.event as event
from nord.sigurd.utils.log import warn
import nord.sigurd.shell.vismap as vm


def show_edge_menu(ei):
    """
    Module level function used by event system
    to show the menu for an edge.
    """
    edge = ei.args[0]
    origin = ei.args[1]
    app = ei.args[2]
    edge.show_menu(origin, app)


class Connection(object):
    """
    Connections are the visualization of edges between nodes.

    This interrogates the type to determine which models will be
    employed to visualize the connection between nodes.
    """

    @funlog
    def __init__(self, ctype, source, target, app, edge):
        """
        Constructor for class connection.

        Takes the connection type as ctype,
        the source and target models and the
        main application as arguments.
        """
        self.source = source
        self.target = target
        self.source_holder = None
        self.target_holder = None
        self.type = ctype
        self.selected = False
        self.name = namegen.get()
        from nord.sigurd.shell.modelmenu import ModelMenu
        self.menu = ModelMenu()

        sspace = "nord"
        tspace = "nord"
        espace = "nord"

        if hasattr(target, "mapentry"):
            if mm.exists('{}conntgt'.format(ctype), space=target.mapentry.get_space()):
                tspace = target.mapentry.get_space()
            if mm.exists('{}connedge'.format(ctype), space=target.mapentry.get_space()):
                espace = target.mapentry.get_space()

        if hasattr(source, "mapentry"):
            if mm.exists('{}connsrc'.format(ctype), space=source.mapentry.get_space()):
                sspace = source.mapentry.get_space()
            if mm.exists('{}connedge'.format(ctype), space=source.mapentry.get_space()):
                espace = source.mapentry.get_space()

        if mm.exists('{}connsrc'.format(ctype)) or\
                mm.exists('{}connsrc'.format(ctype), space=sspace):
            csn = '{}connsrc'.format(ctype)
        else:
            csn = 'defconnsrc'
        idstr = f"{edge.source.id}:{edge.target.id}:{ctype}"
        self.cSrc = ConnSrc('{}->'.format(idstr),
                            csn,
                            app,
                            source,
                            self,
                            space=sspace)

        if mm.exists('{}connedge'.format(ctype)) or\
                mm.exists('{}connedge'.format(ctype), space=espace):
            cen = '{}connedge'.format(ctype)
        else:
            cen = 'defconnedge'
        self.cEdge = ConnEdge(idstr,
                              cen,
                              app,
                              self,
                              space=espace)

        if mm.exists('{}conntgt'.format(ctype)) or\
                mm.exists('{}conntgt'.format(ctype), space=tspace):
            ctn = '{}conntgt'.format(ctype)
        else:
            ctn = 'defconntgt'
        self.cTgt = ConnTgt('->{}'.format(idstr),
                            ctn,
                            app,
                            target,
                            self,
                            space=tspace)

        @event.callback
        def callback(*args, **kwargs):
            warn("==== Calling Connection Selected Callback")
            self.menu.hide()
            em.model_toggle_selected(self)
        self.menu.add_item(MenuItem('menudefselect', callback))

        @event.callback
        def edit_properties(*args, **kwargs):
            warn("===-- Calling Connection Edit Properties Callback")
            self.menu.hide()
            em.show_property_overlay(self)

        if hasattr(self.target, 'mapentry') and\
                isinstance(self.target.mapentry, Receiver) and\
                ctype == 'pass':
            self.menu.add_item(MenuItem('menushowproperties', edit_properties))

        source.link(self.cSrc)
        target.link(self.cTgt)
        self.calcPos()

    @funlog
    def getPos(self):
        """Return the center point of the connection."""
        return (self.cSrc.get_nodep().getPos() + self.cTgt.get_nodep().getPos()) / 2.0

    @funlog
    def reparentTo(self, node):  # noqa: N802
        """Call the member reparentTo methods."""
        self.cSrc.reparentTo(node)
        self.cEdge.reparentTo(node)
        self.cTgt.reparentTo(node)
        self.menu.update_container()

    @funlog
    def calcPos(self):  # noqa: N802
        """
        Calculate member offset positions.

        Updates the member model's sizes, positions,
        orientations accordingly.
        """
        src_size = 0
        s_end_size = 0
        tgt_size = 0
        t_end_size = 0

        wrk_src = self.source
        wrk_tgt = self.target
        if self.source_holder:
            wrk_src = self.source_holder
        if self.target_holder:
            wrk_tgt = self.target_holder

        s_pos = wrk_src.getWorldPos()
        t_pos = wrk_tgt.getWorldPos()

        norm_conn_vec = s_pos - t_pos
        norm_conn_vec.normalize()

        n = wrk_src.getTightBounds()
        if n is not None:
            mn, mx = n[0], n[1]
            src_size = (mx - mn).length() / 6.0

        n = wrk_tgt.getTightBounds()
        if n is not None:
            mn, mx = n[0], n[1]
            tgt_size = (mx - mn).length() / 6.0

        n = self.cSrc.getTightBounds()
        if n is not None:
            mn, mx = n[0], n[1]
            s_end_size = (mx - mn).length() / 6.0

        n = self.cTgt.getTightBounds()
        if n is not None:
            mn, mx = n[0], n[1]
            t_end_size = (mx - mn).length() / 6.0

        s_offset = src_size + s_end_size
        t_offset = tgt_size + t_end_size

        sp = (norm_conn_vec * s_offset * -1.0) + s_pos
        tp = (norm_conn_vec * t_offset) + t_pos

        if not self.cSrc.get_nodep().is_empty():
            self.cSrc.get_nodep().setPos(sp)
        else:
            warn(f"{self.cSrc} NodePointer is empty")
        if not self.cTgt.get_nodep().is_empty():
            self.cTgt.get_nodep().setPos(tp)
        else:
            warn(f"{self.cTgt} NodePointer is empty")
        self.cEdge.updateConnection(sp, tp)
        self.reparentTo(wrk_src.getParent())

    @funlog
    def handle_clicked(self, origin, app):
        """Show the menu on click."""
        em.post_pending('model_toggle_selected', self)
        em.edge_show_menu(self, origin, app)

    @funlog
    def show_menu(self, origin, app):
        """Show the menu."""
        self.menu.update_container()
        # Ignore the passed event origin.... it renders
        # with some sort of transformation I don't feel like digging into now
        self.menu.show(self.getPos(), app)

    @funlog
    def hide_menu(self):
        """Hide the conneciton menu."""
        self.menu.hide()

    @funlog
    def hide(self):
        """Remove the visible artifacts from the screen."""
        self.cSrc.remove()
        self.cTgt.remove()
        self.cEdge.remove()

    @funlog
    def remove(self):
        """Remove the connection visual artifacts and mapentry if possible."""
        # if this edge references a delegate
        # we'll need to clean that up along with removing
        # the visible artifacts
        for itm in (self.source, self.target):
            if hasattr(itm, "delegates"):
                for direction in itm.delegates:
                    for rule in itm.delegates[direction]:
                        for vis in itm.delegates[direction][rule]:
                            vis.rem_delegated_conn(self)
        if hasattr(self.source, 'delegate'):
            self.source.rem_delegated_conn(self)
        if hasattr(self.target, 'delegate'):
            self.target.rem_delegated_conn(self)
        if hasattr(self, 'mapentry'):
            vm.remove_edge(self)
        else:
            self.hide()

    @funlog
    def renderSelected(self):
        self.cSrc.setSelected()
        self.cTgt.setSelected()
        self.cEdge.setSelected()

    @funlog
    def setSelected(self):  # noqa: N802
        """Set this item as selected and affect members accordingly."""
        self.selected = True
        self.renderSelected()

    @funlog
    def renderUnselected(self):
        self.cSrc.setUnselected()
        self.cTgt.setUnselected()
        self.cEdge.setUnselected()

    @funlog
    def setUnselected(self):  # noqa: N802
        """Set this item as unselected and affect members accordingly."""
        self.selected = False
        self.renderUnselected()

    @funlog
    def toggle_selected(self):
        """Toggle the selected state of this connection."""
        if self._toggle_selected():
            em.mouse_select_object(self)
        else:
            em.mouse_deselect_object(self)

    @funlog
    def _toggle_selected(self):
        rv = True
        if self.selected:
            self.setUnselected()
            rv = False
        else:
            self.setSelected()
        return rv

    @funlog
    def getuid(self):
        """
        Assemble the UID from the underlying source and return it.
        """
        return self.cEdge.uid


    @funlog
    def camPlaneMove(self, cam, mpos):  # noqa: N802
        """
        No-op placeholder.

        Permits connections to be operated upon
        as if they were models, without actually doing anything...
        """
        pass

    @funlog
    def move_source_to(self, visnode):
        self.source = visnode
        self.source_holder = None
        self.calcPos()

    @funlog
    def move_target_to(self, visnode):
        self.target = visnode
        self.target_holder = None
        self.calcPos()

    """
    These visonly implementations are insufficient
    in their current form as any move of a connected
    object will reset the pointing.
    """
    @funlog
    def move_target_vis_only(self, visnode):
        self.target_holder = visnode
        self.calcPos()

    @funlog
    def move_source_vis_only(self, visnode):
        self.source_holder = visnode
        self.calcPos()

    @funlog
    def pointed_at(self):
        if self.target_holder is not None:
            return self.target_holder
        else:
            return self.target

    # Appears to be DEPRECATED
    @funlog
    def resolve_delegation(self):
        if not hasattr(self.mapentry, 'delegate'):
            return None
        delegate = self.mapentry.delegate
        if 'uid' in delegate and\
                'rule_type' in delegate and\
                'direction' in delegate:
            if delegate['direction'] == 'as_src':
                end = self.source
            else:
                end = self.target

            end.set_delegated_conn(end.get_delegate(delegate['uid'],
                                                    delegate['direction'],
                                                    delegate['rule_type']),
                                   self)

    @funlog
    def demap(self):
        """
        Remove the connection between the map entity
        and the visible one, most likely for the purposes
        of removing nodes and edges from a graph
        """
        if hasattr(self, "mapentry"):
            if hasattr(self.mapentry, "visualization"):
                del self.mapentry.__dict__["visualization"]
            delattr(self, "mapentry")
