"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The central part of a rendered edge, between the source and target.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.shell.model import Model
from nord.sigurd.utils.log import warn


class ConnEdge(Model):
    """The class which implements the edge."""

    @funlog
    def __init__(self, uid, model, app, connection, space="nord"):
        """Call the super constructor and attaches the model bones."""
        super().__init__(uid, model, app, tag='clickable', space=space)
        self.connection = connection

        self.tgtEnd = self.model.exposeJoint(None, "modelRoot", "tgtBone")
        self.tgtCJ = self.model.controlJoint(self.tgtEnd, "modelRoot", "tgtBone")

        self.srcEnd = self.model.exposeJoint(None, "modelRoot", "srcBone")
        self.srcCJ = self.model.controlJoint(self.srcEnd, "modelRoot", "srcBone")
        # self.data.listJoints()

        ll, ur = self.model.getTightBounds()

        # x1, y1, z1 = ll[0], ll[1], ll[2]
        # x2, y2, z2 = ur[0], ur[1], ur[2]

        self.size = abs((ur - ll).length())
        # max([abs(x1 - x2), abs(y1 - y2), abs(z1 - z2)])

        self.model.setP(self.data, -90)

    @funlog
    def setSrcPos(self, pos):
        """Set the position of the source bone."""
        if not self.data.is_empty():
            self.data.setPos(pos)
        else:
            warn(f"{self} NodePointer is empty")

    @funlog
    def setTgtPos(self, pos):
        """Set the position of the target bone."""
        if not self.data.is_empty():
            self.data.look_at(pos)
        else:
            warn(f"{self} NodePointer is empty")

    @funlog
    def updateConnection(self, srcPos, tgtPos):
        """Update the positions of this edge."""

        vec = tgtPos - srcPos
        self.setSrcPos(srcPos)

        if not self.model.is_empty():
            self.model.setSz(vec.length() / self.size * 0.8)
        else:
            warn(f"{self} NodePointer is empty")

        vec = srcPos - tgtPos
        self.setTgtPos(tgtPos)

    @funlog
    def handle_clicked(self, origin, app):
        """Pass click to the owning connection of this edge."""
        self.connection.handle_clicked(origin, app)

    @funlog
    def update_tooltip(self, tooltip):
        if hasattr(self.connection, 'mapentry'):
            edge = self.connection.mapentry
            src = edge.get_source().__class__.__name__
            tgt = edge.get_target().__class__.__name__

            if edge.source_has_executed():
                src = f"{src}*"
            if edge.target_has_executed():
                src = f"{tgt}*"
            text = f"{src}\n-- {edge.rule} -->\n{tgt}"
            if edge.has_cargo():
                text = f"{text}\npayload: {str(edge.get_cargo())}"

        else:
            text = self.connection.type
            if hasattr(self.connection.source, 'mapentry'):
                src = self.connection.source.mapentry.__class__.__name__
            else:
                src = self.connection.source.__class__.__name__

            if hasattr(self.connection.target, 'mapentry'):
                tgt = self.connection.target.mapentry.__class__.__name__
            else:
                tgt = self.connection.target.__class__.__name__
            text = f"{src} -->\n{text}\n--> {tgt}"
        tooltip.set_text(text)
