"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Base Menu class from which all other menu'self are derived.

This class provides the basic menu functionality and
visualization capability for all other menu classes.
"""

from nord.sigurd.utils.log import warn, note, verb
from nord.sigurd.utils.verboser import funlog, postcallfunlog
from nord.sigurd.config.globalVars import PLACEDIST
from panda3d.core import CollisionTraverser,\
    CollisionHandlerQueue,\
    CollisionNode,\
    CollisionRay,\
    GeomNode
import math
import nord.sigurd.managers.modelmanager as mm
import nord.sigurd.utils.namegen as ng
import time
import nord.sigurd.managers.event as event
import nord.sigurd.managers.containermanager as cm
import nord.nord.utils.SpaceManager as use
import nord.sigurd.shell.vismap as vm
from nord.sigurd.shell.menuItem import MenuItem


class Menu(object):
    """The base menu class."""

    @postcallfunlog
    def __init__(self, space="nord"):
        """Initialize the necessities for the menu to function."""
        self.menuName = "Menu: {}".format(ng.get())
        self.items = []
        self.callBeforeHide = []
        self.submenus = []
        self.itemsByName = {}
        self.maxItemSize = None
        self.data = cm.cwc().attachNewNode(self.menuName)

        self.back = mm.load_model("menuback", space)
        self.back.reparentTo(self.data)
        self.back.setScale(1.5, 0.1, 1.5)
        self.data.setPos(1000000.0, -10000000.0, 90000000.0)
        self.data.hide()

        self.pointerTraverser = CollisionTraverser("pointer traverser")
        self.pickerNode = CollisionNode('mouseRay')
        self.pointerCollisionHandler = CollisionHandlerQueue()
        picker_np = camera.attachNewNode(self.pickerNode)  # noqa: F821
        self.pickerNode.setFromCollideMask(GeomNode.getDefaultCollideMask())
        self.pickerRay = CollisionRay()
        self.pickerNode.addSolid(self.pickerRay)
        self.pointerTraverser.addCollider(picker_np, self.pointerCollisionHandler)

        self.lasthidden = time.time()
        self.visible = False
        self.app = None

    def __repr__(self):
        """Provide summary information to things that try to stringify a Menu."""
        return "[{}](items: {}, subs: {}, hidecbks: {})"\
               .format(self.menuName, len(self.items), len(self.submenus), len(self.callBeforeHide))

    @funlog
    def add_prehide_cbk(self, callback, arg):
        """
        Add a callback to this menu to be called before hiding.

        This is somewhat contrary to the use of StateManager
        and the EventManagement Abstractions. This really
        should only be used by constructs within the "menu"
        subsystem.
        """
        self.callBeforeHide.append((callback, arg))

    @funlog
    def add_submenu(self, menu):
        """Add a submenu ti this menu."""
        self.submenus.append(menu)

    @funlog
    def update_container(self):
        """
        Utility method to set correct renderer.

        Because menu's are, from a panda3d perspective, children
        of the topmost NodePath for a given view, this method
        needs to be called when the menu is expected to be
        visible in a new view.
        """
        self.data.reparentTo(cm.cwc())
        for m in self.submenus:
            m.update_container()

    @funlog
    def getPos(self):  # noqa: N802  # This is a panda3d correlary
        """Return the underlying Panda3d visualizations position."""
        return self.data.getPos()

    @funlog
    def show(self, origin, app, apply_dist=True):
        """Wrapper method to handle some defaults when showing this menu."""
        if apply_dist:
            origin = origin * PLACEDIST
            cam = cm.cwc()
            origin = cam.get_relative_point(camera, origin)  # noqa: F821
        self.show_raw(origin, app)

    @funlog
    def show_raw(self, origin, app):
        """The main "show" method for any menu."""
        self.app = app
        t = time.time() - self.lasthidden
        if t > 0.00001:
            self.visible = True

            nx = origin.getX()
            ny = origin.getY()
            nz = origin.getZ()

            self.data.setPos(nx, ny, nz)
            self.data.setHpr(camera.getHpr())  # noqa: F821
            self.data.show()
            verb("Showing Menu '{}'".format(self.menuName))
        else:
            note("Menu hidden too recently: {}self ago".format(t))

    @funlog
    @event.callback
    def hide(self, *args, **kwargs):
        """Base event callback to hide this menu."""
        if self.visible:
            for func, arg in self.callBeforeHide:
                func(arg)
            warn("Hiding Menu '{}'".format(self.menuName))
            self.visible = False
            self.lasthidden = time.time()
            self.data.hide()
            self.data.setPos(1000000.0, -10000000.0, 90000000.0)

    @funlog
    def add_item(self, menuitem, hide_on_click=None):
        """Add a visible item to this menu."""
        self.itemsByName[menuitem.name] = menuitem
        self.itemsByName[menuitem.modelname] = menuitem
        self.items.append(menuitem)
        menuitem.reparentTo(self.data)
        self.reposition_items()
        mn, mx = menuitem.getTightBounds()
        size = mx - mn
        menuitem.menu = self

        if self.maxItemSize is None or self.maxItemSize < size:
            self.maxItemSize = size

    @funlog
    def rem_item(self, menuitem):
        """Remove an item from this menu and redraw."""
        self.items.remove(menuitem)
        menuitem.remove()
        del self.itemsByName[menuitem.name]
        del self.itemsByName[menuitem.modelname]
        self.reposition_items()

    @funlog
    def has_item(self, itemname):
        """Return whether or not there is an item with the passed name."""
        return itemname in self.itemsByName

    @funlog
    def get_items(self):
        """Get the list of items."""
        return self.items

    @funlog
    def backup_items(self):
        """Save the current set of items for later restoration."""
        return (self.itemsByName, self.items)

    @funlog
    def restore_items(self, bkup):
        """Restore this menu's items from provided backup."""
        for itm in bkup[1]:
            self.add_item(itm)
        # self.itemsByName = bkup[0]
        # self.items = bkup[1]

    @funlog
    def set_items(self, new_items):
        """Replace the current set of items with the one provided."""
        self.items = new_items
        self.reposition_items()

    @funlog
    def clear_items(self):
        """Set the item pointers to empty objects."""
        for itm in self.items:
            itm.detach()
        self.itemsByName = {}
        self.items = []

    @funlog
    def get_nodep(self):
        return self.data

    @funlog
    def reposition_items(self):
        """
        Position all the items about the position of the menu.

        Adjust the scale of the items to accomodate more of them.
        This probably needs some research and smarts around
        optimal sizes and methods to automatically create
        sub-menu's after a certain number of items....

        Also, the some of the hardcoded numbers should be
        moved to globalVars and then config...
        """
        mn, mx = self.back.getTightBounds()
        size = mx - mn

        radius = size.length() / 4.0

        item_space = radius / (len(self.items) * 1.3)

        angle_sep = 360.0 / len(self.items)
        angle = 0.0
        for itm in self.items:
            nx = radius * math.cos(math.radians(angle))
            nz = radius * math.sin(math.radians(angle))

            isize = itm.get_nodep().getTightBounds()
            note("Re-scaling menu item from {} size:{} to {}".format(itm, isize, item_space))
            itm.get_nodep().setScale(item_space)

            angle += angle_sep
            itm.get_nodep().setPos(nx, -1.5, nz)

    @funlog
    def populate_nodes_menu(self, menu, space="nord"):
        """Fill menu with items for each Node type."""
        use.space(space)
        nodes = use.module(f'nord.{space}.nodes')
        contents = use.get_mod_subs(nodes)
        for itm in contents:
            if not itm.startswith('_'):
                @funlog
                def get_cbk(itm):
                    @event.callback
                    @funlog
                    def dyn_node_add(*args, **kwargs):
                        """
                        Instead of creating the visible item,
                        the action should create the map item (If there is one?)
                        Then the map would do it's job from there.

                        There probably is a usecase for rendering just the visible stuff
                        if there are visibles that do not have a node or edge....
                        """
                        # send the name of the node, its position, and the space, if available
                        vm.create_node(itm, args[0], space=space)
                    return dyn_node_add

                menu.add_item(MenuItem(itm.lower(),
                                       get_cbk(itm),
                                       space=space,
                                       title=f"Add {space}.{itm.lower()}"),
                              hide_on_click=True)

    @funlog
    def load_from_file(self, fname, space):
        """Read file and populate menu according to its instructions."""
        ospace = use.current_space()
        use.space(space)
        f = use.file(fname)
        for line in f:
            if line.strip().startswith(';'):
                continue

            # File line format
            # visible name:modelname:hide on click:mouse over text
            line = line.split(';')[0]  # ignore any comments
            visname, modelname, hoc, text = line.split(':')

            if hoc.lower() in ['1', 't', 'true', 'yes', 'y', 'on', 'hide']:
                hoc = True
            else:
                hoc = False

            module = use.module(f"nord.{space}.visibles.{visname.split('.')[0]}")
            visfunc = getattr(module, visname.split('.')[1])
            self.add_item(MenuItem(modelname,
                                   visfunc,
                                   space=space,
                                   title=text
                                   ),
                          hide_on_click=hoc)
        use.space(ospace)
