"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Base class for items to be included in menus.

Provides the rendering and interaction methods.
"""


import nord.sigurd.utils.namegen as namegen
from nord.sigurd.utils.log import err, verb
from nord.sigurd.utils.verboser import funlog, postcallfunlog
import nord.sigurd.managers.modelmanager as mm
import nord.sigurd.managers.objectmanager as objects
import nord.sigurd.managers.event as event
from nord.sigurd.shell.tooltip import Tooltip
# from panda3d.bullet import BulletSphereShape, BulletRigidBodyNode
import nord.sigurd.managers.containermanager as cm


class MenuItem(object):
    """The menu item class."""

    @postcallfunlog
    def __init__(self, model_name, callback, hides_menu=None, space="nord", title=None):
        """Prepare the visualization and other member variables."""
        self.hides_menu = hides_menu
        self.name = namegen.get()
        self.callback = callback
        self.menu = None
        self.modelname = model_name
        if title is None:
            self.title = model_name
        else:
            self.title = title
        self.uid = f"MI:{self.name}-{namegen.id()}"
        self.model = mm.load_model(model_name, space)

        self.data = cm.cwc().attachNewNode(self.uid)

        self.model.reparentTo(self.data)
        self.data.setTag('clickable', self.name)
        self.data.setTag('hoverable', self.name)

        self.tooltip = Tooltip()
        self.tooltip.add_updater(self.update_tooltip)
        # self.tooltip.text_np.reparentTo(cm.c2d())

        objects.store(self.name, self)

    def __repr__(self):
        """Produce a summary string for menu items."""
        return "<<MenuItem:['{}' i({}){}]>>"\
               .format(self.name, self.modelname, "" if not self.hides_menu else "-hides menu")

    @funlog
    def getTightBounds(self):  # noqa: N802
        """Wrap same named method around the panda3d visualization member variable."""
        return self.data.getTightBounds()

    @funlog
    def activate(self, origin, app):
        """Presumably never used, will call the callback associated with this item."""
        err("Clicked on menu item: {} -- SHOULD BE UNREACHABLE".format(self.name))
        self.callback(origin, app)

    @funlog
    def reparentTo(self, node):  # noqa: N802
        """Expose the panda3d visua/izations method of the same name."""
        self.tooltip.text_np.reparentTo(cm.cwc())
        self.data.reparentTo(node)

    @funlog
    def detach(self):
        """Detach the visible Node from it's parent, so as to hide it."""
        self.data.detachNode()

    @funlog
    def remove(self):
        """Disappear this item and delete from panda3d."""
        self.data.detachNode()
        if hasattr(self.data, 'cleanup'):
            self.data.cleanup()
        self.data.removeNode()

    @funlog
    def get_menu(self):
        """Return the menu to which this item belongs."""
        return self.menu

    @funlog
    def set_hide_menu(self, hideme):
        """Set whether or not this item hides the menu when activated."""
        self.hides_menu = hideme

    @funlog
    def handle_clicked(self, origin, app):
        """Trigger the callback for this item."""
        verb("Clicked on menu item: {}".format(self.name))

        self.callback(origin, app)
        if self.hides_menu is not None:
            self.hides_menu.hide(event.EventInstance(self.__class__.__name__, self.hides_menu))

    @funlog
    def getPos(self):
        """Return the data position."""
        return self.data.getPos()

    @funlog
    def get_nodep(self):
        return self.data

    @funlog
    def show_tooltip(self):
        """Show this objects tooltip if it has one."""
        if self.tooltip is not None:
            self.tooltip.show('above', render.getRelativePoint(self.data, self.getPos()))

    @funlog
    def hide_tooltip(self):
        """Hide this objects tooltip if it has one."""
        if self.tooltip is not None:
            self.tooltip.hide()

    @funlog
    def update_tooltip(self, tooltip):
        """Update the contents of this model's tooltip."""
        tooltip.set_text(f"""{self.title}""")

    @funlog
    def set_title(self, title):
        """Change the title of this menu item."""
        self.title = title
