"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Ancestor object for most visual artifacts.

The model is the basic rendered entity for any visual artifact.
It provides the needed methods to interact with the event system.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.utils.log import err, verb, debug, note, warn
from panda3d.core import Point3, Plane
# from panda3d.bullet import BulletSphereShape, BulletRigidBodyNode
import nord.sigurd.shell.application as app
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.modelmanager as mm
import nord.sigurd.managers.event as event
import nord.sigurd.managers.containermanager as cm
from nord.sigurd.shell.menuItem import MenuItem
from nord.sigurd.shell.submenu import SubMenu
from nord.sigurd.managers.callbackmanager import CallbackManager
import nord.nord.utils.SpaceManager as use
from nord.sigurd.shell.tooltip import Tooltip
import nord.sigurd.shell.worldview as wv
from nord.sigurd.config.globalVars import DEFAULTDIMALPHA


@event.callback
def model_clicked(model, origin):
    """
    Module level function used by event system.

    Its purpose is to trigger a specific model'self handle_clicked
    callback.
    """
    model.handle_clicked(origin, app)


def toggle_model_selected(ei):
    """
    Module level function used by event system.

    Its purpose is to toggle a specific model'self selection
    state.
    """
    model = ei.args[0]
    model.toggle_selected()


def show_model_menu(ei):
    """
    Module level function used by event system.

    Its purpose is to show a specific model'self menu.
    """
    model = ei.args[0]
    model.show_menu()


def create_connection_menu_for(self, parent_menu):
    """
    Interrogate the language for possible edges for this node.

    Create a menu of all these edges.

    Based upon the type of the child class (which should call this
    function in its constructor) interrogate the nord.rules submodule
    looking for all the rules whose names start with the classname of
    this instance (meaning there is a rule available that can come
    from this type of Node)
    """
    menu = SubMenu('menudefconnect', parent_menu)

    use.space(self.space)
    rules = use.module(f'nord.{self.space}.rules')
    for rname in (
            x for x in use.get_mod_subs(rules) +
            ['Flow', 'flow'] if x.startswith(self.__class__.__name__) or
            x in ['Flow', 'flow']):
        verb("create_connection_menu evaluating: {} for {}"
             .format(rname, self.__class__.__name__))
        add = False
        if rname.startswith(self.__class__.__name__):
            ctype = rname.replace(self.__class__.__name__, '').lower()
            add = True
        elif rname == 'Flow':
            ctype = 'flow'
            add = True

        if add:
            def gen_callback(rn, ct):

                @event.callback
                def callback(*args, **kwargs):
                    warn("  ---> Calling {} Start Connection Callback"
                         .format(rn))
                    em.model_start_connection(self, ct)

                return callback
            menu.add_item(MenuItem('connmenu{}'.format(rname.lower()),
                                   gen_callback(rname, ctype),
                                   space=self.space),
                          hide_on_click=True)

    return menu


def cam_plane_move(item, cam, mpos):
    """
    Move this item on the plane of the camera.

    Care of:

    https://discourse.panda3d.org/t/super-fast-mouse-ray-collisions-with-ground-plane/5022

    I finally got this function to work...
    """
    np = Point3()
    fp = Point3()
    new_pt = Point3()
    base.camLens.extrude(mpos, np, fp)  # noqa: F821
    np = cm.cwc().getRelativePoint(cam, np)
    fp = cm.cwc().getRelativePoint(cam, fp)

    plane = Plane(np - fp, item.getPos())
    plane.intersectsLine(new_pt, np, fp)
    if hasattr(item, "mapentry"):
        """
        update the position in the map, and let the
        map_change system trigger the vismap move
        """
        item.mapentry.position = [new_pt.getX(), new_pt.getY(), new_pt.getZ()]
    else:
        item.setPos(new_pt.getX(), new_pt.getY(), new_pt.getZ())


class Model(CallbackManager):
    """
    Eponymous class for this module.

    This class should be the ancestor to most visual
    subclasses like Nodes, menuItems, etc.

    This class provides the needed wrapper functions
    to aid in event detection and object management.
    """

    @funlog
    def __init__(self, uid, name, app, tag='clickable', space="nord", hoverable=True, filterable=True):
        """The constructor."""
        super().__init__(space=space, app=app)
        self.uid = uid
        self.name = name
        self.app = app

        self.tag = tag
        self.hoverable = hoverable

        self.model = mm.load_model(name, space)
        self.space = space
        self.data = cm.cwc().attachNewNode(uid)

        self.model.reparentTo(self.data)
        self.data.reparentTo(cm.cwc())
        self.data.setScale(1.0, 1.0, 1.0)
        self.data.setTag(tag, uid)
        if hoverable:
            self.data.setTag('hoverable', uid)

        self.tooltip = Tooltip()
        self.tooltip.add_updater(self.update_tooltip)

        self.independently_movable = True

        import nord.sigurd.managers.objectmanager as objects
        objects.store(self.uid, self)
        self.oldScale = self.data.getColorScale()
        self.selected = False
        self.menu = None
        self.priorAlpha = self.data.getColorScale()[3]
        self.visible = False
        self.filterable = filterable
        if filterable:
            import nord.sigurd.managers.filtermanager as fm
            fm.add_node(self)

    @funlog
    def set_new_model(self, name, space="nord"):
        self.model.cleanup()
        self.model = mm.load_model(name, space)
        self.model.reparentTo(self.data)

    @funlog
    def setPos(self, x, y, z):  # noqa: N802
        """Set the Panda3d NodePath's parent relative position."""
        debug("Setting position of {}: '{}' to ({:.3f}, {:.3f}, {:.3f})"
              .format(self.name, self.uid, x, y, z))
        self.data.setPos(x, y, z)
        for k in self.links:
            link = self.links[k]
            link.setPos(x, y, z)

    @funlog
    def getWorldPos(self):  # noqa: N802
        """Set the Panda3d NodePath's absolute position."""
        node = self.data
        rv = wv.getWorldPos(node)

        dp = node.getPos()
        if dp[0] == 0.0 and dp[1] == 0.0 and dp[2] == 0.0 and\
                hasattr(self, 'position'):
            note("Updating position of {} to {}".format(self, self.position))
            self.data.setPos(self.position[0],
                             self.position[1],
                             self.position[2])
        return rv

    @funlog
    def setScale(self, scale):  # noqa: N802
        """Set the Panda3d NodePath's scale."""
        self.data.setScale(scale)

    @funlog
    def getPos(self):  # noqa: N802
        """Get the Panda3d NodePath's position."""
        if self.data.is_empty():  # This check seems needed if a model is deleted, but still
            # attempted to be referenced. Hopefully the -0.0 etc position is invalid enough...
            return (-9999999.0, -0.0, -0.0)
        dp = self.data.getPos()
        if dp[0] == 0.0 and dp[1] == 0.0 and dp[2] == 0.0 and\
                hasattr(self, 'position'):
            self.data.setPos(self.position[0],
                             self.position[1],
                             self.position[2])
        return self.data.getPos()

    @funlog
    def getX(self):  # noqa: N802
        """Get the Panda3d NodePath's position X."""
        dp = self.data.getPos()
        if dp[0] == 0.0 and dp[1] == 0.0 and dp[2] == 0.0 and\
                hasattr(self, 'position'):
            self.data.setPos(self.position[0],
                             self.position[1],
                             self.position[2])
        return self.data.getX()

    @funlog
    def getY(self):  # noqa: N802
        """Get the Panda3d NodePath's position Y."""
        dp = self.data.getPos()
        if dp[0] == 0.0 and dp[1] == 0.0 and dp[2] == 0.0 and\
                hasattr(self, 'position'):
            self.data.setPos(self.position[0],
                             self.position[1],
                             self.position[2])
        return self.data.getY()

    @funlog
    def getZ(self):  # noqa: N802
        """Get the Panda3d NodePath's position Z."""
        dp = self.data.getPos()
        if dp[0] == 0.0 and dp[1] == 0.0 and dp[2] == 0.0 and\
                hasattr(self, 'position'):
            self.data.setPos(self.position[0],
                             self.position[1],
                             self.position[2])
        return self.data.getZ()

    @funlog
    def getWidth(self):  # noqa: N802
        """Get the Panda3d NodePath's width."""
        c1, c2 = self.getTightBounds()
        return abs(c1.x - c2.x)

    @funlog
    def getHeight(self):  # noqa: N802
        """Get the Panda3d NodePath's height."""
        c1, c2 = self.getTightBounds()
        return abs(c1.z - c2.z)

    @funlog
    def getDepth(self):  # noqa: N802
        """Get the Panda3d NodePath's depth."""
        c1, c2 = self.getTightBounds()
        return abs(c1.y - c2.y)

    @funlog
    def get_nodep(self):
        return self.data

    @funlog
    def setIndependentMotion(self, flag):  # noqa: N802
        """Set the flag indicating if this model can be directly moved."""
        self.independently_movable = flag

    @funlog
    def getTightBounds(self):  # noqa: N802
        """Return the Panda3d nodepath's boundary tuple."""
        if not self.data.has_parent():
            return None
        return self.data.getTightBounds()

    @funlog
    def handle_clicked(self, *args, **kwargs):
        """Placeholder method to be overridden by descendents."""
        verb("{} '{}' clicked, without overriden handler"
             .format(self.name, self.uid))

    @funlog
    def toggle_selected(self):
        """
        Toggle the selection status of this element between off and on.

        Fire off the mouse_select_object event.
        """
        if self.toggleSelected():
            em.mouse_select_object(self)
        else:
            em.mouse_deselect_object(self)

    @funlog
    def show_menu(self):
        """Show this objects menu, if it has one."""
        if self.menu is not None:
            self.menu.show(self.data.getPos(), None)

    @funlog
    def show_tooltip(self):
        """Show this objects tooltip if it has one."""
        if self.tooltip is not None:
            self.tooltip.show('above', self.data.getBounds().getCenter())

    @funlog
    def hide_tooltip(self):
        """Hide this objects tooltip if it has one."""
        if self.tooltip is not None:
            self.tooltip.hide()

    @funlog
    def update_tooltip(self, tooltip):
        """Update the contents of this model's tooltip."""
        title = '--'
        if hasattr(self, 'mapentry'):
            title = self.mapentry.__class__.__name__
        tooltip.set_text("""{}\nname: {}\nid: {}\npos: {:0.4f}, {:0.4f}, {:0.4f}\ntype: {}"""
                         .format(title,
                                 self.name,
                                 self.uid,
                                 self.getX(),
                                 self.getY(),
                                 self.getZ(),
                                 self.__class__.__name__))

    @funlog
    def set_tooltip_updater(self, func):
        """
        expose the tooltip set_updater method.
        """
        self.tooltip.set_updater(func)

    @funlog
    def get_tooltip_updater(self):
        """
        expose the tooltip get_updater method.
        """
        return self.tooltip.get_updater()

    # #  @funlog
    def get_tooltip_vis_secs(self):
        """
        return the tooltip's visibility time in seconds
        """
        return self.tooltip.get_vis_secs()

    @funlog
    def set_tooltip_vis_secs(self, secs):
        """
        Set the tooltip's visibility time in seconds
        """
        self.tooltip.set_vis_secs(secs)

    @funlog
    def toggleSelected(self):  # noqa: N802
        """Toggle the selection status of this element between off and on."""
        rv = True
        if self.selected:
            self.setUnselected()
            rv = False
        else:
            self.setSelected()
        return rv

    @funlog
    def setSelected(self):  # noqa: N802
        """Set this object as selected."""
        self.selected = True
        self.renderSelected()

    @funlog
    def renderSelected(self):
        """
        Display this object as if it were selected
        """
        self.data.setColorScale(255, 255, 255, 128)

    @funlog
    def setUnselected(self):  # noqa: N802
        """Set this object as de-selected."""
        self.selected = False
        self.renderUnselected()

    @funlog
    def renderUnselected(self):
        """
        Display this object as if it were not selected
        """
        self.data.setColorScale(self.oldScale)

    @funlog
    def set_dimmed(self, dimTo=DEFAULTDIMALPHA):
        """Reduce the visibility of this model."""
        if self.priorAlpha is None:
            self.priorAlpha = self.data.getColorScale()[3]
        self.data.setAlphaScale(dimTo)

    @funlog
    def set_undimmed(self):
        """Restore the visibility of this model."""
        if self.priorAlpha is not None:
            self.data.setAlphaScale(self.priorAlpha)
        self.priorAlpha = None

    @funlog
    def hide(self):
        """Remove the panda3d nodepath from the render path."""
        if self.data.has_parent():
            self.data.hide()
            self.visible = False

    @funlog
    def show(self):
        """Re-add the panda3d nodepath to the current render path."""
        self.data.show()
        self.visible = True

    @funlog
    def is_visible(self):
        """Return whether or not this model has been hidden."""
        return self.visible

    @funlog
    def remove(self):
        """If also logical, request removal from map, else, just un-render"""
        if hasattr(self, 'mapentry'):
            import nord.sigurd.shell.vismap as vm
            vm.remove_node(self)
        else:
            self.unrender()

    @funlog
    def unrender(self):
        """Remove this item from the screen and logical map."""
        self.hide()
        self.model.cleanup()
        self.data.removeNode()

        for func in self.onUnmappedFuncs:
            func()

        import nord.sigurd.managers.objectmanager as om
        om.remove(self.uid)

        if self.filterable:
            import nord.sigurd.managers.filtermanager as fm
            fm.rem_node(self)

    @funlog
    def camPlaneMove(self, cam, mpos):  # noqa: N802
        if not self.independently_movable:
            return
        if self.data.is_empty():
            err("Trying to move an empty node... this means the selectionmanager\
                 did not de-select the node upon removal")
            return

        cam_plane_move(self, cam, mpos)

    @funlog
    def create_connection_menu(self, parent_menu):
        return create_connection_menu_for(self, parent_menu)
