"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Render a 2d overlay constructed from the _properties dict of a visnode which allows editing.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import note, err
import nord.sigurd.managers.event as event
import nord.sigurd.managers.containermanager as cm
from nord.sigurd.visibles.ui.text import Text
from nord.sigurd.visibles.ui.label import Label
from nord.sigurd.visibles.ui.button import Button
from panda3d.core import NodePath


_vis_frame = None


@funlog
def _hide(vf):
    note(f"Hiding Property Editor filled with: {vf}")
    vf['parent'].removeNode()


@funlog
def _show(vf):
    note(f"Showing Property Editor filled with: {vf}")
    vf['parent'].setPos(vf['source'].getPos())
    vf['parent'].show()

@event.callback
@funlog
def show(visnode):
    """Show the property widgets for the passed visnode."""
    note("Showing property items for {}".format(visnode))
    global _vis_frame
    if _vis_frame is not None:
        err("Already showing properties, hiding them before showing new ones.")
        _hide(_vis_frame)
        _vis_frame = None

    _vis_frame = {}

    parent = cm.cwc().attachNewNode('__PropertyEditor__')
    _vis_frame['source'] = visnode
    _vis_frame['parent'] = parent
    _vis_frame['properties'] = []

    def closeme(*args, **kwargs):
        global _vis_frame
        _hide(_vis_frame)
        _vis_frame = None

    close = Button('PropEditorClose', text='\n Close \n.',
                   parent=parent,
                   command=closeme
                   )

    _vis_frame['close'] = close
    # now, show all the properties for the visnode:
    kl = list(visnode.mapentry._properties.keys())
    kl.sort()
    scale = 0.1
    z = 0.0
    for pk in kl:
        x = 0.0
        y = -2.0
        p = visnode.mapentry._properties[pk]
        l = Label(p.name,
                  (x, y, z),
                  parent
                  )

        def gen_prop_input_changed(property):

            @event.callback
            def prop_input_changed(new_text):
                if hasattr(visnode, 'handlePropertyChange'):
                    visnode.handlePropertyChange(property, new_text)
                visnode.mapentry.set_property(property.name, new_text)

            return prop_input_changed

        prop = visnode.mapentry.get_property(pk)

        if prop is None:
            prop = 'NOT SET'

        val = Text(f'_property_{p.name}', None, parent=parent, default_contents=prop, one_line=True)
        val.setPos(x+l.getWidth(),y,z)
        val.addCallback('on_change', gen_prop_input_changed(p))
        _vis_frame['properties'].append((l, val))
        z -= val.getHeight()
    close.setPos(x,y,z)
    _show(_vis_frame)


@event.callback
@funlog
def hide(*args, **kwargs):
    """Hide the property overlay for the passed visnode."""
    note("Hiding property screen")
    global _vis_frame
    if _vis_frame is not None:
        _hide(_vis_frame)
        _vis_frame = None
