"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Incorporate a menu and menuitem for submenuing.

The base class in a descendent of MenuItem such that
it can be treated like a menuItem. Also exposes many
of the member menu methods to be able to be operated
upon like a menu.
"""


from nord.sigurd.utils.log import verb
from nord.sigurd.utils.verboser import funlog, postcallfunlog
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.event as event
from nord.sigurd.shell.menuItem import MenuItem


@event.callback
@funlog
def show(subm):
    """Module level show method to facilitate statemanager interaction."""
    subm._show_menu()


# @event.callback
#  @funlog
# def hide(subm):
#     """Module level hide method to facilitate statemanager interaction."""
#     subm.hide_menu()


class SubMenu(MenuItem):
    """The SubMenu class."""

    @postcallfunlog
    def __init__(self, model_name, parent_menu, add_on_init=True, space="nord", title=None):
        """Call parent constructor and prepare menu member."""
        super().__init__(model_name, self._callback, space=space, title=title)
        # from nord.sigurd.shell.menu import Menu
        # self.submenu = Menu()
        self.submenu = None
        self.hasBackItem = False
        self.parent = parent_menu
        self.backItem = None
        self.showPos = None

        @funlog
        def fetch_pos(me):
            me.showPos = me.parent.getPos()
            if self.submenu is None:
                from nord.sigurd.shell.menu import Menu
                me.submenu = Menu()
            me.submenu.update_container()

        self.parent.add_prehide_cbk(fetch_pos, self)
        if add_on_init:
            self.parent.add_submenu(self)

    @funlog
    def load_from_file(self, fname, space):
        if self.submenu is None:
            from nord.sigurd.shell.menu import Menu
            self.submenu = Menu()
        self.submenu.load_from_file(fname, space)

    @funlog
    def _callback(self, *args, **kwargs):
        em.show_sub_menu(self)

    @funlog
    def _show_menu(self):
        """Private show_menu to be called by the statemanager BYO module.show."""
        self.parent.hide(event.EventInstance(self.parent.__class__.__name__))
        self.show_menu(self.showPos, self.parent.app)

    @funlog
    def hide_menu(self):
        """Hide the menu member."""
        self.submenu.hide(self)

    @funlog
    def show_menu(self, pos, app):
        """Exposed method to render the menu, adding back item if need be."""
        if self.backItem is None:
            verb("Initializing subMenu Back Item")

            @event.callback
            @funlog
            def backcbk(*args, **kwargs):
                pos = self.submenu.getPos()
                self.parent.show_raw(pos, self.parent.app)

            self.backItem = MenuItem('submenuback', backcbk, hides_menu=self.submenu)
            self.submenu.add_item(self.backItem)

        @funlog
        def callback():
            verb("The submenu should be hidden after this...")
            self.hide_menu()
        em.listen('hide_sub_menu', 'SubMenu', callback)

        self.submenu.show_raw(pos, app)
        self.submenu.get_nodep().ls()

    @funlog
    def add_item(self, menuitem, hide_on_click=False):
        """Add an item to the member menu."""
        if self.submenu is None:
            from nord.sigurd.shell.menu import Menu
            self.submenu = Menu()
        if hide_on_click:
            menuitem.set_hide_menu(self.submenu)
        self.submenu.add_item(menuitem)

    @funlog
    def has_item(self, itemname):
        """Return whether or not there is an item with the passed name."""
        if self.submenu is None:
            from nord.sigurd.shell.menu import Menu
            self.submenu = Menu()
        return self.submenu.has_item(itemname)

    @funlog
    def rem_item(self, item):
        """Removed the referenced item from the menu."""
        if self.submenu is None:
            return
        self.submenu.rem_item(item)

    @funlog
    def reposition_items(self):
        """Expose the member menu's reposition_items method."""
        if self.submenu is None:
            from nord.sigurd.shell.menu import Menu
            self.submenu = Menu()
        self.submenu.reposition_items()

    @funlog
    def update_container(self):
        """Expose the member menu's update_container method."""
        if self.submenu is None:
            from nord.sigurd.shell.menu import Menu
            self.submenu = Menu()
        self.submenu.update_container()
