"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Enable visualizations to display a two dimensional sheet of information.

The tooltip module provides the basic methods to control
the marshalling and display of 2d info upon the configured
triggering events.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
import nord.sigurd.managers.event as event
from nord.sigurd.utils.log import err
# from direct.gui.DirectGui import DirectFrame, DirectLabel, DirectButton, DirectEntry
from panda3d.core import TextNode
import nord.sigurd.utils.namegen as namegen
import nord.sigurd.managers.containermanager as cm
import datetime
from nord.sigurd.config.globalVars import DEFAULTTIPSHADOWPOS,\
    DEFAULTTIPSHADOWCOL,\
    DEFAULTTIPBGCOL,\
    DEFAULTTIPMARGIN,\
    DEFAULTTIPSCALE,\
    DEFAULTTIPFGCOL,\
    DEFAULTTTVISSECS

_shown_tts = {}


@event.callback
@funlog
def show(*args, **kwargs):
    """The hook to the event system to display a tooltip."""
    global _shown_tts
    err("Tooltip Show requested with: {}, {}".format(args, kwargs))
    model = args[0]
    model.show_tooltip()
    _shown_tts[model] = datetime.datetime.now()


@event.callback
@funlog
def hide(*args, **kwargs):
    """The hook to the event system to hide a tooltip."""
    global _shown_tts
    err("Tooltip hide requested with: {}, {}".format(args, kwargs))
    model = args[0]
    model.hide_tooltip()
    if model in _shown_tts:
        del _shown_tts[model]


#  @  funlog
def expire_tooltips():
    """Hide tooltips that have been onscreen longer than the interval."""
    global _shown_tts
    for model in _shown_tts.copy():
        if hasattr(model, 'get_tooltip_vis_secs'):
            secs = model.get_tooltip_vis_secs()
        else:
            secs = DEFAULTTTVISSECS
        if (datetime.datetime.now() - _shown_tts[model]).total_seconds() > secs:
            hide(model)


class Tooltip(object):
    """
    The base object for creating and rendering tooltips.

    Basic capabilities include text and image display.
    """

    @funlog
    def __init__(self):
        """The tooltip constructor."""
        self.visible = False
        self.text = ""
        self.vis_secs = DEFAULTTTVISSECS
        self.updaters = list()
        self.onsrctxt = TextNode(namegen.get())
        self.onsrctxt.setShadow(DEFAULTTIPSHADOWPOS)
        self.onsrctxt.setShadowColor(DEFAULTTIPSHADOWCOL)
        self.onsrctxt.setCardColor(DEFAULTTIPBGCOL)
        self.onsrctxt.setTextColor(DEFAULTTIPFGCOL)
        self.onsrctxt.setCardAsMargin(*DEFAULTTIPMARGIN)
        self.onsrctxt.setCardDecal(True)
        self.text_np = cm.cwc().attachNewNode(self.onsrctxt)
        self.text_np.setScale(DEFAULTTIPSCALE)
        self.text_np.setBillboardPointEye(-10, fixed_depth=True)

    @funlog
    def show(self, where, pos3d):
        """Relative to the provided 3d position, show the tooltip."""
        self.visible = True
        for upd in self.updaters:
            upd(self)

        # self.text_np.reparentTo(cm.c2d())
        self.text_np.reparentTo(cm.cwc())

        width = self.onsrctxt.getWidth()
        height = self.onsrctxt.getHeight()

        pos3d[0] += width / -2.0
        pos3d[2] += height

        self.text_np.setPos(pos3d)
        self.text_np.show()

    @funlog
    def hide(self):
        """Hide this tooltip if visible."""
        if self.visible:
            pass
        self.visible = False
        self.text_np.hide()

    @funlog
    def clear(self):
        """Clear the contents of this tooltip."""
        self.onsrctxt.setText('')

    @funlog
    def set_text(self, text):
        """Set the contents of the tooltip."""
        self.onsrctxt.setText(text)
        self.text = text

    @funlog
    def get_text(self):
        """Return the most recently set text"""
        return self.text

    @funlog
    def add_updater(self, func):
        """Add the provided function to the list of those to be called before showing."""
        self.updaters.append(func)

    @funlog
    def set_updater(self, func):
        """
        Replace any updaters with the newly provided one.
        """
        self.updaters.clear()
        self.updaters.append(func)

    @funlog
    def get_updater(self):
        """
        return the current updater (or list if there are more than one)
        """
        if len(self.updaters) == 1:
            return self.updaters[0]
        else:
            return self.updaters

    ##  @funlog
    def get_vis_secs(self):
        """
        return the number of seconds the tooltip should be on screen
        """
        return self.vis_secs

    @funlog
    def set_vis_secs(self, n_secs):
        """
        set the number of seconds the tooltip should be on screen
        """
        self.vis_secs = n_secs