"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Primary module to manage the relationship between the map and visualizations.

This is the primary place to manage the connections of visual artifacts
and their correlated graph/map constructs.

VisMap owns the visualization of NoRD'_working_map Maps.
"""


from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import warn, err, info, notice, verbose
from nord.sigurd.config.globalVars import DEFAULTOUTFILENAME, PLACEDIST
import nord.sigurd.managers.objectmanager as om
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.vismanager as vm
from nord.sigurd.managers.connectionmanager import render_completed_conn
from nord.nord.exceptions import UseException
from nord.nord.exceptions import MapException
import nord.nord.InstanceOf as InstanceOf
import json as json
import nord.sigurd.utils.namegen as ng
from panda3d.core import Point3
import nord.nord.utils.SpaceManager as use
from nord.wrigan.visibles.runmenu import load_map, get_running_map_id
from nord.wrigan.visibles.utils import get_changes, ship_changes
from nord.nord.changetrack import merge
from nord.wrigan.visibles.utils import show_error


Map = use.module('nord.nord.Map')
Edge = use.module('nord.nord.Edge')
_working_map = Map(None)

"""
_in_bypass = True is the original behavior of the load routines. When True
load, will directly parse the _working_map. When False, the CDC implementation
is used to enqueue the changes induced by loading the map, and, then apply them
all when the map has been loaded.
"""
_in_bypass = True
_pending_changes = []

"""
NOTE: THIS IS DEPRECATED, AND POSSIBLY EVEN HARMFUL


The purpose of _mod_flags is to track
if it is needed to send the map to the
backend.

any time get_map_dict is called, the changes
are cleared, as it is assumed that a "gotten"
map is being shared with the backend.

There is an inconsistency between
node and edge change tracking.
Nodes are tracked as visibles, whereas
the edges are tracked as mapentries.
This is harmless for now, as this is really
intended to prevent overshipping the map
but, if more robust use cases are attempted
this will need to be rectified.

This is probably insufficient to
track whether or not to send the
entire map to the backend, as
vismap will not have visibility
into whether or not individual node
content has changed.
Which suggests thot we'll need to
integrate vismap into the VisNode
in some way, so that it is easy for
nodes to indicate that they have
been modified...
"""
_mod_flags = {
    "nodes_changed": [],
    "edges_changed": []
}


def node_changed(node):
    global _mod_flags
    if node not in _mod_flags["nodes_changed"]:
        _mod_flags["nodes_changed"].append(node)


def edge_changed(edge):
    global _mod_flags
    if edge not in _mod_flags["edges_changed"]:
        _mod_flags["edges_changed"].append(edge)


def clear_changes():
    global _mod_flags
    _mod_flags["nodes_changed"].clear()
    _mod_flags["edges_changed"].clear()


def has_changes():
    global _mod_flags
    return len(_mod_flags["nodes_changed"]) + len(_mod_flags["edges_changed"])


def synchronize():
    """
    ensure that the current _working_map
    matches the map on the backend
    """
    running_guid = get_running_map_id()
    local_map = get_working_map()
    if running_guid is None:
        return load_map(get_map_dict())
    local_changes = {}
    if local_map.has_changes():
        # send changes to server
        local_changes = local_map.get_changes()
        ship_changes(running_guid, local_changes)  # needs to be implemented in runmenu
    running_changes = get_changes(running_guid)
    running_changes = merge(local_changes, running_changes)  # merge needs to be its own function as running and local changes are likely a dict  # noqa: E501
    local_map.apply_changes(running_changes)
    # handle_map_changes(running_changes)

    return running_guid


def init():
    """Minimal constructor function to allow for future loading."""
    global _working_map, _in_bypass
    _working_map.mapfile = None
    _in_bypass = True


def unload():
    """
    Clear the current Map and prepare for initialization.

    This was created to allow Map's to be opened and closed.
    """
    global _working_map
    for n in [_working_map.nodes[node].visualization for node in
              list(_working_map.nodes.keys())
              if hasattr(_working_map.nodes[node], 'visualization')]:
        removeNode(n)
        om.remove(n.uid)
    _working_map.unload()
    clear_changes()
    _working_map = Map(None)
    _working_map.add_change_handler(handle_map_changes)


# FUNLOG does not work on load as it tries to print the Map
# (not quite sure why ATT, but the Map isn't populated yet...)
@funlog
def load(mapname, app, bypass_cdc=True):
    """
    Load the map and parse it for visual artifacts.

    First, load the runtime map object with the dictionary, then,
    interrogate parent nodes, edges, and rules, which should
    include any visualization fields from the input JSON
    """
    global _working_map, _in_bypass, _pending_changes
    _working_map.mapfile = mapname
    _working_map.app = app
    _in_bypass = bypass_cdc

    try:
        m = json.load(open(mapname))
    except FileNotFoundError:
        warn("File: {} not found, cannot load map".format(mapname))
        """
        Prepare an empty map to "load" such that all needed initialization activities
        take place. When we just returned at this point new, empty, maps could not have any nodes
        added to them.
        """
        _working_map.name = ng.get()
        m = _working_map.to_dict()
    _working_map.load(m, verbose=True, bypass_cdc=bypass_cdc)

    import nord.sigurd.shell.mainmenu as mm
    for extn in _working_map.extensions:
        app.load_space_config(extn)
        mm.add_space_menu(extn, mm)

    x = len(_working_map.nodes) / -4.0
    z = x
    cnt = 0

    mods_to_run_onmapped = list()

    if bypass_cdc:
        for n in _working_map.nodes:
            """
                for each node, extract the
                position and scale information,
                or, generate it on best guess estimation...
                and then render them, storing the
                visual objects in the node instance
                as well as the node instance in the
                visual object

                Nodes (vertices) are rendered by the modelmap. The node type
                is the name of the model looked up in the modelmap

                Edges are visualized through the visualization manager
                their rule corresponds to the visualization functions
                provided by the edgemap managed by the visualization manager

            """

            if not hasattr(_working_map.nodes[n], 'position'):
                info("No position specified for node:{}".format(_working_map.nodes[n]))
                origin = Point3(x, 0.0, z)
            else:
                p = _working_map.nodes[n].position
                origin = Point3(p[0], p[1], p[2])

            x = (cnt % len(_working_map.nodes) / 2.0 - len(_working_map.nodes) / 4.0) * 5.0

            cnt += 1
            if cnt % len(_working_map.nodes) / 4.0 == 0:
                z += 1

            modname = "{}".format(_working_map.nodes[n].type)

            if hasattr(_working_map.nodes[n], "space") and _working_map.nodes[n].space != "nord":
                space = _working_map.nodes[n].space
                pfx = f"nord.{space}."
            else:
                space = "nord"
                pfx = "nord.sigurd."
            use.space(space)

            try:
                module = use.module(f"{pfx}visibles.{modname}")
                modconstructor = getattr(module, _working_map.nodes[n].type.capitalize())
                mod = modconstructor(_working_map.nodes[n].id, _working_map.app)
            except UseException:
                from nord.sigurd.shell.visnode import VisNode
                c = type(_working_map.nodes[n].type.capitalize(), (VisNode, ), {})
                mod = c('{}'.format(_working_map.nodes[n].name),
                        '{}'.format(_working_map.nodes[n].name),
                        _working_map.app)
            _working_map.nodes[n].set_untracked_attr('visualization', mod)

            """
                Is linking the visual object to the logical (the graph'_working_map
                reference to the node)
                best done here in the visualization manager?

                Should this be managed by the object manager?

                Should there be yet another manager...?

                I think it is best in here, and then
                have the model wrapper functions call vismap.add or something
                when creating an element, or removing one...
            """
            mod.mapentry = _working_map.nodes[n]

            om.add_item(origin, _working_map.app, mod, mouseInteractive=False)

            node_changed(mod)

            for func in mod.onMappedFuncs:
                func()
            mods_to_run_onmapped.append(mod)

        # For each edge, pass it to the visualization manager function
        # defined in vismap.conf, that function will be responsible to
        # operate on the source and target Model'_working_map stored in the nodes
        # but also constrained by precedence.conf
        for en in list(_working_map.edges):
            e = _working_map.edges[en]
            renderfunc = getattr(vm, e.rule)
            conn = renderfunc(e)
            edge_changed(e)
            if conn is not None:
                conn.mapentry = e
                e.set_untracked_attr('visualization', conn)
                # This call+function is obsolete now with UI changes driven
                # by Map change data capture (CDC)
                # conn.resolve_delegation()

        # Call all the onmapped functions now that all visualizations
        # (to include edges) have been rendered...
        for mod in mods_to_run_onmapped:
            for func in mod.onAllMappedFuncs:
                func()

    if not bypass_cdc:
        handle_map_changes(_pending_changes, apply_queue=True)
        _pending_changes.clear()

    # Once we are done loading, CDC is no longer bypassable, so, set this flag to True
    # such that handle_map_changes immediately processes changes, rather than saving them
    _in_bypass = True


@funlog
def create_node(nodename, position, space=None, containrule="contains", containername=None):
    """
    Create node in the _working_map, it's callback on onchange
    will then call this object's create_visnode method
    """
    position = position * PLACEDIST
    position = cm.cwc().get_relative_point(camera, position)  # noqa: F821

    n = {"space": space,
         "type": nodename.lower(),
         "position": [position[0], position[1], position[2]],
         "id": ng.id(),
         "name": ng.get()}
    _ = _working_map.load_node(n)
    node = get_node(n["id"])
    if not cm.is_top():
        if containername is None:
            containername = cm.get_current_name()
        container = get_node(containername)
        create_edge(container, node, containrule, None, None)
    return node


@funlog
def create_instance(parent, position, containrule="contains", containername=None):
    position = position * PLACEDIST
    position = cm.cwc().get_relative_point(camera, position)  # noqa: F821

    n = {
         "type": "instanceof",
         "position": [position[0], position[1], position[2]],
         "id": ng.id(),
         "instanceof": parent.uid,
         "name": "*"+parent.mapentry.name,
         "space": parent.space}
    _ = _working_map.load_node(n)
    node = get_node(n["id"])
    if not cm.is_top():
        if containername is None:
            containername = cm.get_current_name()
        container = get_node(containername)
        create_edge(container, node, containrule, None, None)
    return node


@funlog
def create_edge(src, tgt, ctype, src_delegate, tgt_delegate):
    for space in [src.get_space(), tgt.get_space(), "nord"]:
        try:
            use.space(space)
            econnstru = use.module(f'nord.{space}.edges.{ctype.capitalize()}')
            edge = econnstru(_working_map)
            if src_delegate and tgt_delegate:
                raise MapException(_("Delegate to delegate edge requested, but, this is not implemented yet"))  # noqa: F821, E501
            if src_delegate is not None:
                edge.delegate = src_delegate
            if tgt_delegate is not None:
                edge.delegate = tgt_delegate
            edge.set_nodes(src, tgt, ctype)

            _working_map.add_edge(edge.source, edge.target, edge.rule, edge)
            edge_changed(edge)

            return True
        except UseException:
            pass

    return False


@funlog
def handle_add_node(change):
    """A node was added to the graph/map, create associated visualization"""
    create_visible_node(change["item_id"])


@funlog
def handle_change_node(change):
    update_visible_node(change["item_id"], change)


@funlog
def handle_delete_node(change):
    delete_visible_node(change["item_id"])


@funlog
def handle_add_edge(change):
    item_id = change["item_id"]
    if type(change["item_id"]) is not tuple:
        # This is needed as the json converts the item_id to a list
        item_id = tuple(item_id)
    create_visible_edge(item_id)


@funlog
def handle_delete_edge(change):
    itemid = ":".join(change["item_id"])
    delete_visible_edge(itemid)


@funlog
def handle_map_changes(changes, apply_queue=False):
    """
    iterate over and process all map changes.
    """
    global _pending_changes, _in_bypass

    if _in_bypass or apply_queue:
        def raise_unknown_change(change):
            show_error(_("Visual Layer cannot handle map/graph change: {}").format(change))  # noqa: F821
            return
        for change in changes:
            func = {
                "add_node": handle_add_node,
                "change_node": handle_change_node,
                "delete_node": handle_delete_node,
                "add_edge": handle_add_edge,
                "delete_edge": handle_delete_edge
            }.get(change["change_type"], raise_unknown_change)
            func(change)
    else:
        _pending_changes.extend(changes)


@funlog
def create_visible_node(node_id):
    """
    Grab the needed info from the map about the node.
    """
    global _working_map
    import nord.sigurd.shell.application as app
    node = _working_map.get_node(node_id)
    old_space = use.current_space()
    try:
        # So this code probably needs to be in vismap
        # so that it can be called by the map's on_change
        # callback runner. I'm unsure whether or not to abstract
        # the map changed events through the event map and statemanager
        if not hasattr(node, "space"):
            space = "nord"
        else:
            space = node.space
        use.space(space)
        if space != "nord":
            pfx = f"nord.{space}."
        else:
            pfx = "nord.sigurd."

        if node.type == 'instanceof':
            typetouse = _working_map.get_node(node.instanceof).type
        else:
            typetouse = node.type
        module = use.module(f"{pfx}visibles.{typetouse}")
        modconstructor = getattr(module, typetouse.capitalize())
        item = modconstructor(node.id, app)

    except UseException:
        err("Incomplete visual implementation, no visual module " +
            f"named {node.type.lower()} available")
        from nord.sigurd.shell.visnode import VisNode
        item = VisNode(node.id, node.type.lower(), app)

    item.mapentry = node
    item.mapentry.set_untracked_attr("visualization", item)

    if not hasattr(node, 'position'):
        node.position = [0.0, 0.0, 0.0]

    om.add_item(Point3(*node.position), app, item)
    use.space(old_space)

    for func in item.onMappedFuncs:
        func()


@funlog
def update_visible_node(node_id, change):
    global _working_map
    # node = _working_map.get_node(node_id)
    old_space = use.current_space()
    if "attribute" in change and om.has(node_id):
        item = om.fetch(node_id)
        if change["attribute"] == "position":
            p = change["change"]
            item.setPos(p[0], p[1], p[2])
        item.handle_map_node_update(change)

        # raise MapException("NOT IMPLEMENTED")
    elif not om.has(node_id):
        warn(f"Visualization layer cannot update unknown node: {node_id}")
    else:
        notice(f"Incoming change has no attributes for node: {node_id}")
    use.space(old_space)


@funlog
def delete_visible_node(node_id):
    global _working_map
    if not om.has(node_id):
        warn(f"Visualization layer cannot delete unknown node {node_id}")
        return
    node = om.fetch(node_id)
    node.demap()
    node.remove()


@funlog
def create_visible_edge(edge_id):
    global _working_map
    if not om.has(edge_id[0]):
        warn(f"Visualization layer cannot create an edge as node: {edge_id[0]} is not known")
        return
    if not om.has(edge_id[1]):
        warn(f"Visualization layer cannot create an edge as node: {edge_id[1]} is not known")
        return

    src = om.fetch(edge_id[0])
    tgt = om.fetch(edge_id[1])
    # ctype = edge_id[2]
    edge = _working_map.edges.get(edge_id, None)
    if not hasattr(edge.source, "visualization"):
        edge.source.set_untracked_attr("visualization", src)
    if not hasattr(edge.target, "visualization"):
        edge.target.set_untracked_attr("visualization", tgt)
    if edge is None:
        show_error(_("Received create edge from Map but missing edge"))  # noqa: F821
        return
    if hasattr(vm, edge.__class__.__name__.lower()):
        func = getattr(vm, edge.__class__.__name__.lower())
    else:
        func = render_completed_conn
    func(edge)
    src.handle_map_edge_src_create(edge)
    tgt.handle_map_edge_tgt_create(edge)


@funlog
def delete_visible_edge(edge_id):
    global _working_map
    try:
        edge = om.fetch(edge_id)
        src = om.fetch(edge_id.split(":")[0])
        tgt = om.fetch(edge_id.split(":")[1])
        src.handle_map_edge_src_delete(edge)
        tgt.handle_map_edge_tgt_delete(edge)
        edge.connection.demap()
        edge.connection.remove()
    except KeyError:
        """
        Contains edges will not have any visual artifacts stored in the object manager.
        There may well be other edges without visualizations to remove.
        """
        pass


@funlog
def addNode(item):
    """Add node to the map and connect visual to logical entities."""
    global _working_map
    if hasattr(item, 'mapentry'):
        notice("Adding item {} to graph".format(item.mapentry))
        if not hasattr(item.mapentry, 'id'):
            item.mapentry.id = ng.id()
        if not hasattr(item.mapentry, "name"):
            verbose("Item to add to visualization seems not to be named:{}"
                    .format(item))
            item.mapentry.name = item.uid
            # _working_map.set_node(item.mapentry.id, item.mapentry)
            # import nord.sigurd.managers.objectmanager as objects
            # objects.store(item.mapentry.id, item)

        # _working_map.set_node(item.mapentry.id, item.mapentry)
        node_changed(item)
    # check to see if I am in a lower level container
    # if so, I'll need to add the appropriate edge
    # as established in vismap.conf. the addtocontainer
    # mapping calls a function in the mapped module/class
    # that creates the needed edge between the container
    # and the new item
    container_name = cm.get_current_name()
    if not cm.is_top() and container_name in _working_map.nodes:
        edge = Edge(None)  # use None as graph to bypass CDC the edge will be added later
        edge.set_nodes(_working_map.nodes[container_name], item.mapentry, 'NA')
        # call the vismanager wrapped function to convert this edge to the
        # correct type
        edge = vm.addtocontainer(edge)
        _working_map.add_edge(edge.source, edge.target, edge.rule, edge)


@funlog
def isConnectable(src, tgt, ctype):
    """
    Determine if the requested connection is permitted.

    Check to see if there is a rule from the src to the ctype
    and from the ctype to the tgt as well as the ctype of edge
    """

    # the edges of type docs and type flow can be connected to any target
    # type.
    if ctype in ('docs', 'flow') or ctype.endswith('flow'):
        return True
        # Because anything can be flowed to or from, as well as commented
        # upon

    if isinstance(src, InstanceOf):
        # The source is a reference to another node,
        # so we need to use that node to determine connectability
        tmp = src.get_parent()
        if tmp:
            src = tmp

    if isinstance(tgt, InstanceOf):
        # The target is a reference to another node,
        # so we need to use that node to determine connectability
        tmp = tgt.get_parent()
        if tmp:
            tgt = tmp

    rv, _, _, _ = Map.find_edge_rules(src, tgt, ctype, [tgt.get_space(), src.get_space(), "nord"])
    return rv


@funlog
def addConnection(src, tgt, ctype):
    """
    Add the requested connection between visible nodes, if possible.

    Interrogates the source and target. If there aren't any mapentries,
    return False if there are mapentries (in src and target) check to
    see if the needed rules exist between the types of nodes. If so,
    create and store the ctype edge and return True
    """
    global _working_map

    # Check if the src or target are delegates
    # a delegate visually proxies a connection on
    # behalf of another visnode.. The delegate, may, or
    # may not also have a map entry. It is up to the
    # delegate to determine if it is accepting the
    # connection for itself, or not.

    src_delegate = None
    tgt_delegate = None

    if src.is_delegate(ctype, 'as_src'):
        if not src.connect_allowed(tgt, ctype, 'as_src'):
            return False
        src_delegate = {'uid': src.uid.split(':')[0],
                        'rule_type': ctype,
                        'direction': 'as_src'}
        src = src.get_delegator(ctype, 'as_src')

    if tgt.is_delegate(ctype, 'as_tgt'):
        if not tgt.connect_allowed(src, ctype, 'as_tgt'):
            return False
        tgt_delegate = {'uid': tgt.uid.split(':')[0],
                        'rule_type': ctype,
                        'direction': 'as_tgt'}
        tgt = tgt.get_delegator(ctype, 'as_tgt')

    # If the node is just visual and has no corresponding map entry
    # return False.
    if not hasattr(src, 'mapentry') or not hasattr(tgt, 'mapentry'):
        return False

    # check to see if the involved nodes (from a visible perspective) are good with the connection
    if not tgt.connect_allowed(src, ctype, 'as_tgt') or\
            not src.connect_allowed(tgt, ctype, 'as_src'):
        return False

    if isConnectable(src.mapentry, tgt.mapentry, ctype):

        # make sure this Edge is not already in the Map does not
        # already exist:
        if ctype in src.mapentry.get_target_rules():
            for edge in src.mapentry.get_targets(ctype):
                if tgt.mapentry.id == edge.target.id:
                    # Make sure this edge is in the stored set of edges....
                    # This is needed because edges created non-interactively
                    # can not yet exist in the working map (dataobject.show_output
                    #  for instance)
                    if (edge.source.id, edge.target.id, edge.rule) not in _working_map.edges:
                        _working_map.add_edge(edge.source, edge.target, edge.rule, edge)

                    return True

        # we also need to check for operate in reverse edges
        if ctype in src.mapentry.get_source_rules():
            for edge in src.mapentry.get_sources(ctype):
                if hasattr(edge, '_operate_in_reverse') and src.mapentry.id == edge.source.id:

                    if (edge.source.id, edge.target.id, edge.rule) not in _working_map.edges:
                        _working_map.add_edge(edge.source, edge.target, edge.rule, edge)

                    return True

        # if we made it here, it means that there is not already an edge
        # between these nodes.
        return create_edge(src.mapentry, tgt.mapentry, ctype, src_delegate, tgt_delegate)
    return False


def remove_edge(item):
    """
    Remove an edge from the underlying map if it should exist there.
    This will trigger the CDC system to in turn propogate the events
    to visibly remove the edge as appropriate.
    """
    global _working_map
    if hasattr(item, 'mapentry'):
        _working_map.rem_edge(item.mapentry)


def remove_node(item):
    """
    Remove a node from the underlying map if it should exist there.
    This will trigger the CDC system to in turn call removeNode
    as appropriate.
    """
    global _working_map
    if hasattr(item, 'mapentry') and hasattr(item.mapentry, 'id'):
        _working_map.rem_node(item.mapentry.id)
    else:
        removeNode(item)


@funlog
def removeNode(item):
    """Remove node from graph and any connected edges."""
    global _working_map
    if hasattr(item, 'mapentry') and hasattr(item.mapentry, 'id'):
        if item.mapentry.id in _working_map.nodes:
            n = _working_map.nodes[item.mapentry.id]
            for srule in list(n.get_source_rules()):
                for edge in list(n.get_sources(srule)):
                    if hasattr(edge, 'visualization'):
                        edge.visualization.remove()
                    edge.source.rem_target(n, edge)

            for trule in list(n.get_target_rules()):
                for edge in list(n.get_targets(trule)):
                    if hasattr(edge, 'visualization'):
                        edge.visualization.remove()
                    n.rem_target(edge.target, edge)
                    edge_changed(edge)

            del _working_map.nodes[item.mapentry.id]
            node_changed(item)
        else:
            err("Remove Node from visual map requested, but no node found \
                with id: {}"
                .format(item.mapentry.id))
    else:
        err("Remove node from visual map requested of a node with an \
            invalid structure")


@funlog
def removeConnection(conn):
    """Remove the passed connection from the graph/map.

    Appears to be obsolete now....
    """
    global _working_map
    if hasattr(conn, 'mapentry'):
        edge = (conn.mapentry.source.id,
                conn.mapentry.target.id,
                conn.mapentry.rule)

        if edge in _working_map.edges:
            edge_changed(edge)
            del _working_map.edges[edge]
        else:
            err("Remove Edge from visual map requested, but no edge \
                matching {} found"
                .format(conn.mapentry))
    else:
        err("Remove Edge from visual map requested of an edge with an \
            invalid structure")


@funlog
def save():
    """
    Write the map/graph data into its JSON formatted data file.

        Go through the map data

        Should this be implemented entirely in Walujapi?

        or, is it better for the vismap to handle extracting all
        the details out of each node and edge....
        Maybe Node'_working_map and edges in walujapi should have a toJson method

    """
    global _working_map
    if _working_map.mapfile is None:
        _working_map.mapfile = DEFAULTOUTFILENAME
    if not hasattr(_working_map, 'name'):
        _working_map.name = ng.get()
    with open(_working_map.mapfile, 'w') as f:
        info("Saving Map to {}".format(_working_map.mapfile))
        try:
            json_data = json.dumps(_working_map.to_dict(), sort_keys=True, indent=4)
            f.write(json_data)
        except Exception as e:
            err("Failed to save working map to {} due to: {}"
                .format(_working_map.mapfile, e))


@funlog
def get_map_dict():
    """Return the dictionary of the current working map."""
    global _working_map
    if not hasattr(_working_map, 'name'):
        _working_map.name = ng.get()
    clear_changes()
    return _working_map.to_dict()


@funlog
def get_working_map():
    """Return the working map object"""
    global _working_map
    return _working_map


@funlog
def get_nodes():
    """Return the dict of the known nodes."""
    global _working_map
    if _working_map is None:
        return {}
    else:
        return _working_map.nodes


@funlog
def get_edges():
    """Return a dict of the known edges"""
    global _working_map
    if _working_map is None:
        return {}
    else:
        return _working_map.edges


@funlog
def get_node(nid):
    """Return the requested node, or None"""
    global _working_map
    if _working_map is None:
        return None
    else:
        return _working_map.nodes.get(nid)


_working_map.add_change_handler(handle_map_changes)
