"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.utils.log import warn
from nord.sigurd.shell.labeledmodel import LabeledModel
from nord.sigurd.shell.modelmenu import ModelMenu
from nord.sigurd.shell.menuItem import MenuItem
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.event as event
from nord.sigurd.visibles.ui.text import Text
import nord.nord.InstanceOf


class VisNode(LabeledModel):
    @funlog
    def __init__(s, uid, ntype, app, space="nord"):
        super().__init__(uid, ntype, app, space=space)
        s.input = None
        s.menu = ModelMenu()
        s.instances = []

        # have the super.super class (Model) interrogate
        # nord for which connections are available to this
        # Node type, and put them into a submenu
        s.connMenu = s.create_connection_menu(s.menu)

        @event.callback
        def callback(*args, **kwargs):
            warn("==== Calling VisNode Selected Callback")
            em.model_toggle_selected(s)
        s.menu.add_item(MenuItem('menudefselect', callback, title="Toggle Selected"))

        # this adds the Model generated connection menu
        s.menu.add_item(s.connMenu)

        # this has to be in visnode rather than labeledmodel as there is not
        # a menu yet in labeledmodel...
        @event.callback
        def edit_label(*args, **kwargs):
            s.label.handle_clicked(*args, **kwargs)

        s.menu.add_item(MenuItem('textmenuicon', edit_label, title="Re-Name"))

        # after the associated nord.nodes module has been connected,
        # check to see if it has properties, and if so, enable the
        # displaying of the property overlay screen
        def checkForProperties():
            if hasattr(s.mapentry, '_properties'):
                @event.callback
                def show_properties_cbk(*args, **kwargs):
                    warn("----< Displaying VisNode Property Overlay >----")
                    s.menu.hide()
                    em.show_property_overlay(s)

                s.menu.add_item(MenuItem('menushowproperties', show_properties_cbk))

            if isinstance(s.mapentry, nord.nord.InstanceOf):
                # Add a menu item to allow for "instancing" of this visnode, if supported by the visnode
                # or its subclass
                @event.callback
                def cbk_create_instance(*args, **kwargs):
                    s.create_instance(*args)

                s.menu.add_item(MenuItem('menuinstanceof', cbk_create_instance, title="Add Instance Of"))

        s.onMappedFuncs.append(checkForProperties)

    @funlog
    def apply(s):
        """
            This function should be overriden by the child classes
        """
        warn(f"{s.__class__.__name__} did not override the apply method, evn though it was called")

    @funlog
    def revert(s):
        """
            This function should be overriden by the child classes
        """
        warn(f"{s.__class__.__name__} did not override the revert method, evn though it was called")

    @funlog
    def handle_clicked(s, *args, **kwargs):
        em.post_pending('model_toggle_selected', s)
        em.data_show_menu(s)

    @funlog
    def show_output(s, text):
        """Show the passed output."""
        pos = s.getWorldPos()
        if not hasattr(s, '_textout'):
            s._textout = Text(f"Output_{s.mapentry.id}", s.app,
                              parent=cm.get_by_name(s.container),
                              default_contents=text,
                              initscale=0.5,
                              one_line=True)
        else:
            s._textout.setText(s._textout.getString() + "\n" + text)
        s._textout.setPos(pos[0], pos[1] - 15.0, pos[2])

    @funlog
    def show_prompt(s, cbk):
        if s.input:
            cbk(s.input)
        else:
            # No event / state managing?
            # Not sure that is the correct approach, if it is even feasible.
            # s.move_cam_to()
            if not hasattr(s, '_textin'):
                pos = s.getWorldPos()
                s._textin = Text(f"Input_{s.mapentry.id}", s.app,
                                 parent=cm.cwc(),
                                 default_contents="Input required",
                                 initscale=0.5,
                                 one_line=True)
                s._textin.setPos(pos[0], pos[1] - 13.0, pos[2])

                def wrapped_cbk(widget):
                    def rv(new_text):
                        cbk(new_text)
                        # widget.hide()
                    return rv

                s._textin.addCallback('on_change', wrapped_cbk(s._textin))

    @funlog
    def get_new_instance_uid(self):
        """generate a derived uid for a new instance of this visnode"""
        return f'{self.uid}:Instance{len(self.instances)}'

    @funlog
    def _new_copy_of(self):
        """
        Simplistic abililty to create a new instance of self, so long as
        the descendent class does not change the arguments to __init__
        if it does, calls to this method will throw an exception ...
        """
        uid = self.get_new_instance_uid()
        # Most subclasses of visnode should just require
        # these two argments...
        return self.__class__(uid, self.app)

    @funlog
    def create_instance(self, *args, **kwargs):
        """Create an InstanceOf this visnode"""
        # Identify if this node is already an instance of another
        # visnode. If it is not, create a new instance of this node
        # if so, create an instance of the original instance
        if not hasattr(self, '_parent_instance'):
            # figure out the youngest descendant class
            # this visnode is, and, create a new one of them
            # setting the _parent_instance of that to this(self).
            # Creating a new instance of the descendant visnode
            # uses the _new_copy_of visnode method
            # if a descendant's constructor takes more args
            # than visnode, it will have to implement _new_copy

            import nord.sigurd.shell.vismap as vm
            vm.create_instance(self, args[0])
        else:
            # self._parent_instance.create_instance(caller=self)
            self._parent_instance.create_instance(*args, **kwargs)
