"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Utility module to faciltate log messages.

Thiscan filter output, send it to various locations, and
colorize the messages to the terminal based upon severity.
"""


import traceback
import datetime
import sys
import syslog
import os
import inspect
import re

_highlight_on = False
_meta_off = True
_def_show_message = True

DEBUG         = DBUG = DBG = 7  # noqa: E221
VERBOSE       = VERB = VRB = 6  # noqa: E221
INFORMATIONAL = INFO = INF = 5  # noqa: E221
NOTICE        = NOTE = NOT = 4  # noqa: E221
WARNING       = WARN = WRN = 3  # noqa: E221
ERROR         = ERRR = ERR = 2  # noqa: E221
CRITICAL      = CRIT = CRT = 1  # noqa: E221
MATCHED       = MTCH = MTC = 0  # noqa: E221
OFF                        = 99  # noqa: E221


_lvl2text = {
    DBUG: 'debug',
    VERB: 'verbose',
    INFO: 'informational',
    NOTE: 'notice',
    WARN: 'warning',
    ERRR: 'error',
    CRIT: 'critical',
    MTCH: 'matched',
    OFF:  'off'  # noqa: E241
}

_lvl2color = {
    MTCH: '\033[1m\033[44m',  # bold light blue
    CRIT: '\033[1m\033[91m',  # bold light red
    ERRR: '\033[1m\033[31m',  # bold red
    WARN: '\033[1m\033[33m',  # bold yellow
    NOTE: '\033[93m',         # light yellow
    INFO: '\033[36m',         # cyan
    VERB: '\033[2m\033[32m',  # dim, light green
    DBUG: '\033[2m\033[37m',  # dim, light gray
    OFF:  '\033[0m'  # noqa: E241
}

level = DEBUG
defChannel = 'STDERR'
_logfile = None
_lfh = None
_logfileopened = None
_doRotate = False

_dbgFileRegex = None

_lvl2syslog = {
    DEBUG: syslog.LOG_DEBUG,
    VERBOSE: syslog.LOG_DEBUG,
    INFORMATIONAL: syslog.LOG_INFO,
    NOTICE: syslog.LOG_NOTICE,
    WARNING: syslog.LOG_WARNING,
    ERROR: syslog.LOG_ERR,
    CRITICAL: syslog.LOG_CRIT
}


def file_rotate():
    """
    Trigger file rotation on next message output.

    Do so by setting _doRotate boolean.
    """
    global _doRotate
    _doRotate = True


def _flog(x, depth):
    """
    File logger.

    Responsible for opening, closing, and rotating the
    file based log. If configured and enabled.

    By default logfile will be rotated daily, it can be rotated manually, by calling
    file_rotate
    """
    global _doRotate, _logfileopened, _logfile, _lfh
    if _logfile is not None:
        now = datetime.datetime.now()
        if _lfh is None or _logfileopened.day != now.day or _doRotate:
            if _lfh is not None:
                _lfh.close()
                os.rename(_logfile, "{}.{}".format(_logfile, _logfileopened.isoformat()))
                _doRotate = False
            try:
                _lfh = open(_logfile, 'w+')
                _logfileopened = now
            except Exception:
                error("Unable to open logfile '{}'".format(_logfile))
                _logfile = None
                return

        now = datetime.datetime.now()
        m = "{} <{}> - {}".format(now.isoformat(), _get_meta(depth), x)
        _lfh.write("{}\n".format(m))


def _get_meta(depth):
    """
    Use the inspection module to get the file and line where logging was requested.

    This has negative performance consequences. Use only for debugging...
    """
    if depth >= 0:
        """
        the frame should be one item ahead
        of this function in the stack
        """
        frame = inspect.stack()[depth][0]
        fname = inspect.getframeinfo(frame).filename.split(os.path.sep)[-1]
        line = inspect.getframeinfo(frame).lineno
        return "{}:{}".format(fname, line)
    else:
        return " --- "


def _match_meta(rgx):
    """Interrogate the frame data, returning regex matches."""
    for frm in inspect.stack():
        frame = frm[0]
        fname = inspect.getframeinfo(frame).filename.split(os.path.sep)[-1]

        if len(rgx.findall(fname)) > 0:
            return True
    return False


def _print(msg, lvl, depth):
    """Print out the message and meta data to STDERR."""
    now = datetime.datetime.now()
    show_msg = _def_show_message
    if _highlight_on:
        meta = _get_meta(depth)
        # if _dbgFileRegex is not None and _match_meta(_dbgFileRegex):
        if (_dbgFileRegex is not None and
                len(_dbgFileRegex.findall(meta)) > 0) or\
                _dbgFileRegex is None:
            show_msg = True
            lvl = MATCHED
    elif _meta_off:
        meta = " -- "
        show_msg = (_dbgFileRegex is not None and
                    len(_dbgFileRegex.findall(meta)) > 0) or\
            _dbgFileRegex is None
    else:
        meta = _get_meta(depth)
        show_msg = (_dbgFileRegex is not None and
                    _match_meta(_dbgFileRegex)) or\
            _dbgFileRegex is None

    if show_msg:
        m = "{}{} [{}] <{}> - {}{}"\
            .format(_lvl2color[lvl],
                    now.isoformat(),
                    _lvl2text[lvl],
                    meta,
                    msg,
                    _lvl2color[OFF])
        sys.stderr.write("{}\n".format(m))


def _syslog(x, depth):
    global level
    syslog.syslog(_lvl2syslog[level], "{} - {}".format(x, _get_meta(depth)))


_channels = {
    'STDERR': lambda x, lvl, depth: _print(x, lvl, depth),
    'SYSLOG': lambda x, lvl, depth: _syslog("[{}] {}"
                                            .format(_lvl2text[lvl],
                                                    x),
                                            depth),
    'FILE': lambda x, lvl, depth: _flog("{}[{}]{}{}"
                                        .format(_lvl2color[lvl],
                                                _lvl2text[lvl],
                                                x,
                                                _lvl2color[OFF]),
                                        depth)
}

_levels = {
    'STDERR': DEBUG,
    'SYSLOG': OFF,
    'FILE': OFF
}


def setfile(fname):
    """
    Establish the logfile name.

    And enables logging to it.
    """
    global _logfile, _levels, level
    _logfile = fname
    _levels['FILE'] = level


def crit(msg, depth=5):
    """Shortcut to logging at this level."""
    log(CRIT, msg, depth)


def err(msg, depth=5):
    """Shortcut to logging at this level."""
    log(ERR, msg, depth)


def warn(msg, depth=5):
    """Shortcut to logging at this level."""
    log(WARN, msg, depth)


def note(msg, depth=5):
    """Shortcut to logging at this level."""
    log(NOTE, msg, depth)


def info(msg, depth=5):
    """Shortcut to logging at this level."""
    log(INFO, msg, depth)


def verb(msg, depth=5):
    """Shortcut to logging at this level."""
    log(VERB, msg, depth)


def dbug(msg, depth=5):
    """Shortcut to logging at this level."""
    log(DBUG, msg, depth)


critical = crit
error = err
warning = warn
notice = note
information = info
verbose = verb
debug = dbg = dbug


def log(lvl, msg, depth):
    """
    The workhorse function of this module.

    if the input level is more critical (smaller number)
    than the set level for the channel, then call that channel
    to log the message
    """
    global level
    for chan in _levels.keys():
        if _levels[chan] <= level:
            _channels[chan](msg, lvl, depth)


def set_level(l, chan=defChannel):
    """Set the logging level."""
    global level
    if l >= CRIT and l <= DBUG or l == OFF:
        level = l
        if chan in _levels:
            _levels[chan] = l
        else:
            error("Invalid log channel '{}' requested".format(chan))
    else:
        error("Invalid log level '{}' requested".format(l))


def set_dbg_filter(rgex):
    """"Set the regular expression to use for filtering messages."""
    global _dbgFileRegex
    try:
        _dbgFileRegex = re.compile(rgex)
    except Exception:
        error("Invalid debug regex requested: '{}'".format(rgex))
        pass


def tb(t, tp=None, val=None, msg=None):
    """Emit a stack trace, typically in **except** blocks."""
    stck = None
    if tp is None or val is None:
        stck = traceback.format_tb(t)
    else:
        stck = traceback.format_exception(tp, val, t)
    cnt = 0
    s = None
    for i in stck:
        if s is None:
            s = "Stack Trace [%d] %d: %s" % (os.getpid(), cnt, i)
        else:
            s = "%s Stack Trace [%d] %d: %s" % (s, os.getpid(), cnt, i)
        cnt = cnt + 1
    if msg is not None and s is not None:
        s = "%s: %s" % (msg, s)
    if msg is not None and s is None:
        s = msg

    err(s)
