"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The decorators in this module are used throughout this program to aid troubleshooting.

The funlog methods decorate most functions so that the function call and
its arguments, and possibly return values, will be logged.
"""

import nord.sigurd.utils.log as log
import re
import time

_matcher = re.compile(r'.*')
_nestlvl = 0

_time_summary = {}


def postcallfunlog(func):
    """
    Decorator to print function call details.

    This decorator outputs parameters names, effective values and their return values

    This prints nested function calls before their caller is printed
    """
    def wrapper(*func_args, **func_kwargs):
        global _nestlvl
        global _matcher
        _nestlvl += 1
        rv = func(*func_args, **func_kwargs)
        if log.level >= log.DEBUG and log.level != log.OFF:
            if len(_matcher.findall(func.__name__)) > 0:
                arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
                args = func_args[:len(arg_names)]
                defaults = func.__defaults__ or ()
                args = args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):]
                params = list(zip(arg_names, args))
                args = func_args[len(arg_names):]
                if args:
                    params.append(('args', args))
                if func_kwargs:
                    params.append(('kwargs', func_kwargs))
                spacer = "^"
                for i in range(0, _nestlvl):
                    spacer += "^"
                log.debug('⭅' + spacer + ' ' +
                          func.__module__ + ":" +
                          func.__name__ + '( ' + ', '
                          .join('%s = %r' % p for p in params) +
                          ' ) returned: ' + repr(rv), depth=6)
        _nestlvl -= 1
        return rv
    return wrapper


def sandwichfunlog(func):
    """
    Decorator to print function call details.

    This outputs parameters names, effective values and their return values

    This prints nested function calls before their caller is printed,
    and then after as well
    """
    def wrapper(*func_args, **func_kwargs):
        global _nestlvl
        global _matcher
        global _time_summary

        _nestlvl += 1
        if log.level >= log.DEBUG and log.level != log.OFF:
            if len(_matcher.findall(func.__name__)) > 0:
                arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
                args = func_args[:len(arg_names)]
                defaults = func.__defaults__ or ()
                args = args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):]
                params = list(zip(arg_names, args))
                args = func_args[len(arg_names):]
                if args:
                    params.append(('args', args))
                if func_kwargs:
                    params.append(('kwargs', func_kwargs))
                spacer = "⇨"
                for i in range(0, _nestlvl):
                    spacer += "⇨"
                log.debug(" ")
                log.debug(spacer + '⭆ ' +
                          func.__module__ + ":" +
                          func.__name__ + '( ' + ', '
                          .join('%s = %r' % p for p in params) +
                          ' ) ', depth=6)
        tstart = time.time()
        rv = func(*func_args, **func_kwargs)
        tend = time.time()

        if "{}.{}".format(func.__module__, func.__name__) not in _time_summary:
            _time_summary[
                "{}.{}".format(func.__module__, func.__name__)
            ] = (1, tend - tstart, None, None)

        else:
            runt = tend - tstart
            cnt, avg, mn, mx = _time_summary["{}.{}".format(func.__module__, func.__name__)]
            avg = ((avg * cnt) + runt) / (cnt + 1)
            cnt += 1
            if mn is None or runt < mn:
                mn = runt
            if mx is None or runt > mx:
                mx = runt

            _time_summary["{}.{}".format(func.__module__, func.__name__)] = (cnt, avg, mn, mx)

        if log.level >= log.DEBUG and log.level != log.OFF:
            if len(_matcher.findall(func.__name__)) > 0:
                arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
                args = func_args[:len(arg_names)]
                defaults = func.__defaults__ or ()
                args = args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):]
                params = list(zip(arg_names, args))
                args = func_args[len(arg_names):]
                if args:
                    params.append(('args', args))
                if func_kwargs:
                    params.append(('kwargs', func_kwargs))
                spacer = "⇦"
                for i in range(0, _nestlvl):
                    spacer += "⇦"
                log.debug("[{}s]".format(tend - tstart) +
                          '⭅' + spacer + ' ' +
                          func.__module__ + ":" +
                          func.__name__ + '( ' + ', '
                          .join('%s = %r' % p for p in params) +
                          ' ) returned: ' + repr(rv), depth=6)
        _nestlvl -= 1
        return rv
    return wrapper


def precallfunlog(func):
    """
    Decorator to print function call details.

    Outputs parameters names and effective values WITHOUT return values

    This prints nested function calls AFTER their caller is printed.
    """
    def wrapper(*func_args, **func_kwargs):
        global _matcher
        if log.level >= log.DEBUG and log.level != log.OFF:
            if len(_matcher.findall(func.__name__)) > 0:
                arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
                args = func_args[:len(arg_names)]
                defaults = func.__defaults__ or ()
                args = args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):]
                params = list(zip(arg_names, args))
                args = func_args[len(arg_names):]
                if args:
                    params.append(('args', args))
                if func_kwargs:
                    params.append(('kwargs', func_kwargs))
                log.debug('⭆ ' +
                          func.__module__ + ":" +
                          func.__name__ + '( ' + ', '
                          .join('%s = %r' % p for p in params) +
                          ' ) ', depth=6)

        rv = func(*func_args, **func_kwargs)
        return rv
    return wrapper


funlog = sandwichfunlog


def set_matcher(rgex):
    """Set the regex to limit output only to function names which match the expression."""
    global _matcher
    try:
        tmp = re.compile(rgex)
        _matcher = tmp
    except Exception:
        log.error("Unable to compile regex: '{}'".format(rgex))


def show_profile():
    """Print out the summary profile information."""
    global _time_summary

    for fk in sorted(_time_summary,
                     key=lambda kv: (_time_summary[kv][0], _time_summary[kv][1])):
        cnt, avg, mn, mx = _time_summary[fk]
        print("  {}: Ran: {}, Fastest runtime: {}s, Avg: {}s, longest: {}s"
              .format(fk, cnt, mn, avg, mx))
