"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The node managing the visualization of containers.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import warn, info
from nord.sigurd.shell.menuItem import MenuItem
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.event as event
from nord.sigurd.shell.visnode import VisNode
from nord.sigurd.exceptions import UIException


class Container(VisNode):
    """This class manages the special activities related to the container."""

    @funlog
    def __init__(self, uid, app, originof=None):
        """
        Initialize the Container instance.

        Since we will not have the mapentry attribute
        until the vismap has added this object, we'll add this callback
        to the Model, so when added, it will then establish the visual
        aspects of this object
        """
        super().__init__(uid, "container", app)
        if originof is None:
            self._originmodel = None

            def on_add_to_map():
                cm.add(self.mapentry.id)

            def on_all_add_to_map():
                self.establish_delegates()

            self.onMappedFuncs.append(on_add_to_map)
            self.onAllMappedFuncs.append(on_all_add_to_map)

            @event.callback
            @funlog
            def callback2(*args, **kwargs):
                warn("  ---> Calling Enter Container Callback")
                em.enter_container(self.mapentry.id)

            self.menu.add_item(MenuItem("entercontainer", callback2))
        else:
            self._originof = originof

    def show_origin(self):
        self.establish_origin_model()
        self._originmodel.show()

    def hide_origin(self):
        self.establish_origin_model()
        self._originmodel.hide()

    def establish_origin_model(self):
        if self._originmodel is not None:
            return
        self._originmodel = Container(f"origin_{self.uid}", self.app, originof=self)
        self._originmodel.setPos(0.0, 0.0, 0.0)
        self._originmodel.setScale(0.5)
        # This puts the namegen generated UID (readable name...) As the label of the origin
        # minus the prefixed "Container: " bit.
        if self.name == "container":
            label = self.uid.replace("Container: ", '')
        else:
            label = self.name
        self._originmodel.setLabelText(f"-> {label} <-")
        self._originmodel.menu.clear_items()
        self._originmodel.menu.add_item(MenuItem('exitcontainer', em.exit_container, title="Exit Container"))

        @event.callback
        @funlog
        def exit_current_container(*args, **kwargs):
            em.exit_container()

    @funlog
    def establish_delegates(self):
        """
        Find all the "sourceless" nodes contained within
        and set those as decorations/argument candidates
        if those nodes are eligible to be passed to
        """
        # We need to backup the delegated edges prior to
        # clearing out the delegates
        edges = {}
        for dirk, dirv in self.delegates.items():
            for rulek, rulev in dirv.items():
                for dgate in rulev:
                    if dirk not in edges:
                        edges[dirk] = {}
                    if rulek not in edges[dirk]:
                        edges[dirk][rulek] = []
                    for rulevis in rulev:
                        edges[dirk][rulek].extend(rulevis.get_delegated_conns())

        self.clear_delegates()
        targets = list()
        for edge in self.mapentry.get_targets('contains'):
            targets.append(edge.get_target())

        for tgt in sorted(targets):
            if len(tgt.get_source_rules()) == 1 and tgt.eligible_for("assign"):
                dec = self.add_delegate(tgt.id,
                                        tgt.visualization.name,
                                        "pass",
                                        "as_tgt",
                                        tiptext=f"Pass arg into {tgt.name}",
                                        space=tgt.get_space())

                tgt.visualization.track_decoratee(self.decorationmanager, dec.uid, "pass", "as_tgt")
                if "as_tgt" in edges and "pass" in edges["as_tgt"]:
                    for visedge in edges["as_tgt"]["pass"]:
                        self.set_delegated_conn(dec, visedge)
                    for k, v in edges["as_tgt"].items():
                        if k != "pass":
                            raise UIException(f"Unknown {k} in cached connections ... probably an implementation error")
                # Should probably raise some sort of implementation exception
                # if there are non-as_tgt directions and non-pass rules
                # in edges, as we don't do anything wth them...
                for k, v in edges.items():
                    if k != "as_tgt":
                        raise UIException(f"Unknown {k} in cached connections/edges, probably an implementation error")

    @funlog
    def enter(self):
        """
        Upon entering a container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode apply function passing this container
        as the argument
        """
        for edge in self.mapentry.get_sources("use"):
            info("Container rendering edge: {}".format(edge))
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.apply()

        self.show_origin()

    @funlog
    def exit(self):
        """
        Upon leaving the container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode revert function passing this container
        as the argument
        """
        for edge in self.mapentry.get_sources("use"):
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.revert()

        self.establish_delegates()
        self.hide_origin()
