"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Class as module providing visualization for importing other graphs.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import debug, err, critical
from nord.sigurd.shell.visnode import VisNode
from nord.sigurd.shell.menuItem import MenuItem
from nord.sigurd.shell.submenu import SubMenu
from nord.sigurd.visibles.reference import Reference
import nord.sigurd.managers.objectmanager as om
import nord.sigurd.utils.namegen as namegen
import nord.sigurd.shell.mainmenu as mainmenu
import nord.sigurd.managers.event as event
import nord.nord.utils.SpaceManager as use
import nord.nord.Map as Map
import json as json


class Graph(VisNode):
    """Visualization to connect an external graph with the current one."""

    def __init__(self, uid, app):
        """Initialize the graph visualization."""
        super().__init__(uid, 'graph', app)
        self.menuItem = None

    @funlog
    def apply(self):
        """
        Create a SubMenu and inject it into the mainmenu.

        Extract all the top level containers from the graph
        and add a menuItem for each of them to the SubMenu
        """
        if self.mapentry.is_set_property('mapfile'):
            path = use.path(self.mapentry.get_property('mapfile'))
            icon_name = self.mapentry.get_property('menuicon')

            with open(path, 'r') as f:
                submap = Map(json.load(f))
                debug("Loaded the map in {} resulting in graph: \n{}".format(path, submap))

                # go through all the Nodes in the submap
                # for each Node of type Container in the Graph
                # which does not have any sourceself, add a menuItem
                # for that container.
                # For each of the toplevel containers
                # recursively add submenu's for each contained container
                # and create menu_items for each contained container

                menu_items = list()
                for nk in submap.nodes:
                    node = submap.nodes[nk]
                    if node.__class__.__name__ == 'Container' and\
                            len(node.get_sources('contains')) == 0:
                        debug("Found a container node named: {}".format(node.name))

                        def gen_callback(name, refid, menuicon):

                            @event.callback
                            @funlog
                            def func(*args, **kwargs):
                                """
                                Add a node to the graph.

                                It is a special node from the runtime perspective in that it
                                references the selected container in the imported Map..


                                ADD CODE TO add the item to the Screen...
                                """
                                critical("Called an click of: {} with id {}".format(name, refid))
                                reference = Reference(namegen.get(), self.app, menuicon)
                                om.add_item(args[0], self.app, reference)
                                reference.mapentry.set_property('parentid', self.mapentry.id)
                                reference.mapentry.set_property('itemid', refid)

                            return func

                        name = node.get_property('menuicon')
                        rname = node.name
                        nid = node.id

                        mi = MenuItem(name, gen_callback(rname, nid, node.get_property('menuicon')))
                        menu_items.append(mi)

                if len(menu_items) > 1:
                    submenu = SubMenu(icon_name, mainmenu)
                    for mi in menu_items:
                        submenu.add_item(mi, hide_on_click=True)
                    mainmenu.add_item(submenu)
                    self.menuItem = submenu
                elif len(menu_items) == 1:
                    mainmenu.add_item(menu_items[0], hide_on_click=True)
                    self.menuItem = menu_items[0]

        else:
            err("Graph used by container, but no 'mapfile' set for the graph")

    @funlog
    def revert(self):
        """Remove the injected submenu item from the mainmenu."""
        if self.menuItem is not None:
            mainmenu.rem_item(self.menuItem)
            self.menuItem.remove()
            del self.menuItem
            self.menuItem = None
