"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Decorate a parent model with this other model.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
import nord.sigurd.managers.ui.decorationmanager as dm
from nord.sigurd.shell.model import Model
from nord.sigurd.config.globalVars import DEFAULTDECORATIONSCALE
import nord.sigurd.managers.statemanager as sm
# import nord.sigurd.managers.containermanager as cm


class Decoration(Model):
    """Extend/restrict the model class implementing decoration specifics."""

    @funlog
    def __init__(self, name, model, attachment, app, decoratee, tier=1, space="nord", delegate=None, tiptext=None):
        """Initialize the decoration, calling the parent constructor as well."""
        uid = f"{name}:dec@{attachment}__{decoratee.uid}"
        super().__init__(uid, model, app, tag='clickable', space=space, filterable=False)
        self.decoratee = decoratee
        self.attachment = attachment
        self.tier = tier
        self.data.setScale(DEFAULTDECORATIONSCALE)
        self.data.hide()
        self.delegate = delegate
        self.reparentTo(decoratee.getParent())
        if tiptext is not None:
            self.ttext = tiptext
        else:
            self.ttext = f"{decoratee.__class__.__name__}:\n   {name}\n   '{model}'"

    @funlog
    def getWorldPos(self):
        """
        This was overridden from Model as the worldPos defaults
        to 0.0, 0.0, 0.0 when hidden. This, in turn, caused edges
        pointed at the delegate to point to origin, instead of
        the decoratee upon load.
        """
        if self.data.is_hidden():
            return self.decoratee.getPos()
        else:
            return self.data.getPos()

    @funlog
    def update_tooltip(self, tooltip):
        tooltip.set_text(self.ttext)

    @funlog
    def dump_to_dict(self, out_dict):
        """Override dump_to_dict to prevent the decoration from entering the graph."""
        pass

    @funlog
    def show(self):
        """Render the decoration at the attachment point."""
        x, y, z = dm.Attachment.to_pos(self.decoratee, self.attachment, self.tier)
        self.data.setPos(x, y, z)
        self.data.show()

    @funlog
    def is_delegate(self, *args, **kwargs):
        return self.delegate is not None

    @funlog
    def get_delegator(self, *args, **kwargs):
        return self.delegate

    @funlog
    def connect_allowed(self, other, ctype, direction, delegate=None):
        """Check delegatee if connecting to me is allowed, if there is one."""
        if self.delegate:
            return self.delegate.connect_allowed(other, ctype, direction, self)
        return True

    @funlog
    def handle_clicked(self, *args, **kwargs):
        """
        Handle clicking on the decoration....
        """
        if self.delegate:
            sm.post_pending('decoration_clicked', self)
            # sm.data_show_menu(self)
            self.delegate.handle_delegate_click(self.uid)
