"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Text module manages in-world text.
"""

from nord.sigurd.utils.verboser import funlog  # noqa: F401
from nord.sigurd.utils.log import warn
import nord.sigurd.shell.model as model
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.modelmanager as mm
from nord.sigurd.config.globalVars import DEFAULTTEXTSPACEWIDTH,\
    DEFAULTTEXTHSPACING,\
    DEFAULTTEXTHEIGHT
from nord.sigurd.shell.scale import Scale
import nord.sigurd.managers.containermanager as cm
from nord.sigurd.managers.callbackmanager import CallbackManager
from panda3d.core import TextNode, CardMaker, NodePath, DecalEffect
import nord.sigurd.handlers.keyboardhandler as keyboardhandler
from nord.sigurd.shell.glyph import Glyph
import nord.sigurd.shell.worldview as wv
import nord.sigurd.managers.objectmanager as om


remaps = {
    "~": "tilde",
    "`": "backtick",
    "!": "bang",
    "@": "at",
    "#": "octothorpe",
    "$": "dollar",
    "%": "percent",
    "^": "caret",
    "&": "ampersand",
    "*": "asterisk",
    "(": "leftparen",
    ")": "rightparen",
    "{": "leftcurly",
    "}": "rightcurly",
    "[": "leftsquare",
    "]": "rightsquare",
    '"': "doublequote",
    "<": "leftangle",
    ">": "rightangle",
    "'": "singlequote",
    ",": "comma",
    ".": "period",
    "/": "forwardslash",
    "=": "equals",
    "\\": "backslash",
    "?": "question",
    "+": "plus",
    "|": "pipe",
    "-": "dash",
    "_": "underscore",
    ";": "semicolon",
    ":": "colon",
}

mapping = {
    "space": " ",
    "return": "\n",
    "tilde": "tilde",
    "backtick": "backtick",
    "bang": "bang",
    "at": "at",
    "octothorpe": "octothorpe",
    "dollar": "dollar",
    "percent": "percent",
    "caret": "caret",
    "ampersand": "ampersand",
    "asterisk": "asterisk",
    "leftparen": "leftparen",
    "rightparen": "rightparen",
    "leftcurly": "leftcurly",
    "rightcurly": "rightcurly",
    "leftsquare": "leftsquare",
    "rightsquare": "rightsquare",
    "doublequote": "doublequote",
    "leftangle": "leftangle",
    "rightangle": "rightangle",
    "singlequote": "singlequote",
    "comma": "comma",
    "period": "period",
    "forwardslash": "forwardslash",
    "equals": "equals",
    "backslash": "backslash",
    "question": "question",
    "plus": "plus",
    "pipe": "pipe",
    "dash": "dash",
    "underscore": "underscore",
    "semicolon": "semicolon",
    "colon": "colon",
}

name2char = {
    "space": " ",
    "return": "\n",
    "tilde": "~",
    "backtick": "`",
    "bang": "!",
    "at": "@",
    "octothorpe": "#",
    "dollar": "$",
    "percent": "%",
    "caret": "^",
    "ampersand": "&",
    "asterisk": "*",
    "leftparen": "(",
    "rightparen": ")",
    "leftcurly": "{",
    "rightcurly": "}",
    "leftsquare": "[",
    "rightsquare": "]",
    "doublequote": '"',
    "leftangle": "<",
    "rightangle": ">",
    "singlequote": "'",
    "comma": ",",
    "period": ".",
    "forwardslash": "/",
    "equals": "=",
    "backslash": "\\",
    "question": "?",
    "plus": "+",
    "pipe": "|",
    "dash": "-",
    "underscore": "_",
    "semicolon": ";",
    "colon": ":",
}


class Anchor(object):
    """The vertical anchor types."""

    TOP = 1
    CENTER = 2
    BOTTOM = 3
    LEFT = 4
    RIGHT = 5


class HAlign(object):
    """Horizontal alignment types."""

    RIGHT = 1
    CENTER = 2
    LEFT = 3


class Cursor(object):
    """The location of the current edit in the text string."""

    @funlog
    def __init__(self, parent, initscale=1.0, space="nord"):
        """The cursor constructor."""
        self.cursor = mm.load_model("textcursor", space)
        self.cursor.reparentTo(parent)
        self.cursor.setScale(initscale)
        self.cursor.hide()
        self.currentLine = 0

    @funlog
    def setScale(self, scale):  # noqa: N802
        """Expose the setScale method."""
        self.cursor.setScale(scale)

    @funlog
    def hide(self):
        """Hide the cursor."""
        self.cursor.hide()

    @funlog
    def show(self):
        """Show the cursor."""
        self.cursor.show()
        self.cursor.loop("animate")

    @funlog
    def getPos(self):  # noqa: N802
        """Return the world position of the cursor."""
        return self.cursor.getPos()

    @funlog
    def setPos(self, x, y, z):  # noqa: N802
        """Set the world position of the cursor."""
        self.cursor.setPos(x, y, z)


class RowProps(object):
    """The row properties class."""

    @funlog
    def __init__(self, num, start=0):
        """Constructor for the row properties class."""
        self.rownum = num
        self.totalWidth = 0
        self.startIdx = start
        self.endIdx = self.startIdx
        self.cnt = 0

    @funlog
    def add_char(self, glyph):  # noqa: N802
        """Add a character to the row."""
        self.totalWidth += glyph.getWidth() + DEFAULTTEXTHSPACING
        self.endIdx += 1
        self.cnt += 1

    @funlog
    def add_space(self):
        """Add a space to the row."""
        self.totalWidth += DEFAULTTEXTSPACEWIDTH
        self.endIdx += 1
        self.cnt += 1

    @funlog
    def rem_char(self, glyph):
        """Remove the passed glyph from the row."""
        if glyph.char == " ":
            self.totalWidth -= DEFAULTTEXTSPACEWIDTH
        else:
            self.totalWidth -= glyph.getWidth() + DEFAULTTEXTHSPACING
        self.endIdx -= 1
        self.cnt += 1

    @funlog
    def merge_with(self, rp):
        """Merge this row with the passed row."""
        if self.startIdx > rp.startIdx:
            self.startIdx = rp.startIdx
            self.rownum = rp.rownum
        else:
            self.endIdx = rp.endIdx

        self.totalWidth += rp.totalWidth

    @funlog
    def width(self):
        """Return the width of this row."""
        return self.totalWidth


class Text(CallbackManager):
    def __init__(
        self,
        uid,
        app,
        parent=None,
        default_contents="Text",
        initscale=1.0,
        anchor_h=Anchor.LEFT,
        anchor_v=Anchor.TOP,
        align_h=HAlign.LEFT,
        one_line=False,
        space="nord"
    ):
        """Initialize the text instance."""
        CallbackManager.__init__(self, space, app)
        if parent is None:
            parent = cm.cwc()
        self.parent = parent
        self.text_np = None
        self.uid = uid
        # self.data = parent.attachNewNode(tnp)
        self.uid = uid
        self.name = uid
        self.app = app
        self.hasContent = False
        self.cursor = '˙'  # unicode dot above
        self.editMode = False
        self.contents = default_contents
        self.line_pos = 0
        self.tail = ""
        self.anchor_h = anchor_h
        self.anchor_v = anchor_v
        self.align_h = HAlign
        self.anchor_pos = None
        self.one_line = one_line
        self.addEvent("on_change")
        om.store(f"{uid}", self)
        self.draw()

    def show_tooltip(self):
        pass

    def hide_tooltip(self):
        pass

    def getPos(self):
        return self.data.getPos()

    def getX(self):
        return self.data.getX()

    def getY(self):
        return self.data.getY()

    def getZ(self):
        return self.data.getZ()

    def getWorldPos(self):
        return wv.getWorldPos(self.data)

    def getParent(self):
        """
        Override CallbackManager's getParent
        as that references self.text_np, which is the
        wrong type in this object.
        """
        return self.data.getParent()

    """@funlog
    #def addCallback(self, name, func):
    #    pass
    """
    @funlog
    def get_nodep(self):
        return self.data

    @funlog
    def getTightBounds(self):  # noqa: N802
        """Return the bounding box."""
        return self.data.getTightBounds()

    @funlog
    def getString(self):  # noqa: N802
        """Return the text string managed by this instance."""
        return self.contents + self.tail

    @funlog
    def clear(self):
        """Remove contents of this text area."""
        self.contents = ""
        self.tail = ""
        self.draw()

    @funlog
    def centerHorizontally(self):
        """Center the text horizontally, but not vertically."""
        self.center(cent_horiz=True, cent_vert=False)

    @funlog
    def center(self, cent_horiz, cent_vert):
        """
        Center the text according to flags cent_horiz and cent_vert.

        calculate the size of each row of text
        and re-position the glyphs on both sides of zero X
        """
        pass

    @funlog
    def draw(self):
        """
        Method to trigger the rendering of the glyphs.

        Render the contents as glyphs
        parented by self.text_np
        """
        uid = self.uid
        parent = self.parent
        oldPos = None
        if self.text_np is not None:
            oldPos = self.data.getPos()
            self.data.removeNode()
        self.text_np = TextNode(uid)
        cm = CardMaker(f"cm--{uid}")
        if self.editMode:
            contents = self.contents + self.cursor + self.tail
        else:
            contents = self.contents + self.tail
        self.text_np.setText(contents)
        cm.setFrame(self.text_np.getFrameActual())
        tgm = cm.generate()
        self.card = NodePath(tgm)
        self.card.setColor(0.1, 0.1, 0.1, 0.0)
        self.text_np.setFrameColor(1.0, 1.0, 1.0, 1.0)
        self.text_np.setFrameAsMargin(0.2, 0.2, 0.1, 0.1)
        self.text_np.setTextColor(1.0, 1.0, 1.0, 1.0)
        self.card.attachNewNode(self.text_np)
        if parent is not None:
            self.data = parent.attachNewNode(tgm)
            self._oldScale = self.data.getColorScale()
        self.card.setTag("clickable", uid)
        self.card.setEffect(DecalEffect.make())
        self.cardMaker = cm
        self.data.setScale(0.33)
        if oldPos is not None:
            self.data.setPos(oldPos)

        self.data.show()

    @funlog
    def setText(self, text):
        """Allow external entities to set the text contents."""
        self.text_np.setText(text)
        if text != self.contents + self.tail:
            self.contents = text
            self.tail = ""
            self.draw()

    @funlog
    def setPos(self, x, y, z):  # noqa: N802
        """Expose the internal set position method."""
        self.data.setPos(x, y, z)

    @funlog
    def getWidth(self):
        """Return the width of the text"""
        return self.text_np.getWidth()

    @funlog
    def getHeight(self):
        """Return the height of the text"""
        return self.text_np.getHeight()

    @funlog
    def setScale(self, scale):  # noqa: N802
        """Expose internal set-scale methods."""
        raise Exception("Not implemented!!!")
        # self.scale.set(scale)
        # self.cursor.setScale(scale)

    @funlog
    def toggleEdit(self):  # noqa: N802
        """Toggle the editmode of this text area."""
        if self.editMode:
            if len(self.contents) < 1:
                self.contents = "Text"
            self.editMode = False
            self.triggerEvent("on_change", self.getString())
        else:
            # self.data.setHpr(camera.getHpr())
            self.editMode = True

        if not self.hasContent:
            self.contents = ""

        self.draw()

    @funlog
    def handle_clicked(self, *args, **kwargs):
        """Handle a click on any of the visible text."""
        # em.post_pending('hide_model_menu', self.menu)
        # em.post_pending('model_toggle_selected', self)
        keyboardhandler.set_target(self)
        em.toggle_text_edit(self)

    @funlog
    def handle_named_char(self, char):
        """Given the named character input, handle it as needed."""

        if char in ("up", "down", "left", "right", "backspace", "delete"):

            """

                    UP

                    ##
                   #  #
                  #    #
                    ##
                    ##
            """
            if char == "up":
                if "\n" in self.contents:

                    prev_line_end = self.contents.rfind('\n')
                    prev_line_start = self.contents[:prev_line_end].rfind('\n')
                    if prev_line_start < 0:
                        prev_line_start = 0
                    prev_line = self.contents[prev_line_start:prev_line_end]

                    if self.line_pos > len(prev_line) and prev_line_start == 0:
                        prev_line_end -= 1

                    if len(prev_line) > self.line_pos:
                        self.tail = self.contents[prev_line_start + self.line_pos + 1:] + self.tail
                        self.contents = self.contents[:prev_line_start + self.line_pos + 1]
                    else:
                        self.tail = self.contents[prev_line_start + len(prev_line):] + self.tail
                        self.contents = self.contents[:prev_line_start + len(prev_line)]

                return ''

                """

                    DOWN

                    ##
                    ##
                  #    #
                   #  #
                    ##

                """
            elif char == "down":
                if "\n" in self.tail:
                    next_line_start = self.tail.find('\n')
                    if next_line_start < 0:
                        return ''
                    next_line_end = self.tail[next_line_start+1:].find('\n')
                    if next_line_end < 0:
                        next_line_end = len(self.tail[next_line_start+1:])

                    next_line = self.tail[next_line_start:next_line_start + next_line_end + 1]

                    if len(next_line) > self.line_pos:
                        self.contents += self.tail[:next_line_start + self.line_pos]
                        self.tail = self.tail[next_line_start + self.line_pos:]
                    else:
                        self.contents += self.tail[:next_line_start + len(next_line)]
                        self.tail = self.tail[next_line_start + len(next_line):]

                return ''

                """

                    LEFT - or - BACKSPACE


                       ##
                      ##
                     ##  ####
                      ##
                       ##

                """
            elif char in ("left", "backspace") and len(self.contents) > 0:

                last = self.contents[-1]
                self.contents = self.contents[:-1]

                if last == '\n':
                    self.line_pos = len(self.contents[:self.contents.rfind('\n')])

                elif self.line_pos > 0:
                    self.line_pos -= 1
                if char == 'backspace':
                    pass
                else:
                    self.tail = last + self.tail
                return ''

                """


                    RIGHT - or - DELETE

                         ##
                          ##
                     ####  ##
                          ##
                         ##

                """
            elif char in ("right", "delete") and len(self.tail) > 0:
                first = self.tail[0]
                self.tail = self.tail[1:]
                if first == '\n':
                    self.line_pos = 0
                elif len(self.tail) > 0:
                    self.line_pos += 1
                if char == 'delete':
                    pass
                else:
                    self.contents += first
                return ''

        rv = name2char.get(char, '')

        return rv

    """

        ###################################

            HANDLE THE INPUT CHARACTERS

        ###################################

    """

    @funlog
    def handle(self, char):
        """Deal with the provided input."""
        if self.one_line and char == 'return' and self.editMode:
            em.toggle_text_edit(self)
            return

        if len(char) > 1:
            char = self.handle_named_char(char)
            if char is None:
                return
        elif char in remaps.keys():
            char = remaps[char]

        if not self.hasContent:
            self.hasContent = True
            self.contents = ""

        # add the new character to the content list
        if char == "\n":
            self.line_pos = 0
        elif char != '':
            self.line_pos += 1
        self.contents += char
        self.draw()

    @funlog
    def get_tooltip_updater(self):
        """
        expose the tooltip get_updater method.
        """
        pass

    # #  @funlog
    def get_tooltip_vis_secs(self):
        """
        return the tooltip's visibility time in seconds
        """
        return 0.0
        return self.tooltip.get_vis_secs()

    @funlog
    def set_tooltip_vis_secs(self, secs):
        """
        Set the tooltip's visibility time in seconds
        """
        return
        self.tooltip.set_vis_secs(secs)

    @funlog
    def set_tooltip_updater(self, func):
        """
        expose the tooltip set_updater method.
        """
        return
        self.tooltip.set_updater(func)

    @funlog
    def setSelected(self):  # noqa: N802
        """Set this object as selected."""
        self.selected = True
        self.renderSelected()

    @funlog
    def renderSelected(self):
        """
        Display this object as if it were selected
        """
        self._oldScale = self.data.getColorScale()
        self.data.setColorScale(255, 255, 255, 128)

    @funlog
    def setUnselected(self):  # noqa: N802
        """Set this object as de-selected."""
        self.selected = False
        self.renderUnselected()

    @funlog
    def renderUnselected(self):
        """
        Display this object as if it were not selected
        """
        self.data.setColorScale(self._oldScale)


class Text3d(model.Model):
    """The class which manages the on-screen text."""

    def __init__(
        self,
        uid,
        app,
        parent=None,
        default_contents="Text",
        initscale=1.0,
        anchor_h=Anchor.LEFT,
        anchor_v=Anchor.TOP,
        align_h=HAlign.LEFT,
    ):
        """Initialize the text instance."""
        # super().__init__(uid, 'T', app)
        CallbackManager.__init__(self)
        if parent is None:
            parent = cm.cwc()
        self.text_np = parent.attachNewNode(uid)
        self.uid = uid
        self.name = uid
        self.app = app
        self.hasContent = False
        self.editMode = False
        self.contents = list(default_contents)
        self.rightContents = list()
        self.tail = list()
        self.lines = list()
        self.currentRP = RowProps(0)
        self.cursor = Cursor(self.text_np, initscale)
        self.glyphs = []
        self.keycnt = 0
        self.scale = Scale()
        self.scale.set(initscale)
        self.anchor_h = anchor_h
        self.anchor_v = anchor_v
        self.align_h = HAlign

        # So that the text is always looking at the viewer
        self.text_np.setHpr(camera.getHpr())  # noqa: F405, F821
        self.draw()
        self.addEvent("on_change")

    @funlog
    def getTightBounds(self):  # noqa: N802
        """Return the bounding box."""
        return self.text_np.getTightBounds()

    @funlog
    def getString(self):  # noqa: N802
        """Return the text string managed by this instance."""
        rv = ""
        for i in self.contents:
            if len(i) == 1:
                rv += i
            else:
                rv += name2char[i]
        for i in self.rightContents:
            if len(i) == 1:
                rv += i
            else:
                rv += name2char[i]

        return rv

    @funlog
    def clear(self):
        """Remove contents of this text area."""
        for g in self.glyphs:
            g.remove()
        del self.glyphs[:]
        del self.lines[:]
        self.currentRP = RowProps(0)

    @funlog
    def centerHorizontally(self):
        """Center the text horizontally, but not vertically."""
        self.center(cent_horiz=True, cent_vert=False)

    @funlog
    def center(self, cent_horiz, cent_vert):
        """
        Center the text according to flags cent_horiz and cent_vert.

        calculate the size of each row of text
        and re-position the glyphs on both sides of zero X
        """
        rows = []
        rsize = 0
        rwidth = 0
        work = self.glyphs + self.tail
        for g in work:
            if g.char == " ":
                rsize += 1
                rwidth += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                continue
            if g.char == "\n":
                rows.append((rsize, rwidth))
                rsize = 0
                rwidth = 0
                continue

            rsize += 1
            rwidth += g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)

        rows.append((rsize, rwidth))

        ridx = 0
        rsize, rwidth = rows[ridx]
        x = z = 0

        def mod_h(x, w):
            if cent_horiz:
                rv = x - (w / 2)
            else:
                rv = x
            return rv

        def mod_v(z, h):
            if cent_vert:
                rv = z + (h / 2)
            else:
                rv = z
            return rv

        for g in work:

            if g.char == " ":
                x += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                continue
            if g.char == "\n":
                rsize, rwidth = rows[ridx]
                ridx += 1
                z -= self.scale.of(DEFAULTTEXTHEIGHT)
                x = 0
                continue

            g.setPos(mod_h(x, rwidth), 0, mod_v(z, len(rows) + 1))
            x += g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)

    @funlog
    def draw(self):
        """
        Method to trigger the rendering of the glyphs.

        Render the contents as glyphs
        parented by self.text_np
        """
        self.clear()
        for c in self.contents:
            self.handle(c)

        return
        i = 0
        z = 0
        for c in self.contents:
            if c == " ":
                i += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                continue
            if c == "\n":
                z -= 1
                i = 0
                continue
            g = Glyph(c, self.keycnt, self, self.app)
            g.setScale(self.scale.get())
            g.setTextParent(self)
            self.glyphs.append(g)
            g.reparentTo(self.text_np)
            if not self.centerHorizontal:
                g.setPos(i, 0, z)
                i += g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)
            self.keycnt += 1

    @funlog
    def setPos(self, x, y, z):  # noqa: N802
        """Expose the internal set position method."""
        self.text_np.setPos(x, y, z)

    @funlog
    def setScale(self, scale):  # noqa: N802
        """Expose internal set-scale methods."""
        self.scale.set(scale)
        self.cursor.setScale(scale)

    @funlog
    def toggleEdit(self):  # noqa: N802
        """Toggle the editmode of this text area."""
        if self.editMode:
            if len(self.contents) < 1:
                self.contents = list("Text")
                self.draw()
            self.editMode = False
            self.cursor.hide()
            self.triggerEvent("on_change", self.getString())
        else:
            self.text_np.setHpr(camera.getHpr())  # noqa: F821
            self.editMode = True
            self.cursor.show()

        if not self.hasContent:
            self.contents = list()
            self.draw()

    @funlog
    def handle_clicked(self, *args, **kwargs):
        """Handle a click on any of the visible text."""
        # em.post_pending('hide_model_menu', self.menu)
        # em.post_pending('model_toggle_selected', self)
        import nord.sigurd.shell.application as APP

        APP.set_keyboard_text_target(self)
        em.toggle_text_edit(self)

    @funlog
    def handle_named_char(self, char):
        """Given the named character input, handle it as needed."""
        x, y, z = self.cursor.getPos()
        """
            We'll need to translate this character
        """
        if char in ("up", "down", "left", "right", "backspace", "delete"):

            """

                    UP

                    ##
                   #  #
                  #    #
                    ##
                    ##
            """
            if char == "up":
                warn("Text input does not yet support 'up' input")
                # z+=1
                # self.cursor.setPos(x, y, z)

                """

                    DOWN

                    ##
                    ##
                  #    #
                   #  #
                    ##

                """
            elif char == "down":
                warn("Text input does not yet support 'down' input")
                # z-=1
                # self.cursor.setPos(x, y, z)

                """

                    LEFT - or - BACKSPACE


                       ##
                      ##
                     ##  ####
                      ##
                       ##

                """
            elif char in ("left", "backspace") and len(self.glyphs) > 0:
                g = self.glyphs.pop()
                nc = self.contents.pop()

                # handle backing over a newline
                if g.char == "\n":

                    z += self.scale.of(DEFAULTTEXTHEIGHT)

                    # recompute the X location from the prior newline or beginning of list
                    rp = self.lines[self.currentRP.rownum - 1]
                    x = self.scale.of(rp.width())

                    priorll = 0
                    idx = len(self.glyphs) - 1
                    while idx > 0 and not self.glyphs[idx].char == "\n":
                        if self.glyphs[idx].char == " ":
                            priorll += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                        else:
                            priorll += (self.glyphs[idx].getWidth()) + self.scale.of(
                                DEFAULTTEXTHSPACING
                            )
                        idx -= 1

                    x = priorll

                elif g.char == " ":
                    x -= self.scale.of(DEFAULTTEXTSPACEWIDTH)
                else:
                    x -= g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)

                #
                if char == "left":
                    self.tail.insert(0, g)
                    self.rightContents.insert(0, nc)
                else:
                    if g.char not in (" ", "\n"):
                        cwidth = (
                            g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)
                        ) * -1.0
                    elif g.char == "\n":
                        cwidth = priorll
                    elif g.char == " ":
                        cwidth = self.scale.of(DEFAULTTEXTSPACEWIDTH) * -1.0
                    # move any things in tail
                    # up to the first newline
                    z2 = z
                    for i in self.tail:
                        if i.char == " ":
                            continue
                        if i.char == "\n":
                            z2 -= self.scale.of(DEFAULTTEXTHEIGHT)
                            cwidth = 0
                            continue
                        x1, y1, z1 = i.getPos()
                        i.setPos(x1 + cwidth, y1, z2)

                    if g.char not in (" ", "\n"):
                        g.remove()
                    else:
                        del g

                self.cursor.setPos(x, y, z)

                """


                    RIGHT - or - DELETE

                         ##
                          ##
                     ####  ##
                          ##
                         ##

                """
            elif char in ("right", "delete") and len(self.tail) > 0:
                g = self.tail.pop(0)
                nc = self.rightContents.pop(0)

                if g.char == "\n" and not char == "delete":
                    z -= self.scale.of(DEFAULTTEXTHEIGHT)
                    x = 0
                elif g.char == "\n" and char == "delete":

                    # recompute the X location from the prior newline or beginning of list
                    priorll = 0
                    idx = len(self.glyphs) - 1
                    while idx >= 0 and not self.glyphs[idx].char == "\n":
                        if self.glyphs[idx].char == " ":
                            priorll += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                        else:
                            priorll += (self.glyphs[idx].getWidth()) + self.scale.of(
                                DEFAULTTEXTHSPACING
                            )
                        idx -= 1

                    cwidth = -1.0 * priorll

                #
                elif g.char == " " and char == "delete":
                    pass
                elif g.char == " ":
                    x += self.scale.of(DEFAULTTEXTSPACEWIDTH)
                elif not char == "delete":
                    x += g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)

                if char == "right":
                    self.glyphs.append(g)
                    self.contents.append(nc)
                else:

                    if g.char not in (" ", "\n"):
                        cwidth = g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)
                    elif g.char == " ":
                        cwidth = self.scale.of(DEFAULTTEXTSPACEWIDTH)

                    # move any things in tail
                    # up to the first newline
                    z2 = z
                    for i in self.tail:
                        if i.char == " ":
                            continue
                        elif i.char == "\n":
                            cwidth = 0
                            z2 -= self.scale.of(DEFAULTTEXTHEIGHT)
                            continue

                        x1, y1, z1 = i.getPos()
                        i.setPos(x1 - cwidth, y1, z2)

                    if g.char not in (" ", "\n"):
                        g.remove()
                    else:
                        del g

                self.cursor.setPos(x, y, z)

            return None

        return mapping[char]

    """

        ###################################

            HANDLE THE INPUT CHARACTERS

        ###################################

    """

    @funlog
    def handle(self, char):
        """Deal with the provided input."""
        x, y, z = self.cursor.getPos()

        if len(char) > 1:
            char = self.handle_named_char(char)
            if char is None:
                return
        elif char in remaps.keys():
            char = remaps[char]

        if not self.hasContent:
            self.hasContent = True
            self.contents = []

        # add the new character to the content list
        i = len(self.contents)
        self.contents.append(char)
        if char == " ":
            self.currentRP.add_space()
            cwidth = self.scale.of(DEFAULTTEXTSPACEWIDTH)
            x += cwidth
            g = Glyph(char, self.keycnt, self, self.app, visible=False)
        elif char == "\n":
            rp = RowProps(self.currentRP.rownum + 1, i)
            self.lines.append(rp)
            self.currentRP = rp
            g = Glyph(char, self.keycnt, self, self.app, visible=False)
            cwidth = x * -1.0
            x = 0
            z -= self.scale.of(DEFAULTTEXTHEIGHT)
        else:
            # create a new glyph
            g = Glyph(char, self.keycnt, self, self.app)
            self.currentRP.add_char(g)
            g.setScale(self.scale.get())
            g.setTextParent(self)

            g.reparentTo(self.text_np)
            g.setPos(x, y, z)
            cwidth = g.getWidth() + self.scale.of(DEFAULTTEXTHSPACING)
            x += cwidth

        self.glyphs.append(g)

        # move any things in tail
        # up to the first newline
        z2 = z
        for i in self.tail:
            if i.char == " ":
                continue
            if i.char == "\n":
                cwidth = 0
                z2 -= self.scale.of(DEFAULTTEXTHEIGHT)
                continue
            x1, y1, z1 = i.getPos()
            i.setPos(x1 + cwidth, y1, z2)

        self.cursor.setPos(x, y, z)
        self.keycnt += 1
