"""
Copyright 2020 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
"""
import pytest
import nord.sigurd.managers.objectmanager as om
import nord.sigurd.shell.vismap as vm
import nord.nord.Runtime as RT
from panda3d.core import Point3
from pathlib import Path
import json
import os


Runtime = RT.Runtime


@pytest.fixture(scope="function")
def cleanup_testgraph():
    testfname = f"{os.getpid()}.n"
    yield testfname
    os.unlink(testfname)


@pytest.mark.panda3d()
def test_application_save(init_panda3d, cleanup_testgraph):
    """
    Starting with an empty mapfile:
    add a data node, and instance of the data node
    and a static assigning to each of them.

    save the map, then load it into a dict and ensure
    the expected nodes and edges are in place.
    """
    app = init_panda3d
    testfname = cleanup_testgraph
    app.load_map(testfname)

    karte = vm.get_working_map()
    # vessal = Runtime()

    vm.create_node("Data", Point3(-1.0, 0.0, 0.5), space="nord")
    data = karte.nodes[list(karte.nodes.keys())[0]]
    vm.create_node("Static", Point3(1.0, 0.0, -0.5), space="nord")
    static = karte.nodes[list(karte.nodes.keys())[1]]
    vm.create_edge(static, data, 'assign', None, None)
    vm.create_instance(om.fetch(data.id), Point3(-1.0, 0.0, 0.0))
    inst = karte.nodes[list(karte.nodes.keys())[2]]
    vm.create_edge(static, inst, 'assign', None, None)

    karte.ship_changes()

    vm.save()

    m = json.load(open(testfname, 'r'))

    datafound = False
    staticfound = False
    instfound = False
    for item in m["map"]["nodes"]:
        if item["id"] == data.id:
            datafound = True
            assert "instanceof" not in item
        if item["id"] == static.id:
            staticfound = True
            assert "instanceof" not in item
        if item["id"] == inst.id:
            instfound = True
            assert "instanceof" in item
            assert item["instanceof"] == data.id

    assert datafound
    assert staticfound
    assert instfound

    assert len(m["map"]["edges"]) == 2

    todatafound = False
    toinstfound = False

    for item in m["map"]["edges"]:
        if item["source"].split(":")[0] == static.id and item["target"].split(":")[0] == data.id:
            todatafound = True
        if item["source"].split(":")[0] == static.id and item["target"].split(":")[0] == inst.id:
            toinstfound = True

    assert todatafound
    assert toinstfound

    # while True:
    #     app.one_loop_iteration()


@pytest.mark.panda3d()
def test_application_load(init_panda3d):
    """
    Starting with an empty mapfile:
    add a data node, and instance of the data node
    and a static assigning to each of them.

    save the map, then load it into a dict and ensure
    the expected nodes and edges are in place.
    """
    app = init_panda3d
    if Path('tests/maps/test_load.n').exists():
        testfname = 'tests/maps/test_load.n'
    else:
        testfname = 'sigurd/tests/maps/test_load.n'

    app.load_map(testfname)

    karte = vm.get_working_map()
    # vessal = Runtime()

    data = karte.get_node("788566d7-8feb-4911-83c6-78e2e1af74c2")
    inst = karte.get_node("948f87a3-d05d-49b7-baf3-6dc93e2b6edb")
    static = karte.get_node("04d1756e-e1d3-4f6f-8f63-2dbb85ec19be")

    todata = karte.get_edge(static.id, data.id, "assign")
    toinst = karte.get_edge(static.id, inst.id, "assign")

    assert om.fetch(data.id).mapentry == data
    assert om.fetch(inst.id).mapentry == inst
    assert om.fetch(static.id).mapentry == static
    assert om.fetch(f"{static.id}:{data.id}:assign").connection == todata.visualization
    assert om.fetch(f"{static.id}:{inst.id}:assign").connection == toinst.visualization

    # while True:
    #     app.one_loop_iteration()
