"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Provide node class which provides variable binary data.
"""

import nord.sushrut.nodes._data_complex as Data
import nord.nord.nodes.Data as _basedata
from nord.nord.exceptions import RuntimeException
import sys
import json


class Dataarray(Data):
    """Facilitate variable arrays of data."""

    _type = list

    def __init__(self, graph):
        super().__init__(graph)
        self._sushrut_da_todict_rcrsv_trk = None
        self._current_idx = 0

    def prep_to_execute(self, runtime, verbose):
        # Ensure none of the array contents are scheduled
        # because they are contained by this array.
        for edge in self.get_targets("contains"):
            edge.flag_followed(runtime)

        super().prep_to_execute(runtime, verbose)

    def _load_contents_into_list(self, list_to_load):
        list_to_load.clear()
        tmp = list()
        for edge in self.get_targets("contains"):
            tmp.append(edge.get_target())
        list_to_load.extend(sorted(tmp))

    def append_to_array(self, rule):
        """This is called by the edge.apply_delegate via object interrogation."""
        if not hasattr(self, '_list'):
            self._list = list()
            self._load_contents_into_list(self._list)
        val = rule.edge.get_cargo()
        k = len(self._list)
        self._list.append(val)

        # Now, we need to modify the graph such that the
        # passed value, is now contained within this Node
        self._contain_value_element(f"idx_{k}", val)

    def set_current_idx(self, rule):
        nidx = rule.edge.get_cargo()
        if not hasattr(self, '_list'):
            self._list = list()
            self._load_contents_into_list(self._list)
        if type(nidx) == int and nidx < len(self._list):
            self._current_idx = nidx
        else:
            raise RuntimeException(_("Invalid array or array location to set"))  # noqa: F821

    def get_current_itm(self, rule):
        if hasattr(self, '_list') and self._current_idx < len(self._list):
            itm = self._list[self._current_idx]
            if isinstance(itm, _basedata) or (
                    type(itm).__module__.startswith('nord.') and
                    hasattr(itm, 'get')):
                val = itm.get()
            else:
                val = itm
            rule.edge.set_cargo(val)
        else:
            raise RuntimeException(_("Invalid array or array location to fetch from"))  # noqa: F821

    def get_size(self, rule):
        if hasattr(self, '_list'):
            rule.edge.set_cargo(len(self._list))
        else:
            rule.edge.set_cargo(-1)

    def to_dict(self, include_contents=False, recurse_tracker=None):
        """
        Extend the builtin (Node) to_dict functionality, which
        has the primary use case of serializing nodes into the
        format needed for map serialization, with the ability to
        translate this instance for communicating with the front-end (FE)

        (The latter part is TBD)
        """

        rv = super().to_dict(self)
        if include_contents:
            if recurse_tracker is None:
                self._sushrut_da_todict_rcrsv_trk = {}
            else:
                self._sushrut_da_todict_rcrsv_trk = recurse_tracker

            rv["contents"] = {}
            for edge in self.get_targets("contains"):
                tgt = edge.get_target()
                # if tgt not in self._sushrut_da_todict_rcrsv_trk:
                #     self._sushrut_da_todict_rcrsv_trk[tgt] = tgt.id
                if hasattr(tgt, '_sushrut_da_todict_rcrsv_trk'):
                    do_contents = tgt not in self._sushrut_da_todict_rcrsv_trk
                    val = tgt.to_dict(include_contents=do_contents, recurse_tracker=self._sushrut_da_todict_rcrsv_trk)
                    if not do_contents:
                        val["contents"] = {}
                else:
                    val = tgt.to_dict(include_contents=True)
                if tgt not in self._sushrut_da_todict_rcrsv_trk:
                    self._sushrut_da_todict_rcrsv_trk[tgt] = tgt.id
                # val["name"] = tgt._sushrut_dataobject_keyname
                if not hasattr(tgt, '_sushrut_dataobject_keyname') and hasattr(tgt, 'name'):
                    tgt._sushrut_dataobject_keyname = tgt.name
                rv["contents"][tgt._sushrut_dataobject_keyname] = val

            if recurse_tracker is None:
                # As above, this will leave the recursion tracker
                # referenced in the descendent data objects
                del self._sushrut_da_todict_rcrsv_trk
                self._sushrut_da_todict_rcrsv_trk = None

        return rv

    def parse_value(self):
        """
            Data array values are stored as lists.
        """
        if not hasattr(self, 'value'):
            return
        self.parse_from_list()

    def json_serialize(self):
        """
        Translate this instance for communicating with the front-end (FE)
        """
        return json.dumps(self.to_dict(include_contents=True))

    def get_index(self, index):
        """
        if the requested value exists in the contained
        data, return it (the wrapped object, not the raw value)
        """
        rv = None
        contents = self.get_targets('contains')
        if len(contents) > index:
            rv = contents[index].get_target()
        return rv

    def set(self, value, recurse_tracker=None, dont_parse=False):
        """
        Clear out contents and value prior to assigning a new value

        Then call _data_complex's set method to handle parsing
        and repopulating. (This may not be entirely appropriate. Some
        introspection of value may be needed in the future.)
        """
        if self.get_targets("contains"):
            self.rem_all_targets("contains")
        self.clear_dc_trackers()
        # if hasattr(self, "value"):
        #     del self.value
        #     self.value = []
        super().set(value, recurse_tracker, dont_parse)


sys.modules[__name__] = Dataarray
