"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Provide node class which provides a variable which handles functions.
"""
import nord.sushrut.nodes._data as Data
import nord.nord.Receiver as Receiver
import nord.sushrut.nodes.Dataobject as Dataobject
from nord.nord.exceptions import RuntimeException as RuntimeException
import sys


class Datafunction(Data, Receiver):
    """Facilitate variable for functions."""

    _type = "function"

    def __init__(self, graph):
        super().__init__(graph)

    def is_external(self):
        """
        Indicate if this function has a Nord implemtation (False)
        or, if it has a Python (or other) implementation (True)
        """
        if len(self.get_targets("contains")) == 0\
                and hasattr(self, 'value') or hasattr(self, '_function_from_payload'):
            return True
        else:
            return False

    def is_native(self):
        """
        Indicate if this function has a Nord implemtation (True)
        or, if it has a Python (or other) implementation (False)
        """
        if not hasattr(self, 'value') and not hasattr(self, '_function_from_payload'):
            return True
        else:
            return False

    def call_value_as_function(self):
        """
        Facilitate the initial determination whether
        the payload is still the function. If it is, relocate it,
        then execute it, and set the value as the result of the
        function.

        Should this item have it's value set, then, reset the internal state
        such that this method will relocate the value again.
        """
        if not hasattr(self, '_function_from_payload') or \
                (self._function_from_payload is None and
                 self.value is not None):
            self._function_from_payload = self.value

        if self._function_from_payload is None:
            if self.get_unfollowed_sources():
                return False
            else:
                raise RuntimeException(_("No function can be found for: {}").format(self))  # noqa: F821

        """
        We need to figure out the arguments to the function,
        organize and pass them as needed... But, thankfully,
        that is implemented by the Receiver class...
        """
        args, kwargs = self.get_arg_values()
        self._sushrut_datafunction_returnvalue = self._function_from_payload(*args, **kwargs)
        return True

    def set(self, value):
        super().set(value)
        if hasattr(self, '_function_from_payload'):
            delattr(self, '_function_from_payload')
        if hasattr(self, '_sushrut_datafunction_returnvalue'):
            delattr(self, '_sushrut_datafunction_returnvalue')

    def change_attribute(self, attrname, value):
        """
        This override's Node's method which is called
        by the CDC system (which can be operating over the network)
        If this is passed a dict,, then this is a member of a dataobject
        if this has _run_state, then it is in a running map.
        so, we need to ask the dataobject for our implementation
        """
        if type(value) is dict and self.get_op_runtime() is not None:
            got_implementation = False
            for edge in self.get_sources("contains"):
                if type(edge.get_source()) == Dataobject:
                    datao = edge.get_source()
                    val = datao.get_implementation(value["value"])
                    if val is not None:
                        self.set_untracked_attr("value", val)
                        got_implementation = True
                        break

            if not got_implementation:
                raise RuntimeException(f"{_('Could not find implementation for:')} {self}")  # noqa: F821
        else:
            setattr(self, attrname, value)

    def prep_to_execute(self, runtime, verbose):
        """
        Pre-empt and clear preempt if we are expected
        to run as a function.
        """

        # if not hasattr(self, "value") and not self.operate_as_variable:
        if not hasattr(self, "value") and (len(self.get_sources('pass')) or
                                           len(self.get_sources('flow'))):
            self.flag_preempted()
            return
        # elif hasattr(self, 'value') and not self.operate_as_variable:
        elif hasattr(self, 'value') and (len(self.get_sources('pass')) or
                                         not len(self.get_sources('flow'))):
            self.clear_preempted()

        super().prep_to_execute(runtime, verbose)

    def execute(self, runtime):
        """
        Override the Data execute method
        such that an exception is raised
        when operate_as_variable is False and no
        arguments are available, yet they are required???

        Really, this should be where the function actually executes.

        Not in the Rules.

        So, call_value_as_function should not be called
        in DatafunctionAssign and DatafunctionPass, but, in
        here, the execute method. So, the question becomes:

        Should those classes do anything special?
        """
        if not len(self.get_sources('pass')) and not \
                len(self.get_sources('flow')):
            pass
        else:
            if not self.call_value_as_function():
                self.flag_preempted()
                return
        self.flag_exe(runtime)

    def json_serialize(self):
        rv = {"type": self.__class__.__name__}
        if hasattr(self, "value") and self.value is not None:
            rv["value"] = self.value.__name__
        elif hasattr(self, '_function_from_payload'):
            rv["value"] = self._function_from_payload.__name__
        return rv


sys.modules[__name__] = Datafunction
