"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Extend the Data node type with "object" behaviors.
"""

import nord.sushrut.nodes._data_complex as Data
import json
import sys


class Dataobject(Data):
    """Facilitate complex object data."""

    _type = object

    def __init__(self, graph):
        super().__init__(graph)
        self._sushrut_do_todict_rcrsv_trk = None

    def to_python(self):
        """
        Method to convert this entity into a python compatible
        structure. (For Shoshoni to work...)
        """
        # if self.value is not None:
        # self.value = DefinableObject(self)
        return self.value

    def get_storage_content(self):
        """
        build and return a dictionary (or other JSON serializable) representation
        of the contents of this object. Graph's (sub-graph's and methods) are to be sent
        in their JSON form.
        """
        return self.to_dict()

    def apply_storage_content(self, content):
        pass

    def to_dict(self, include_contents=True, recurse_tracker=None):
        """
        Extend the builtin (Node) to_dict functionality, which
        has the primary use case of serializing nodes into the
        format needed for map serialization, with the ability to
        translate this instance for communicating with the front-end (FE)

        (The latter part is TBD)
        """

        rv = super().to_dict(self)
        if include_contents:
            if recurse_tracker is None:
                self._sushrut_do_todict_rcrsv_trk = {}
            else:
                self._sushrut_do_todict_rcrsv_trk = recurse_tracker

            rv["contents"] = {}
            for edge in self.get_targets("contains"):
                tgt = edge.get_target()
                do_contents = None
                if tgt not in self._sushrut_do_todict_rcrsv_trk:
                    self._sushrut_do_todict_rcrsv_trk[tgt] = tgt.id
                    do_contents = True
                if hasattr(tgt, '_sushrut_do_todict_rcrsv_trk'):
                    if do_contents is None:
                        do_contents = tgt not in self._sushrut_do_todict_rcrsv_trk
                    val = tgt.to_dict(include_contents=do_contents, recurse_tracker=self._sushrut_do_todict_rcrsv_trk)
                    if not do_contents:
                        val["contents"] = {}
                else:
                    val = tgt.to_dict(include_contents=True)
                if tgt not in self._sushrut_do_todict_rcrsv_trk:
                    self._sushrut_do_todict_rcrsv_trk[tgt] = tgt.id
                # val["name"] = tgt._sushrut_dataobject_keyname
                if not hasattr(tgt, '_sushrut_dataobject_keyname') and hasattr(tgt, 'name'):
                    tgt._sushrut_dataobject_keyname = tgt.name
                rv["contents"][tgt._sushrut_dataobject_keyname] = val

            if recurse_tracker is None:
                # As above, this will leave the recursion tracker
                # referenced in the descendent data objects
                self._sushrut_do_todict_rcrsv_trk.clear()
                del self._sushrut_do_todict_rcrsv_trk
                self._sushrut_do_todict_rcrsv_trk = None

        return rv

    def parse_value(self):
        """
            Data object values are stored as dicts.
            The value's within the dict are linked to
            a submap (contained within the dataobject)
            which aligns sushrut types to the contents
            of the dict.
        """
        if not hasattr(self, 'value'):
            return
        if type(self.value) is dict:
            self.parse_from_dict()
        else:
            self.parse_from_object()

    def json_serialize(self):
        """
        Translate this instance for communicating with the front-end (FE)
        """
        return json.dumps(self.to_dict(include_contents=True))

    def execute(self, runtime):
        """
        Override the data execute as the Dataobject itself
        does nothing.
        """
        self.flag_exe(runtime)


sys.modules[__name__] = Dataobject
