"""
Copyright 2020 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

import nord.sushrut.nodes._data as _Data
import nord.nord.edges.Contains as Contains
from nord.nord.exceptions import RuntimeException as RuntimeException
import sys


class DataComplex(_Data):
    def __init__(self, graph):
        super().__init__(graph)
        self.__dict__["type"] = self.__class__.__name__.lower()
        self.clear_dc_trackers()

    def clear_dc_trackers(self):
        self._sushrut_dc_init_rcrsv_trk = {}
        self._sushrut_dc_members = {}

    def _contain_value_element(self, k, v):
        edge = None
        wrk_node = None
        recurse_detected = False

        if type(v) not in (int, str, float, dict, list, type(None)) and\
                v in self._sushrut_dc_init_rcrsv_trk:
            # We've already digested this value once,
            # so, just reference that and stop recursing
            wrk_node = self._sushrut_dc_init_rcrsv_trk[v]
            recurse_detected = True

        if k in self._sushrut_dc_members and wrk_node is None:
            old_v = self._sushrut_dc_members[k].get()
            if type(old_v) not in (int, str, float, dict, list, type(None)) and\
                    old_v in self._sushrut_dc_init_rcrsv_trk and v != old_v:
                del self._sushrut_dc_init_rcrsv_trk[old_v]
            wrk_node = self._sushrut_dc_members[k]

        if wrk_node is None:
            for e in self.get_targets("contains"):
                n = e.get_target()
                if ((hasattr(n, '_sushrut_dataobject_keyname') and
                        n._sushrut_dataobject_keyname == k) or
                        (hasattr(n, 'name') and k == n.name)):
                    wrk_node = n
                    break

        if wrk_node is None:
            # We could not find an already contained representation of this
            # member, so, we'll create one.
            from nord.sushrut.utils import py_to_sushrut
            wrk_node = py_to_sushrut(v, self.get_graph())
            wrk_node.set_untracked_attr('_sushrut_dataobject_keyname', k)
            wrk_node.set_untracked_attr('name', k)

        if edge is None:
            if self.targets(wrk_node, "contains"):
                edge = self.get_target_edges(wrk_node, "contains")[0]
            else:
                edge = Contains(self.get_graph())
                edge.set_nodes(self, wrk_node, "contains")
                self.set_target(wrk_node, "contains", edge)

        # If this has been called during execution
        # it is this node's responsibility to "execute"
        # all the added nodes, and follow the edges
        if hasattr(self, '_current_runtime'):
            wrk_node.flag_exe(self._current_runtime)
            edge.flag_followed(self._current_runtime)

        # only track for recursion on types that are fully "digested"
        if type(v) not in (int, str, float, dict, list, type(None)):
            self._sushrut_dc_init_rcrsv_trk[v] = wrk_node
        if hasattr(wrk_node, '_sushrut_dc_init_rcrsv_trk'):
            # this means the work node is also a Dataobject
            # and, it needs to not recursively parse any objects
            # already parsed by this, or parent Dataobjects
            wrk_node.set(v, recurse_tracker=self._sushrut_dc_init_rcrsv_trk, dont_parse=recurse_detected)
        else:
            wrk_node.set(v)

        self._sushrut_dc_members[k] = wrk_node

    def parse_from_list(self):
        """
        parses the list of DataArray elements
        and contains the values.
        """
        for k, v in enumerate(self.value, start=0):
            # Ignore hidden stuff, this should be made runtime configurable
            # so user's can elect to see all the internals ... someday...
            self._contain_value_element(f"idx_{k}", v)

    def parse_from_object(self):
        """
        When the data object value is a complex
        python object, this function will interrogate
        that object and develop the submap to represent
        the contents of said complex object.
        """
        for k in dir(self.value):
            # Ignore hidden stuff, this should be made runtime configurable
            # so user's can elect to see all the internals ... someday...
            if k.startswith('_'):
                continue
            v = getattr(self.value, k)
            self._contain_value_element(k, v)

    def parse_from_dict(self):
        """
            When data object values are stored as dicts.
            The value's within the dict are linked to
            a submap (contained within the dataobject)
            which aligns sushrut types to the contents
            of the dict.
        """
        for k, v in self.value.items():
            # first check to see if we already have a Node pointing to this item
            # If already contained, just set the value, if not
            # then create a new Node and contain it.
            self._contain_value_element(k, v)

        for edge in self.get_targets("contains"):
            # Since the value could be represented as an object or a dict, the following
            # complex logic is required.... this could probably use a refactoring
            #  of how dicts and objects are being lumped together for introspection
            #  purposes
            if not hasattr(edge.get_target(), '_sushrut_dataobject_keyname') or \
                    (edge.get_target()._sushrut_dataobject_keyname not in dir(self.value)
                        and type(self.value) is not dict) or (type(self.value) is dict and
                                                              edge.get_target()._sushrut_dataobject_keyname
                                                              not in self.value):
                self.rem_target(edge.get_target(), edge)

        # Now, we need to remove any items which were not in the
        # input data.
        for edge in self.get_targets("contains"):
            if not hasattr(edge.get_target(), '_sushrut_dataobject_keyname') or \
                    edge.get_target()._sushrut_dataobject_keyname not in self.value:
                self.rem_target(edge.get_target(), edge)

    def parse_value(self):
        raise RuntimeException(f"Complex Data parse_value not implemented.... on {self.__class__.__name__}")

    def get_implementation(self, attrname):
        """
        Get the requested attribute from
        the underlying python object.
        """
        if hasattr(self.value, attrname):
            return getattr(self.value, attrname)
        else:
            return None

    def get(self):
        """
        Assemble all the child contents into a python object.

        This seems like a really bad idea. What happens if the members are changed in graph land.
        How do we ensure that the corresponding graph objects are also changed?

        Should there be getattr/setattr wrappers put around a simulated object, such that any
        python code that interacts with the instance, is actually changing contents held
        within the graph?

        So, the returned object would recursively go through all contained elements, and
        wrap their get/set methods inside the returned object's getattr/setattr calls.

        If we did that, it might look like this code below... (now commented out....)
        """

        # self.value = DefinableObject(self)
        return self.value

    def set(self, value, recurse_tracker=None, dont_parse=False):
        """
        Align the passed object's fields with what is contained in this dataobject.

        Then assign the aligned values. If the passed object has un-aligned
        attributes, raise an exception.


        SEEMS THIS IS NEXT ON THE LIST....

        Need to sort out the re-aligning vs just fully reset the object

        A full, from scratch re-interrogate seems to make more sense for arrays

        The root of the issue seems to be converging content that was loaded
        from a saved map, with the content from a fresh execution. It could be that
        this issue is really just an integration / shoshoni (or should be) issue

        This needs some thought about the design and implementation.

        Is this as simple as detecting if running? That is to say:
        If not running attempt to merge? If running, replace ...

        And the answer to that is No. This issue manifests when running for
        the first time...? Is that the optimization? Merge on first run... seems like the
        opposite of an optimization.

        All this parsing stuff, probably belongs in shoshoni...

        The data types should provide the basic methods, but, the interrogation
        and parsing of python objects and probably dicts as well does not
        really belong here.

        That could be a problem for another day as DataComplex could be
        refactored into shoshoni...

        """
        if self.get_targets("contains"):
            # This is a pre-populated object
            # So, we need to link up the member names
            # to their node's such that we don't
            # recreate the nodes, thus generating
            # new ID's and breaking the connection
            # with the visible nodes
            for e in self.get_targets("contains"):
                tgt = e.target
                self._sushrut_dc_members[tgt.name] = tgt
        # Bypass CDC for the value, as the nodes themselves will
        # carry the changes
        self.set_untracked_attr("value", value)
        if recurse_tracker:
            self._sushrut_dc_init_rcrsv_trk = recurse_tracker
        if not hasattr(self, '_sushrut_dc_init_rcrsv_trk') or self._sushrut_dc_init_rcrsv_trk is None:
            self._sushrut_dc_init_rcrsv_trk = {}

        if not dont_parse:
            self.parse_value()

        if recurse_tracker is None:
            # NOTE: this leaves the dictionary in child
            # Dataobjects, it only deletes the top
            # level tracking dictionary.
            self._sushrut_dc_init_rcrsv_trk.clear()
            del self._sushrut_dc_init_rcrsv_trk
            self._sushrut_dc_init_rcrsv_trk = None

    def reset_target_edges(self, recurse=False, clear_exe=False):
        """
        override Node's reset_target_edges so that
        data complex can prevent contents from being scheduled
        by not resetting any "contains" targets.

        This situation occurs when kamal comparer nodes flow down
        a given branch.
        """
        if not hasattr(self, '_followedTargets'):
            return
        bkup = [x for x in self._followedTargets if x.rule != "contains"]
        tmp = [x for x in self._followedTargets if x.rule == "contains"]

        self._followedTargets = list(bkup)
        super().reset_target_edges(recurse=recurse)
        if tmp:
            self._followedTargets.extend(tmp)

    def set_output_for_fe(self, runtime):
        """
        We don't want to send all the data via the output channel
        as it has been delivered via the map changes capability.

        So, the output should just be an indicator that this node
        has changed.
        """
        runtime.set_output(self.id, f"Look at me: {self.get_targets('contains')}\n\n{self._graph}")

    def update_from_output(self, output):
        """
        Override the _data method such that we can
        interrogate the contents of output (assuming there are some...)
        """
        items_to_add = list()
        if "contents" in output:
            for k, v in output["contents"].items():
                """
                Look for a matching contained item.... if not present, prepare it for adding
                """
                match = None
                for edge in self.get_targets('contains'):
                    tgt = edge.get_target()
                    if hasattr(tgt, "id") and tgt.id == v["id"]:
                        """
                        We've found a matching item in the output
                        """
                        match = tgt
                        break

                if match:
                    match.update_from_output(v)
                    break
                else:
                    """
                    This node does not contain this output item yet, so, we need to
                    create it and then add it as a target once done going through
                    all the contents of the target
                    """
                    from nord.sushrut.utils import mapentry_from_output
                    items_to_add.append((k, mapentry_from_output(self.get_graph(), v, None), v["type"]))

            for name, item, etype in items_to_add:
                wrk_node = item
                wrk_node._sushrut_dataobject_keyname = name
                wrk_node.name = name
                wrk_node.type = etype
                edge = Contains()
                edge.set_nodes(self, wrk_node, "contains")
                self.set_target(wrk_node, "contains", edge)
        else:
            """
            This seems like some malformed output
            """
            raise Exception(f"Dataobject could not process given output: {output}")


sys.modules[__name__] = DataComplex
