"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

from keyword import iskeyword
import re
import importlib
import inspect
from nord.nord.exceptions import RuntimeException as RuntimeException
import nord.sushrut.nodes as sushrut_nodes


_bad_char = re.compile(r'[^\w0-9_]')
_bad_start = re.compile(r'[0-9]')
_white_space = re.compile(r'[\s]+')
_py_to_sushrut_tp_map = {}
_py_to_sushrut_tp_name = {}


"""
Nodes in Sushrut should have a _type
attribute, which this function will use
to cache for later mapping purposes
"""


def populate_type_map():
    global _py_to_sushrut_tp_map, _py_to_sushrut_tp_name
    for name in sushrut_nodes.__all__:
        mod = importlib.import_module(f"nord.sushrut.nodes.{name}")
        if hasattr(mod, '_type') and not name == "Dataobject":
            _py_to_sushrut_tp_map[mod._type] = mod
            _py_to_sushrut_tp_name[mod._type] = name


def get_modname_for_type(ptype):
    global _py_to_sushrut_tp_name

    # temp class def for method and function type mapping
    class __tmp():
        def a(self):
            pass
    if ptype in (type(__tmp.a), type(__tmp().a)):
        ptype = "function"
    return _py_to_sushrut_tp_name.get(ptype, None)


def py_to_sushrut(input, graph):
    cls = _py_to_sushrut_tp_map.get(type(input), None)
    if cls is None:
        if callable(input) and not inspect.isclass(input):
            cls = sushrut_nodes.Datafunction
        else:
            cls = sushrut_nodes.Dataobject
    return cls(graph)


def prepare_output(output, graph):
    if type(output) is not dict:
        x = py_to_sushrut(output, graph)
        x.set(output)
        return x.json_serialize()
    else:
        return output


def make_transportable(val):
    """
    Given an input value, using the known types
    to sushrut, convert to the matching known type
    and, call that type's json_serialize method to get
    a transportable value, which is then returned
    """
    cls = _py_to_sushrut_tp_map.get(type(val), None)
    if cls is None:
        if callable(val) and not inspect.isclass(val):
            cls = sushrut_nodes.Datafunction
        elif callable(val):
            cls = sushrut_nodes.Dataobject
        else:
            return val
    inst = cls(None)
    inst.set(val)
    return inst.json_serialize()


def mapentry_from_output(graph, output, sushrut_obj=None):
    """
    Given the output from the BE, create (or potentially update)
    the object from the contents in the provided dictionary (output)

    if the object is None, create a new one, populate and return it.
    """
    if sushrut_obj is not None and hasattr(sushrut_obj, "id") and "id" in output and output["id"] == sushrut_obj.id:
        """
        this means we have a populated mapentry for the output receiving visible whose ID matches
        that of the output, so, the mapentry should update itself from the output.
        """
        sushrut_obj.update_from_output(output)
        return sushrut_obj
    elif sushrut_obj is None:
        """
        We need to interrogate the output, and create the appropriate mapentry type, and return it.
        """
        space = output['space']
        nodename = output['type'].capitalize()
        mod = importlib.import_module(f"nord.{space}.nodes.{nodename}")
        rv = mod(graph)
        rv.update_from_output(output)
        return rv
    else:
        """
        It may be that this is an erroneous state, so, notify the user
        that we don't like this....
        """
        from nord.wrigan.visibles import show_error
        show_error("Unable to process output from runtime.")
        return None

    for k, v in output["contents"].items():
        if v["id"] not in sushrut_obj._contents_dictionary:  # ????????????????
            """
            then do a brute force search of the contains targets?

            But, what if we are dealing with a simple type (like int, str etc)
            """

    return None


def sanitize_name(inname):
    rv = inname
    if not rv.isidentifier():
        rv.replace(' ', '_')
        rv = re.sub(_bad_char, ' ', inname)
        rv = rv.strip()
        rv = re.sub(_white_space, '_', rv)
        if re.match(_bad_start, rv[0]):
            rv = "v_" + rv
        if iskeyword(rv):
            rv = 'v_' + rv
    return rv


# NOT TO BE USED. SHOULD BE REMOVED
class DefinableObject:
    """
    Converts a dataobject into a "thing" which
    behaves like a python object and a python dictionary.

    When it is used like an object, modifications to the
    attributes will be reflected in the nodes in the graph.

    Meaning, their values will be modified (more accurately,
    their set methods will be called....)

    When used like a dictionary, gets will always reflect
    the values from the contained nodes' get function

    assigning values via the dictionary approach will....
    also modify the originating node's data (via its set method)
    """

    def __new__(cls, dataobject):
        """
        Given a connected dataobject, evaluate its contained
        nodes, and internalize their get/set methods for future
        utilization through the returned class's container
        methods.
        """
        doa = dict()
        if not hasattr(cls, '__DefObjAttrs'):
            setattr(cls, '__DefObjAttrs', doa)
        for edge in dataobject.get_targets('contains'):
            fname = sanitize_name(edge.get_target().name)
            cnt = 1
            while fname in doa:
                fname = f"{fname}_{cnt}"
                cnt += 1

            def setter(name, value):
                raise RuntimeException(f"Value Setter not defined for {edge.get_target()}")
            if hasattr(edge.get_target(), 'set'):
                setter = edge.get_target().set  # noqa: F811

            def getter(name, value):
                raise RuntimeException(f"Value Getter not defined for {edge.get_target()}")
            if hasattr(edge.get_target(), 'get'):
                getter = edge.get_target().get  # noqa: F811

            doa[fname] = {'get': getter,
                          'set': setter}

            # setattr(cls, fname, None)

        return super(DefinableObject, cls).__new__(cls)
        # return cls

    def __init__(self, dataobject):
        pass

    def __getitem__(self, name):
        if name in getattr(self, '__DefObjAttrs'):
            return getattr(self, '__DefObjAttrs')[name]['get']()
        elif type(name) == int and name < len(getattr(self, '__DefObjAttrs').keys()):
            kn = list(getattr(self, '__DefObjAttrs').keys())[name]
            return (kn, getattr(self, '__DefObjAttrs')[kn]['get']())
        elif type(name) == int:
            raise StopIteration
        else:
            raise AttributeError(f"No such attribute named {name}")

    def __setitem__(self, name, value):
        if name in getattr(self, '__DefObjAttrs'):
            getattr(self, '__DefObjAttrs')[name]['set'](value)
        else:
            raise AttributeError(f"No such attribute named {name}")

    def __len__(self):
        return len(getattr(self, '__DefObjAttrs'))

    def __getattr__(self, name):
        if name in getattr(self, '__DefObjAttrs'):
            return getattr(self, '__DefObjAttrs')[name]['get']()
        elif name in ('keys',
                      'values',
                      'items',
                      'get',
                      'clear',
                      'setdefault',
                      'pop',
                      'popitem',
                      'copy',
                      'update'):
            return getattr(self, f"_{name}")
        else:
            raise AttributeError(f"No such attribute named {name}")

    def __setattr__(self, name, value):
        if name in getattr(self, '__DefObjAttrs'):
            getattr(self, '__DefObjAttrs')[name]['set'](value)
        else:
            raise AttributeError(f"No such attribute named {name}")

    def _keys(self):
        return getattr(self, '__DefObjAttrs').keys()

    def _values(self):
        return [self.__getattr__(k) for k in getattr(self, '__DefObjAttrs').keys()]

    def _items(self):
        return [(k, self.__getattr__(k)) for k in getattr(self, '__DefObjAttrs').keys()]

    def _get(self, name, default):
        if name not in getattr(self, '__DefObjAttrs'):
            return default
        else:
            return self.__getattr__(name)

    def _clear(self):
        raise RuntimeException("Not implemented")

    def _setdefault(self):
        raise RuntimeException("Not implemented")

    def _pop(self):
        raise RuntimeException("Not implemented")

    def _popitem(self):
        raise RuntimeException("Not implemented")

    def _copy(self):
        raise RuntimeException("Not implemented")

    def _update(self):
        raise RuntimeException("Not implemented")


"""
Interrogate the nodes for the _type attribute
which is used to align python native types
with each Sushrut type node.
"""
populate_type_map()
