"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Provide visualization for typed data.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.shell.visnode import VisNode
from nord.sigurd.shell.menuItem import MenuItem
import nord.sigurd.managers.event as event
from nord.sigurd.utils.log import warn, info
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.statemanager as em
import nord.sigurd.managers.statemanager as sm
from nord.sigurd.config.globalVars import PLACEDIST
import nord.nord.exceptions.RuntimeException as RuntimeException


class Dataarray(VisNode):
    """Implement the visualization."""

    def __init__(self, uid, app, originof=None):
        """Wrap the parent class constructor."""
        super().__init__(uid, 'dataarray', app, space="sushrut")

        self.add_delegate("append_to_array", "dataarrayappend", 'pass', 'as_tgt',
                          tiptext=_("pass item to\nappend here"))  # noqa: F821
        self.add_delegate("set_current_idx", "dataarraysetidx", 'assign', 'as_tgt',
                          tiptext=_("assign current\narray location here"))  # noqa: F821
        self.add_delegate("get_current_itm", "dataarraygetcurrent", 'assign', 'as_src',
                          tiptext=_("current item\nassigned from here"))  # noqa: F821
        self.add_delegate("get_size", "dataarraygetsize", 'pass', 'as_src',
                          tiptext=_("size of array\npassed from here"))  # noqa: F821

        # self.onMappedFuncs.append(self.gen_contents_menu)
        # self.append_func = VisNode(uid+".append", 'arrayappendmethod', app, space="sushrut")
        # self.append_func.hide()

        if originof is None:
            self._originmodel = None

            def on_add_to_map():
                cm.add(self.mapentry.id)

            self.onMappedFuncs.append(on_add_to_map)

            @event.callback
            @funlog
            def callback2(*args, **kwargs):
                warn("  ---> Calling Enter Dataarray Callback")
                em.enter_container(self.mapentry.id)

            self.menu.add_item(MenuItem("entercontainer", callback2))
        else:
            self._originof = originof

    def show_origin(self):
        self.establish_origin_model()
        self._originmodel.show()

    def hide_origin(self):
        self.establish_origin_model()
        self._originmodel.hide()

    def establish_origin_model(self):
        if self._originmodel is not None:
            return
        self._originmodel = Dataarray(f"origin_{self.uid}", self.app, originof=self)
        self._originmodel.setPos(0.0, 0.0, 0.0)
        self._originmodel.setScale(0.5)
        # This puts the namegen generated UID (readable name...) As the label of the origin
        # minus the prefixed "Container: " bit.
        if self.name == "dataarray":
            label = self.uid.replace("Dataarray: ", '')
        else:
            label = self.name
        self._originmodel.setLabelText(f"-> {label} <-")
        self._originmodel.menu.clear_items()
        self._originmodel.menu.add_item(MenuItem('exitcontainer', em.exit_container, title="Exit Data Array"))

        @event.callback
        @funlog
        def exit_current_container(*args, **kwargs):
            em.exit_container()

    @funlog
    def update_tooltip(self, tooltip):
        if hasattr(self, "mapentry"):
            n_elems = len(self.mapentry.get_targets("contains"))
            if n_elems > 0:
                tooltip.set_text(f"{self.mapentry.name}: {n_elems} " + _("items"))  # noqa: F821
                return
        tooltip.set_text(f"{self.name} " + _("** Empty **"))  # noqa: F821

    @funlog
    def enter(self):
        """
        Upon entering a container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode apply function passing this container
        as the argument
        """
        for edge in self.mapentry.get_sources("use"):
            info("Dataarray rendering edge: {}".format(edge))
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.apply()

        self.show_origin()

    @funlog
    def exit(self):
        """
        Upon leaving the container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode revert function passing this container
        as the argument
        """
        for edge in self.mapentry.get_sources("use"):
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.revert()

        self.hide_origin()

    def process_delegated_click(self, mapentry, payload):
        """
        Placeholder for descendents to override.
        """
        dec = mapentry
        rule = {'dataarraygetcurrent': 'assign',
                'dataarraygetsize': 'pass'}.get(dec.name, None)
        if rule is None:
            raise Exception("IMPLEMENTATION ERROR")
        sm.model_start_connection(dec, rule)

    def connect_allowed(self, other, ctype, direction, delegate=None):
        """
        Do not allow connections of type pass to attach to this node directly.

        Only delegates can be passed to. However, delegates are resolved
        after graph/map load. So, in order to prevent, live-edit time
        connections, only check when called by the delegate (which means)
        that delegate will not be none.
        """
        if direction == 'as_tgt' and delegate is not None:
            if ctype == 'pass':
                if delegate.uid.split(':')[0] in ('append_to_array'):
                    return True
                else:
                    return False
            elif ctype == 'assign':
                if delegate.uid.split(':')[0] in ('set_current_idx'):
                    return True
                else:
                    return False
            else:
                return False
        elif direction == 'as_src' and delegate is not None:
            if ctype == 'pass':
                if delegate.uid.split(':')[0] in ('get_size'):
                    return True
                else:
                    return False
            elif ctype == 'assign':
                if delegate.uid.split(':')[0] in ('get_current_itm'):
                    return True
                else:
                    return False
        else:
            return True

    def gen_contents_menu(self):
        def gen_hide_cbk(itm):
            """This is called by the menu item added to the shown member's model menu."""
            def cbk(*args, **kwargs):
                itm.hide()
            return cbk

        def gen_show_append(i):

            @event.callback
            def show_append(*args, **kwargs):
                """Render the array member and update the member's model menu."""
                warn(f"-.-.-.-.- Calling Dataarray show member {i}")
                # if not hasattr(i, 'visualization'):
                #     om.add_item(self.getPos(), self.app, i)

                # args[0] is the position of the event click, relative to the
                # menu, so it needs to be converted to world coordinates.
                pos = args[0] * PLACEDIST
                pos = cm.cwc().get_relative_point(base.camera, pos)  # noqa: F821
                i.show()
                i.setPos(pos[0], pos[1], pos[2])
                i.reparentTo(cm.cwc())
                if not hasattr(i, '_dyn_hide_item'):
                    mi = MenuItem('anomaly', gen_hide_cbk(i), title="hide")
                    i.menu.add_item(mi)
                    i._dyn_hide_item = True

                    # add a callback to the visualization such that after it is
                    # removed, regenerate the contents menu.
                    def gen_remove_cbk(dobj, item):
                        """Generate the on_remove handler for the visualization."""
                        def on_remove():
                            dobj.gen_contents_menu()
                        return on_remove
                    i.onUnmappedFuncs.append(gen_remove_cbk(self, i))

                    # Also, add a callback to the visualization such that if it's name is changed,
                    # update the menuitem title.
                    def gen_lbl_change_cbk(item):
                        @event.callback
                        def on_change(newtext):
                            item._container_menu_item.set_title(f"Show {newtext}")
                        return on_change
                    i.add_label_change_callback(gen_lbl_change_cbk(i))

                self.menu.hide()

            return show_append

        self._mi_append = MenuItem("dataarrayappend",
                                   gen_show_append(self),
                                   hides_menu=None,
                                   space="sushrut",
                                   title="Append to array")
        self.menu.add_item(self._mi_append)

    """

    Visual features of an array:

    expandable/collapsable
    automatic pagination

    There probably need to be some edges that allow
    append, search, insert, remove, pop, slice

    or, are edges the correct way to do this?

    passing to an array might append?
    how would insert look?



    """
    @funlog
    def show_output(self):
        self.gen_contents_menu()
        self.move_cam_to()
        raise RuntimeException("show_output called")
