"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
Provide visualization for typed data.
"""

from nord.sigurd.utils.verboser import funlog
from nord.sigurd.utils.log import info
from nord.sigurd.shell.visnode import VisNode
from nord.sigurd.shell.menuItem import MenuItem
from nord.sigurd.shell.submenu import SubMenu
import nord.sigurd.managers.event as event
from nord.sigurd.utils.log import warn
import nord.sigurd.managers.containermanager as cm
import nord.sigurd.managers.statemanager as em
from nord.sigurd.shell.connection import Connection
from nord.sigurd.config.globalVars import PLACEDIST
import nord.nord.exceptions.RuntimeException as RuntimeException
import nord.nord.Edge as Edge


def recurse_dig(val):
    if type(val) == list and len(val) <= 2:
        return recurse_dig(val[0])
    else:
        return val


def repoint_edges(itm, dataobject, hidden_only=False):
    for rule in itm.mapentry.get_target_rules():
        for edge in itm.mapentry.get_targets(rule):
            if hasattr(edge, 'visualization') and\
                    (not hidden_only or not edge.get_source().visualization.is_visible()):
                edge.visualization.move_source_vis_only(dataobject)
    for rule in itm.mapentry.get_source_rules():
        if rule == 'contains':
            continue
        for edge in itm.mapentry.get_sources(rule):
            if hasattr(edge, 'visualization') and\
                    (not hidden_only or not edge.get_target().visualization.is_visible()):
                edge.visualization.move_target_vis_only(dataobject)
    if type(itm) == Dataobject:
        # This item could have edges on behalf of members
        # so, those need to be repointed to the passed dataobject
        # as well.
        for edge in itm.mapentry.get_targets('contains'):
            if edge.get_source() == edge.get_target():
                return
            if hasattr(edge.get_target(), 'visualization') and\
                    not edge.get_target().visualization.is_visible():
                repoint_edges(edge.get_target().visualization, dataobject, hidden_only=hidden_only)


class Dataobject(VisNode):
    """Implement the visualization."""

    def __init__(self, uid, app, originof=None):
        """Wrap the parent class constructor."""
        super().__init__(uid, 'dataobject', app, space="sushrut")
        self.contentsMenu = SubMenu("container", self.menu, title="Object Contents")
        self.menu.add_item(self.contentsMenu)
        self.onAllMappedFuncs.append(self.gen_contents_menu)
        self.edges_to_vis_members = {}

        # self.append_func = VisNode(uid+".append", 'arrayappendmethod', app, space="sushrut")
        # self.append_func.hide()

        if originof is None:
            self._originmodel = None

            def on_add_to_map():
                cm.add(self.mapentry.id)

            self.onMappedFuncs.append(on_add_to_map)

            @event.callback
            @funlog
            def callback2(*args, **kwargs):
                warn("  ---> Calling Enter Dataobject Callback")
                em.enter_container(self.mapentry.id)

            self.menu.add_item(MenuItem("entercontainer", callback2))
        else:
            self._originof = originof

    def show_origin(self):
        self.establish_origin_model()
        self._originmodel.show()

    def hide_origin(self):
        self.establish_origin_model()
        self._originmodel.hide()

    def establish_origin_model(self):
        if self._originmodel is not None:
            return
        self._originmodel = Dataobject(f"origin_{self.uid}", self.app, originof=self)
        self._originmodel.setPos(0.0, 0.0, 0.0)
        self._originmodel.setScale(0.5)
        # This puts the namegen generated UID (readable name...) As the label of the origin
        # minus the prefixed "Container: " bit.
        if self.name == "dataobject":
            label = self.uid.replace("Dataobject: ", '')
        else:
            label = self.name
        self._originmodel.setLabelText(f"-> {label} <-")
        self._originmodel.menu.clear_items()
        self._originmodel.menu.add_item(MenuItem('exitcontainer', em.exit_container, title="Exit Data Object"))

        @event.callback
        @funlog
        def exit_current_container(*args, **kwargs):
            em.exit_container()

    @funlog
    def update_tooltip(self, tooltip):
        if hasattr(self, "mapentry"):
            n_elems = len(self.mapentry.get_targets("contains"))
            if n_elems > 0:
                tooltip.set_text(f"{self.mapentry.name}: " + f"{n_elems} " + _("members"))  # noqa: F821
                return
        tooltip.set_text(f"{self.name}: " + _("** No Members or Methods **"))  # noqa: F821

    @funlog
    def enter(self):
        """
        Upon entering a container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode apply function passing this container
        as the argument
        """
        self.show_origin()
        if not hasattr(self, 'mapentry'):
            return
        for edge in self.mapentry.get_sources("use"):
            info("Dataobject rendering edge: {}".format(edge))
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.apply()

    @funlog
    def exit(self):
        """
        Upon leaving the container give connected nodes the chance to react.

        If this container is using any Graphs, Externals or other Nodes
        with a *Use -> UseContainer edge relationship with this container
        then for each of them call their VisNode revert function passing this container
        as the argument
        """
        self.hide_origin()
        if not hasattr(self, 'mapentry'):
            return
        for edge in self.mapentry.get_sources("use"):
            if hasattr(edge.source, "visualization"):
                edge.source.visualization.revert()

    @funlog
    def show_output(self):
        self.gen_contents_menu()
        self.move_cam_to()
        raise RuntimeException("Dataobject show_output called")

    def gen_contents_menu(self):

        def gen_hide_cbk(itm):
            """This is called by the menu item added to the shown member's model menu."""
            dataobject = self

            def cbk(*args, **kwargs):
                itm.hide()
                repoint_edges(itm, dataobject)
                self.edges_to_vis_members[itm].remove()
                del self.edges_to_vis_members[itm]
            return cbk

        def gen_cbk(i):
            """Return the callback for the given member object's menu item."""
            @event.callback
            # @funlog
            def show_member_callback(*args, **kwargs):
                """Render the object member and update the member's model menu."""
                warn(f"-.-.-.-.- Calling Dataobject show member {i}")
                # args[0] is the position of the event click, relative to the
                # menu, so it needs to be converted to world coordinates.
                pos = args[0] * PLACEDIST
                pos = cm.cwc().get_relative_point(base.camera, pos)  # noqa: F821
                i.visualization.show()
                i.visualization.setPos(pos[0], pos[1], pos[2])
                i.visualization.reparentTo(cm.cwc())
                if not hasattr(i.visualization, '_dyn_hide_item'):
                    mi = MenuItem('anomaly', gen_hide_cbk(i.visualization), title="hide")
                    i.visualization.menu.add_item(mi)
                    i.visualization._dyn_hide_item = True

                    # add a callback to the visualization such that after it is
                    # removed, regenerate the contents menu.
                    def gen_remove_cbk(dobj, item):
                        """Generate the on_remove handler for the visualization."""

                        def on_remove():
                            dobj.contentsMenu.rem_item(item._container_menu_item)
                            dobj.gen_contents_menu()
                        return on_remove
                    i.visualization.onUnmappedFuncs.append(gen_remove_cbk(self, i))

                    # Also, add a callback to the visualization such that if it's name is changed,
                    # update the menuitem title.
                    def gen_lbl_change_cbk(item):
                        @event.callback
                        def on_change(newtext):
                            item._container_menu_item.set_title(f"Show {newtext}")
                        return on_change
                    i.visualization.add_label_change_callback(gen_lbl_change_cbk(i))

                repoint_edges(i.visualization, i.visualization)

                self.menu.hide()
                self.contentsMenu.hide_menu()

                pseudoedge = Edge(None)
                pseudoedge.source = self.mapentry
                pseudoedge.target = i
                # pseudoedge.id = ng.id()

                self.edges_to_vis_members[i.visualization] = Connection('memberof',
                                                                        i.visualization,
                                                                        self,
                                                                        self.app,
                                                                        pseudoedge)
            return show_member_callback

        for edge in self.mapentry.get_targets('contains'):
            tgt = edge.get_target()
            if not self.contentsMenu.has_item(tgt.id):
                # We haven't "contained" this target yet
                name = tgt.name
                # if the below is true, then the target has just been
                # added in the UI.
                if type(name) is not str:
                    name = tgt.visualization.name
                menuitem = MenuItem(tgt.type,
                                    gen_cbk(tgt, ),
                                    title=f"Show {name}",
                                    space=tgt.get_space())
                menuitem.name = tgt.id
                self.contentsMenu.add_item(menuitem)
                tgt._container_menu_item = menuitem
                text = tgt.name
                if hasattr(tgt, 'get_value') and tgt.get_value() is not None:
                    text = f"{text} = {tgt.get_value()}"
                self.add_decoration(name, tgt.type, tiptext=text)

                # Go through all the
                # contained targets
                # and repoint edges to this dataobject.
                if hasattr(tgt, 'visualization'):
                    repoint_edges(tgt.visualization, self)

    # def enter(self):
    #     pass

    # def exit(self):
    #     pass

    def on_connect(self, conn):
        """Called when this databaject is connected to another node.

        THIS MIGHT BE REDUNDANT (or even harmful) NOW
        """
        warn(f"ON_CONNECT CALLED on {conn.source} --- {conn.type} ----> {conn.target}")
        if conn.type == "contains":
            if type(conn.target) == Dataobject:
                conn.target.hide_children()
            self.gen_contents_menu()
            conn.target.hide()
            conn.hide()

    def setPos(self, x, y, z):
        super().setPos(x, y, z)

        if not hasattr(self, 'mapentry'):
            return
        for edge in self.mapentry.get_targets('contains'):
            if hasattr(edge.get_target(), 'visualization'):
                if type(edge.get_target().visualization) == Dataobject and\
                        edge.get_target().visualization.is_visible():
                    continue
                repoint_edges(edge.get_target().visualization, self, hidden_only=True)

    def hide_children(self):
        """Hide any members that might be visible, repointing edges as needed."""
        if not hasattr(self, 'mapentry'):
            return
        for edge in self.mapentry.get_targets('contains'):
            if hasattr(edge.get_target(), 'visualization') and\
                    edge.get_target().visualization.is_visible():
                repoint_edges(edge.get_target().visualization, self, hidden_only=False)
                edge.get_target().visualization.hide()
        for itm in list(self.edges_to_vis_members.keys()):
            self.edges_to_vis_members[itm].remove()
            del self.edges_to_vis_members[itm]

    def hide(self):
        """Before hiding oneseff, hide all children first."""
        self.hide_children()
        super().hide()

    def handle_map_edge_src_create(self, update):
        """
        To be overridden by elements which need to additionally process
        new edges leaving this node as provided by backend
        """
        self.gen_contents_menu()

    def handle_map_edge_src_delete(self, update):
        """
        To be overridden by elements which need to additionally process
        deleted edges leaving this node as indicated by backend
        """
        self.gen_contents_menu()
