"""
Copyright 2020 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""
import nord.sushrut.nodes.Dataarray as Dataarray
import nord.nord.Map as Map
import nord.nord.Rule as Rule
import nord.nord.Edge as Edge
import json
import nord.sigurd.shell.vismap as vm
from pathlib import Path
import nord.nord.Runtime as RT


Runtime = RT.Runtime


def test_to_simple_dict():
    da = Dataarray(Map(None))
    da.set([1, 2, 3, 4])
    dct = da.to_dict()
    assert "id" in dct
    assert "type" in dct and dct["type"] == "dataarray"
    assert "space" in dct and dct["space"] == "sushrut"

    assert len(da.get_targets("contains")) == 4

    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert 'idx_0' in dct["contents"]
    assert 'idx_1' in dct["contents"]
    assert 'idx_2' in dct["contents"]
    assert 'idx_3' in dct["contents"]
    assert dct['contents']['idx_0']["value"] == 1
    assert dct['contents']['idx_1']["value"] == 2
    assert dct['contents']['idx_2']["value"] == 3
    assert dct['contents']['idx_3']["value"] == 4
    assert dct['contents']['idx_0']["type"] == "dataint"
    assert dct['contents']['idx_1']["type"] == "dataint"
    assert dct['contents']['idx_2']["type"] == "dataint"
    assert dct['contents']['idx_3']["type"] == "dataint"


def test_repeat_int_vals_to_dict():
    da = Dataarray(Map(None))
    da.set([1, 1, 1, 1])
    dct = da.to_dict()
    assert "id" in dct
    assert "type" in dct and dct["type"] == "dataarray"
    assert "space" in dct and dct["space"] == "sushrut"

    assert len(da.get_targets("contains")) == 4

    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert 'idx_0' in dct["contents"]
    assert 'idx_1' in dct["contents"]
    assert 'idx_2' in dct["contents"]
    assert 'idx_3' in dct["contents"]
    assert dct['contents']['idx_0']["value"] == 1
    assert dct['contents']['idx_1']["value"] == 1
    assert dct['contents']['idx_2']["value"] == 1
    assert dct['contents']['idx_3']["value"] == 1
    assert dct['contents']['idx_0']["type"] == "dataint"
    assert dct['contents']['idx_1']["type"] == "dataint"
    assert dct['contents']['idx_2']["type"] == "dataint"
    assert dct['contents']['idx_3']["type"] == "dataint"


def test_repeat_none_vals_to_dict():
    da = Dataarray(Map(None))
    da.set([None, None, None, None])
    dct = da.to_dict()
    assert "id" in dct
    assert "type" in dct and dct["type"] == "dataarray"
    assert "space" in dct and dct["space"] == "sushrut"

    assert len(da.get_targets("contains")) == 4

    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert 'idx_0' in dct["contents"]
    assert 'idx_1' in dct["contents"]
    assert 'idx_2' in dct["contents"]
    assert 'idx_3' in dct["contents"]
    assert dct['contents']['idx_0']["type"] == "datavoid"
    assert dct['contents']['idx_1']["type"] == "datavoid"
    assert dct['contents']['idx_2']["type"] == "datavoid"
    assert dct['contents']['idx_3']["type"] == "datavoid"


def test_to_complex_dict():
    da = Dataarray(Map(None))
    da.set([
        {"ted": 1, "bill": {
            "a": 3, "b": "A STRING"}
         },
        {"somelist": [5, 6, 7]},
        "STRINGONTOP",
        {"list of lists": [[1, 2], ["a", "b"], [0.1, 0.3], [None, None]]}
    ])
    dct = da.to_dict()
    assert "id" in dct
    assert "type" in dct and dct["type"] == "dataarray"
    assert "space" in dct and dct["space"] == "sushrut"

    assert len(da.get_targets("contains")) == 4

    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert dct['contents']['idx_0']["type"] == "dataobject"
    assert dct['contents']['idx_0']["contents"]["ted"]["value"] == 1
    assert dct['contents']['idx_0']["contents"]["ted"]["type"] == "dataint"

    assert dct['contents']['idx_0']["contents"]["bill"]["contents"]["a"]["value"] == 3
    assert dct['contents']['idx_0']["contents"]["bill"]["contents"]["a"]["type"] == "dataint"
    assert dct['contents']['idx_0']["contents"]["bill"]["contents"]["b"]["value"] == "A STRING"
    assert dct['contents']['idx_0']["contents"]["bill"]["contents"]["b"]["type"] == "datastr"
    assert dct['contents']['idx_1']["contents"]["somelist"]["type"] == "dataarray"
    # assert dct['contents'][0]["contents"]["somelist"]["value"] == [5, 6, 7]
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_0']["value"] == 5
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_0']["type"] == "dataint"
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_1']["value"] == 6
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_1']["type"] == "dataint"
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_2']["value"] == 7
    assert dct['contents']['idx_1']["contents"]["somelist"]["contents"]['idx_2']["type"] == "dataint"
    assert dct['contents']['idx_2']["value"] == "STRINGONTOP"
    assert dct['contents']['idx_2']["type"] == "datastr"

    assert dct['contents']['idx_3']["type"] == "dataobject"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["type"] == "dataarray"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_0']["type"] == "dataarray"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_0']["contents"]['idx_0']["type"] == "dataint"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_0']["contents"]['idx_0']["value"] == 1
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_0']["contents"]['idx_1']["type"] == "dataint"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_0']["contents"]['idx_1']["value"] == 2

    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_1']["type"] == "dataarray"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_1']["contents"]['idx_0']["type"] == "datastr"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_1']["contents"]['idx_0']["value"] == "a"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_1']["contents"]['idx_1']["type"] == "datastr"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_1']["contents"]['idx_1']["value"] == "b"  # noqa: E501

    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_2']["type"] == "dataarray"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_2']["contents"]['idx_0']["type"] == "datafloat"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_2']["contents"]['idx_0']["value"] == 0.1  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_2']["contents"]['idx_1']["type"] == "datafloat"  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_2']["contents"]['idx_1']["value"] == 0.3  # noqa: E501

    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_3']["type"] == "dataarray"
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_3']["contents"]['idx_0']["type"] == "datavoid"  # noqa: E501
    assert "value" not in dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_3']["contents"]['idx_0']  # noqa: E501
    assert dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_3']["contents"]['idx_1']["type"] == "datavoid"  # noqa: E501
    assert "value" not in dct['contents']['idx_3']["contents"]["list of lists"]["contents"]['idx_3']["contents"]['idx_1']  # noqa: E501


def test_append_to_array():
    da = Dataarray(Map(None))
    da.set(["a"])
    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert dct['contents']['idx_0']["type"] == "datastr"
    assert dct['contents']['idx_0']["value"] == 'a'
    assert len(dct["contents"]) == 1
    assert len(da.get_targets("contains")) == 1
    assert da.get_targets("contains")[0].get_target().get() == 'a'
    edge = Edge(None)
    rule = Rule(None, None, edge)
    edge.set_cargo("b")
    da.append_to_array(rule)

    dct = da.to_dict(include_contents=True)
    assert 'contents' in dct
    assert dct['contents']['idx_0']["type"] == "datastr"
    assert dct['contents']['idx_0']["value"] == 'a'
    assert dct['contents']['idx_1']["type"] == "datastr"
    assert dct['contents']['idx_1']["value"] == 'b'
    assert len(dct["contents"]) == 2
    assert len(da.get_targets("contains")) == 2
    assert da.get_targets("contains")[0].get_target().get() == 'a'
    assert da.get_targets("contains")[1].get_target().get() == 'b'


def test_json_serialize():
    da = Dataarray(Map(None))
    da.set([1, 2, 3])
    res = da.json_serialize()
    # Since there are randomized contents in the string
    # we need to make sure the core elements are in there...
    assert "idx_0" in res
    assert "idx_1" in res
    assert "idx_2" in res
    assert '"value": 1' in res
    assert '"value": 2' in res
    assert '"value": 3' in res
    recon = json.loads(res)
    assert recon["contents"]["idx_0"]["value"] == 1
    assert recon["contents"]["idx_1"]["value"] == 2
    assert recon["contents"]["idx_2"]["value"] == 3
    assert recon["contents"]["idx_0"]["name"] == "idx_0"
    assert recon["contents"]["idx_1"]["name"] == "idx_1"
    assert recon["contents"]["idx_2"]["name"] == "idx_2"
    assert recon["contents"]["idx_0"]["type"] == 'dataint'
    assert recon["contents"]["idx_1"]["type"] == 'dataint'
    assert recon["contents"]["idx_2"]["type"] == 'dataint'
    assert recon["contents"]["idx_0"]["space"] == 'sushrut'
    assert recon["contents"]["idx_1"]["space"] == 'sushrut'
    assert recon["contents"]["idx_2"]["space"] == 'sushrut'
    assert "id" in recon["contents"]["idx_0"]
    assert "id" in recon["contents"]["idx_1"]
    assert "id" in recon["contents"]["idx_2"]
    assert "name" in recon["contents"]["idx_0"]
    assert "name" in recon["contents"]["idx_1"]
    assert "name" in recon["contents"]["idx_2"]


def test_fetch_by_position():
    da = Dataarray(Map(None))
    da.set([1, 2, 3])
    val = da.get_index(0).get()
    assert val == 1
    val = da.get_index(1).get()
    assert val == 2
    val = da.get_index(2).get()
    assert val == 3


def test_fetch_next():
    assert 1 == 0


def test_array_of_arrays():
    da = Dataarray(Map(None))
    da.set([
        [
            [4, 5],
            [6, 7, 8, 9]
        ],
        [
            ['a', 'b', 'c', 'd']
        ]
    ])
    dct = da.to_dict()
    assert "id" in dct
    assert "type" in dct and dct["type"] == "dataarray"
    assert "space" in dct and dct["space"] == "sushrut"

    assert len(da.get_targets("contains")) == 2

    dct = da.to_dict(include_contents=True)

    da.set([[[[[2]]]]])
    assert len(da.get_targets("contains")) == 1
    first = da.get_index(0)
    second = first.get_index(0)
    third = second.get_index(0)
    fourth = third.get_index(0)
    val = fourth.get_index(0)
    assert val.get() == 2


# @pytest.mark.xfail  # When run with all other tests, getTightBounds returns None...
def test_visible_array_contents(init_panda3d):
    app = init_panda3d
    mappath = '/test/maps/test_dataarray_add.n'
    if Path(f"./sushrut{mappath}").exists():
        app.load_map(f"./sushrut{mappath}")
    elif Path('.{mappath}').exists():
        app.load_map(f".{mappath}")
    else:
        raise Exception("Cannot find test map to load")
    karte = vm.get_working_map()
    vessal = Runtime()
    array = karte.get_node('1f9c3d97-9476-4998-ac4a-776f6a401547')
    acontents = array.get_targets('contains')
    assert len(acontents) == 0
    karte.connect_cdc_loopback()  # to ensure that requested graph changes exist locally
    karte.run(vessal, None)
    acontents = array.get_targets('contains')
    assert len(acontents) == 3
    assert acontents[0].target.value == 12
    assert acontents[1].target.value == 99
    assert acontents[2].target.value == 8

    import nord.sigurd.managers.objectmanager as om
    idx_0 = om.fetch(acontents[0].target.id)
    idx_1 = om.fetch(acontents[1].target.id)
    idx_2 = om.fetch(acontents[2].target.id)

    assert idx_0.mapentry.value == 12
    assert idx_1.mapentry.value == 99
    assert idx_2.mapentry.value == 8
    # while True:
    #     app.one_loop_iteration()


def test_append_in_loop():
    try:
        m = json.load(open('test/maps/test_dataarray_add_loop.n', 'r'))
    except FileNotFoundError:
        m = json.load(open('sushrut/test/maps/test_dataarray_add_loop.n', 'r'))

    karte = Map(m, verbose=True)

    vessal = Runtime()
    karte.run(vessal, None)
    # Notice the lack of connecting the map's cdc loopback handler
    # this means that the contents of the dataarray
    # will exist in the targets of the dataarray, but, the map (karte)'s nodes
    # member will not contain a reference to these objects
    # assert vessal.get_exec_count() == 5
    assert vessal.get_anomaly() is None
    vals = karte.nodes['d8638e04-e90e-4a17-ad08-0bbc1ade8575'].get_targets("contains")
    assert len(vals) == 4
    for item in vals:
        assert item.get_target().id not in karte.nodes
        assert item.rule == "contains"


def test_append_in_loop_cdc():
    try:
        m = json.load(open('test/maps/test_dataarray_add_loop.n', 'r'))
    except FileNotFoundError:
        m = json.load(open('sushrut/test/maps/test_dataarray_add_loop.n', 'r'))

    vessal = Runtime()

    karte = Map(m, verbose=True)
    # this time, we'll establish the CDC connection
    # and, this time the contents of the data array will
    # be referenced in karte.nodes
    karte.connect_cdc_loopback()
    karte.run(vessal, None)

    # assert vessal.get_exec_count() == 5
    assert vessal.get_anomaly() is None
    assert len(karte.nodes['d8638e04-e90e-4a17-ad08-0bbc1ade8575'].get_targets("contains")) == 4
    vals = karte.nodes['d8638e04-e90e-4a17-ad08-0bbc1ade8575'].get_targets("contains")
    assert len(vals) == 4
    for item in vals:
        assert item.get_target().name.startswith('idx_')
        assert item.get_target().id in karte.nodes
