"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--
The base class for the secondary construct of the Graph "lanquage", the Edge.

Edges are the connections between vertices of the graph. Edges move data and
execution between nodes in Nord. Like Nodes, Edges are expected to be subclassed.
This module provides the common capabilities all edges should have. Subclasses
are welcome to override these methods as needed.
"""
import nord.nord.exceptions.UseException as UseException
import nord.nord.exceptions.RuntimeException as RuntimeException
import nord.nord.exceptions.EdgeException as EdgeException
import nord.nord.utils.SpaceManager as use
import sys
import nord.nord.Base as Base


class Edge(Base):
    """
    Edges are the connections between vertices of the graph. Edges connect Nodes in the Map.

    This class is the base class used to implement relationships between different types of Nodes.
    An example is the Assign subclass.

    >>> Edge() #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Edge: [None]  --Edge-->  [None]
    """

    def set_nodes(self, src, tgt, rule):
        """
        Assign the nodes at either end of the edge.

        >>> from nord.Node import Node
        >>> src = Node()
        >>> src.apply({"id": 1, "name": "fred", "value": 77})
        >>> tgt = Node()
        >>> tgt.apply({"id": 2, "name": "larry", "value": 92})
        >>> e = Edge()
        >>> e.set_nodes(src, tgt, 'Dummy')
        >>> repr(e) #doctest: +NORMALIZE_WHITESPACE
        'Edge: [Node(Node): <id: 1, name: fred>]  --Edge-->  [Node(Node): <id: 2, name: larry>]'

        """
        self.source = src
        self.target = tgt
        self.rule = rule
        src.set_target(tgt, rule, self)

    def flag_followed(self, runtime):
        """Flag this edge as having been followed."""
        if not hasattr(self, "_followed"):
            self._followed = runtime.get_exec_count()

        # tell the source about this edge having been followed
        self.source.set_followed(self, source=True)
        self.target.set_followed(self, source=False)

    def clear_followed(self):
        """Remove the Flag that this edge has been followed."""
        if hasattr(self, "_followed"):
            delattr(self, "_followed")

        # tell the source about this edge having been unfollowed
        self.source.set_unfollowed(self, source=True)
        self.target.set_unfollowed(self, source=False)

    def been_followed(self, runtime=None):
        """Indicate if this edge has been followed yet."""
        if hasattr(self, "_followed"):
            if runtime is not None:
                return self._followed >= runtime.get_exec_count()
            else:
                return True
        else:
            return False

    def set_cargo(self, cargo):
        """
        Set this edge's payload to the passed cargo data.

        Used for the source rules to store the meaningful
        data for use by the target rules
        """
        self._cargo = cargo

    def get_cargo(self):
        """
        Return this edge's cargo, if it exists.

        Used for the target rules to retrieve the meaningful
        data placed by the target rules
        """
        if hasattr(self, '_cargo'):
            return self._cargo
        else:
            return None

    def has_cargo(self):
        """Return whether or not this edge has cargo."""
        return hasattr(self, '_cargo')

    def apply_delegate(self, rule):
        """Given the input rule, attempt to employ the delegation requested by its edge."""
        if hasattr(rule.edge, 'delegate'):
            method = rule.edge.delegate['uid']
            if rule.edge.delegate['direction'] == 'as_src':
                node = rule.edge.source
            else:
                node = rule.edge.target
            if hasattr(node, method):
                func = getattr(node, method)
                func(rule)
            else:
                raise RuntimeError(_("Extension implementation error."  # noqa: F821
                                     " no method named {} in delegate: {}").format(method, node))
        else:
            raise RuntimeError(_("Extension implementation error."  # noqa: F821
                                 " apply_delegate requested, but no delegates available."))

    def traverse(self, srule, trule):
        """
        Abstract method called when the edge is followed.

        This method should be overridden by all subclasses to perform whatever special magi they
        will perform to take the return from the source rule and pass that to the target rule
        """
        raise EdgeException("{} is using the base Edge Traverse method...!"
                            .format(str(self)))

    def follow(self, runtime, verbose=False):
        """
        Main "execution" method for all edges.

        Following an edge is a recursive function that makes ensures all the source node source
        edges have been followed. Then signals the child (subclass) to:
        """
        if verbose:
            print("   ---> following: {}".format(self))

        # Only add edges that terminate with the current node, if they haven't been followed before
        src_modname = "{}{}".format(type(self.source).__name__, type(self).__name__)
        if 'space' in dir(self.source):
            provider = self.source.space
        else:
            provider = 'nord'

        use.space(provider)
        try:
            mod = use.module('nord.{}.rules.{}'.format(provider, src_modname))
        except UseException:
            raise RuntimeException(
                "Executing node: {} - No src rule execution source module named '{}.rules.{}' available"  # noqa: E501
                .format(str(runtime.current), provider, src_modname))

        if mod is not None:

            srule = mod(self.source, runtime, self)

        else:
            raise RuntimeException(
                "Executing node: {} - No src rule execution source module named '{}.rules.{}' available"  # noqa: E501
                .format(str(runtime.current), provider, src_modname))  # pragma: no cover

        tgt_modname = "{}{}".format(type(self).__name__, type(self.target).__name__)
        if 'space' in dir(self.target):
            provider = self.target.space
        else:
            provider = 'nord'

        use.space(provider)

        try:
            mod = use.module('nord.{}.rules.{}'.format(provider, tgt_modname))
        except UseException:
            raise RuntimeException(
                "Executing node: {} - No tgt rule execution source module named '{}.rules.{}' available"  # noqa: 501
                .format(str(runtime.current), provider, tgt_modname))  # pragma: no cover

        if mod is not None:

            trule = mod(self.target, runtime, self)

        else:
            raise RuntimeException(
                "Executing node: {} - No tgt rule execution source module named '{}.rules.{}' available"  # noqa: 501
                .format(str(runtime.current), provider, tgt_modname))  # pragma: no cover

        self.traverse(srule, trule)

        self.flag_followed(runtime)

    def to_dict(self):
        """
        Convert this edge to a dictionary for saving / serializing.

        Return a dictionary of all simple datatype members, or
        known dictionaries, as well as the source, target, and rule
        """
        rv = {}
        for k in (k for k in sorted(self.__dict__.keys()) if not k.startswith("_")):
            v = getattr(self, k)
            if type(v) in (str, int, float, bool):
                rv[k] = v
            elif k in ("delegate",) and type(v) == dict:
                rv[k] = v.copy()

        rv["source"] = "{}:{}".format(self.source.id, self.source.name)
        rv["target"] = "{}:{}".format(self.target.id, self.target.name)
        rv["rule"] = self.rule.lower()
        if self._spacename != "nord" or\
                self.source.get_space() != 'nord' or\
                self.target.get_space() != 'nord':
            rv["space"] = self._spacename
            rv["to_space"] = self.target.get_space()
            rv["from_space"] = self.source.get_space()

        return rv

    def __repr__(self):
        """Return the representation of this edge."""
        src = None
        if hasattr(self, "source"):
            src = self.source

        tgt = None
        if hasattr(self, "target"):
            tgt = self.target

        followed = ""
        if self.been_followed():
            followed = "**"
        rv = "{}: [{}]  --{}-->  [{}]{}"\
             .format(type(self).__name__, str(src), type(self).__name__, str(tgt), followed)
        return rv


sys.modules[__name__] = Edge
