"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

The Map is the analog of program, script, module and/or library from other languages.

Map'self are graph'self made up of nodes connected by edges which have rules relating
the types of nodes with the type of edge. Map'self can reference other maps.
"""

import uuid
import sys
import nord.nord.utils.SpaceManager as use
import nord.nord.InstanceOf as InstanceOf


class Map(object):
    """
    Map'self are the representation of the programs to be executed.

    A graph stored within a dictionary is used to
    generate the element mesh of objects which are then executed or debugged.
    """

    def __init__(self, inp, verbose=False):
        """Create the map by traversing the input dictionary."""
        self.prep_for_init()
        if inp is not None:
            self.load(inp, verbose)

    def prep_for_init(self):
        """Prepare the map to be executed."""
        if not hasattr(self, '_initialized'):
            use.space('nord')
            self.Container = use.module('nord.nord.nodes.Container')
            self.nodes = dict()
            self.edges = dict()
            self.rules = dict()
            self.extensions = dict()
            self.start_node = None
            self._initialized = True

    def unload(self):
        """Clear out the current map's data."""
        del self.Container
        self.nodes.clear()
        self.edges.clear()
        self.rules.clear()
        self.start_node = None
        self.Container = use.module('nord.nord.nodes.Container')

    def load(self, inp, verbose):
        """Process the input dict structure into Nodes and Edges."""
        self.verbose = verbose

        # Do not overwrite the name if it has already been set.
        if not hasattr(self, 'name'):
            if "name" in inp:
                self.name = inp['name']
            else:
                self.name = str(uuid.uuid4())

        # prepare the spaces with any extensions requested by the map/graph.
        # permissions and compatible versions are too be implemented.
        # as are overrides
        if "extensions" in inp:
            for extn in inp["extensions"]:
                ext = inp["extensions"][extn]
                self.extensions[extn] = ext
                use.space(extn)
                if "locations" in ext:
                    for provides in ext["locations"]:
                        for loc in ext["locations"][provides]:
                            use.add_location(loc, provides)

        # If payment details are included in the input
        # data, establish an EconomicSession
        if "signed_tx_block" in inp:
            # A transaction block has been provided by the
            # requestor. This is a prepayment, the runtime
            # will need to refund the unused portion when
            # the map is done.
            import nord.pasion.visibles.wallet as wallet
            wallet.establish_session_from_tx(inp["signed_tx_block"], inp["refunds_to"], inp["payment_to"])

        if "map" in inp:
            if "nodes" in inp['map']:
                delayed_nodes = []
                for n in inp['map']['nodes']:
                    if "id" in n and "name" in n and "type" in n and "instanceof" not in n:
                        if 'space' not in n:
                            provider = 'nord'
                        else:
                            provider = n['space']

                        use.space(provider)

                        mod = use.module('nord.{}.nodes.{}'.format(provider, n['type'].title()))
                        if mod is not None:
                            # this will look through the loaded modules for a module
                            # named after the type field in the node element
                            # it will create a new instance of that type
                            # and apply the data provided by the map to that node
                            # and lastly store the node by it'self ID in the nodes dictionary

                            try:
                                node = mod()
                            except TypeError:  # pragma: no cover
                                print("Invalid node initialization for node of type: {}"
                                      .format(n['type'].title()))  # pragma: no cover
                            node.apply(n)
                            self.nodes[n['id']] = node

                        else:  # pragma: no cover
                            print("{} node type not implemented".format(n['type'].title()))
                    elif "id" in n and "instanceof" in n:
                        # This node is an instance of a prior node. Associate if
                        # the referenced node is instanceable.
                        delayed_nodes.append(n)
                    else:  # pragma: no cover
                        print("Malformed node detected: {}".format(n))

                for n in delayed_nodes:
                    if n['instanceof'] in self.nodes and\
                            isinstance(self.nodes[n['instanceof']], InstanceOf):
                        parent = self.nodes[n['instanceof']]
                        child = parent.add_instance(n)
                        child.type = parent.type
                        self.nodes[n['id']] = child
                    else:
                        print("Instance of requested without known parent: {}".format(n))

                if "edges" in inp['map']:
                    for e in inp['map']['edges']:
                        if "source" in e and "target" in e and "rule" in e:
                            if e['source'].split(":")[0] in self.nodes and\
                               e['target'].split(":")[0] in self.nodes:
                                if 'space' not in e:
                                    provider = 'nord'
                                else:
                                    provider = e['space']

                                use.space(provider)

                                mod = use.module('nord.{}.edges.{}'.format(provider, e['rule'].title()))
                                if mod is not None:

                                    # this will create a new edge of type "rule"
                                    # between the source and the target nodes

                                    edge = mod()
                                    edge.apply(e)
                                    source = self.nodes[e['source'].split(":")[0]]
                                    target = self.nodes[e['target'].split(":")[0]]
                                    edge.set_nodes(source, target, e['rule'])
                                    # this suggests that a source and target can only be
                                    # connected once per rule
                                    self.edges[(source, target, e['rule'])] = edge
                                    if e['rule'] not in self.rules:
                                        self.rules[e['rule']] = list()
                                    self.rules[e['rule']].append(edge)

                                else:  # pragma: no cover
                                    print("{} edge type not implemented".format(e['rule'].title()))
                            else:  # pragma: no cover
                                print(f"Unknown node {e['source'].split(':')[0]} or"
                                      f" {e['target'].split(':')[0]} requested by edge: {e}."
                                      f" Known nodes are: {self.nodes}")
                                # sys.exit(1)
                        else:  # pragma: no cover
                            print("Malformed edge detected: {}".format(e))

                    # pick a node to begin map execution.
                    self.detect_start_node()

                else:
                    print("No edge data found in input data")
                    # sys.exit(1)
            else:
                print("No node data found in input data")
                # sys.exit(1)
        else:
            print("No map found in input data")
            # sys.exit(1)

        if verbose:
            print("Map created with {} Nodes and {} Edges".format(len(self.nodes), len(self.edges)))

    def get_sourceless_nodes(self):
        """
        Get all nodes that have no source edges.
        """
        candidates = []
        found_one = False
        for nname in self.nodes:
            node = self.nodes[nname]
            if len(node.get_source_rules()) == 0:
                candidates.append(node)
                found_one = True

        candidates = sorted(candidates)
        return candidates, found_one

    def detect_start_node(self):
        """
        Find nodes with no edges targetting it.
        Make it/them the start node(s).

        Should this approach not find a node, set the first node in the
        map as the starting node.... who knows what the ramifications
        of this approach are...?
        """
        if self.start_node is None:
            self.start_node = []
            candidates, found_one = self.get_sourceless_nodes()
            if not found_one:
                print("Run requested for map, but no initial container specified")

            # first look for any containers with no inbound edges.
            # if there is one named "main" (any case), set it to start
            # otherwise, choose the first container
            # if there are no containers, then, choose the
            # first non-container node with no inbound edges.
            found_main = False
            found_a_container = False
            for node in candidates:
                if type(node) == self.Container and\
                        hasattr(node, 'name') and\
                        node.name.lower() == 'main':
                    self.start_node.insert(0, node)
                    found_main = True
                    found_a_container = True
                else:
                    self.start_node.append(node)

            if not found_main:
                for node in candidates:
                    if type(node) == self.Container:
                        self.start_node.remove(node)
                        self.start_node.insert(0, node)
                        found_a_container = True
                        break

            if len(self.start_node) == 0 and self.nodes:
                # fallback to the first node in the list....
                self.start_node = [self.nodes[list(self.nodes.keys())[0]]]

        return self.start_node

    def run(self, runtime, args):
        """
        Execute this map when called from the World (top level, not from another Map).

        In short this method just starts the flow of rules. From that point execution
        is handled within all the rules modules.

        """
        if self.start_node is not None:
            runtime.initialize(self, verbose=self.verbose)
            runtime.begin(self.start_node)

            # self.start_node.execute(runtime=runtime, args=args)
        else:
            # There is no container at the top level
            # So, find a node that is not a target
            # and start the execution there...
            self.detect_start_node()
            if self.start_node is not None:
                runtime.initialize(self, verbose=self.verbose)
                runtime.begin(self.start_node)
            else:
                print("Run requested for map, but no initial container specified")
                raise RuntimeError("Unable to detect a node from which to start the map.")

    def init_debug(self, runtime, args):
        """.The init_debug method starts the map in debugging mode."""
        if self.start_node is not None:
            import nord.nord.utils.Debugger as dbg
            runtime.initialize(self, verbose=True)
            dbg.start(self, runtime, args)

    def debug_next(self):
        """Step to the next element in the flow, and execute whatever rules it has defined."""
        pass

    def get_node(self, nid):
        """
        Fetch the node with id nid from the map, if it exists.

        Return None if it does not.
        """
        if nid in self.nodes:
            return self.nodes[nid]
        else:
            return None

    def __repr__(self):
        """Traverse the map (graph) and print out all the nodes, edges and rules."""
        rv = "Nodes\n================\n"
        for k in sorted(self.nodes.keys()):
            rv += "-----\n{}\n-----\n".format(repr(self.nodes[k]))

        rv += "\n\n\nEdges\n===============\n"

        for k in sorted(self.edges.keys()):
            rv += "  ---->\n{}\n".format(repr(self.edges[k]))

        rv += "\n\nRules\n===============\n"

        for k in sorted(self.rules.keys()):
            rv += "\n(====)\n-- {} Rule --\n".format(k.title())
            for i in self.rules[k]:
                rv += "\n  R: {}".format(str(i))

        return rv

    def to_dict(self):
        """
        Return the dictionary representation of this map.

        Traverse the map (graph) and convert all the nodes and
        edges into a dictionary
        """
        rv = dict()
        rv["map"] = {"nodes": [], "edges": []}
        rv["name"] = self.name
        rv["extensions"] = {}
        for n in sorted(self.nodes.keys()):
            rv["map"]["nodes"].append(self.nodes[n].to_dict())
        for e in sorted(self.edges.keys()):
            rv["map"]["edges"].append(self.edges[e].to_dict())
        for s in sorted(use.get_spaces()):
            if s == "nord":
                continue
            sd = use.space_to_dict(s)
            if sd is not None:
                rv["extensions"][s] = sd

        return rv


sys.modules[__name__] = Map
