"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

The base class for the primary construct of the Graph "language", the Node.

Nodes are the vertices of the graph. Nodes should perform most of the work
in Nord. Nodes are intended / expected to be subclassed. This module just provides
the common capabilities to support the runtime and fundamental operations
all subclasses should have. Subclasses are welcome to override nearly any of the
methods as appropriate.
"""

import sys
import nord.nord.Base as Base
import nord.nord.RunState as RunState


class Node(Base):
    """
    Nodes are the vertices of the graph.

    They are connected by edges.

    A Node is the base class for all map objects. The primary usage of a node
    is to create subclasses of it. Data, Static and Container are examples of
    direct Node subclasses.

    >>> Node() #doctest: +ELLIPSIS
    Node:

    """
    def handles_events(self):
        """Check if this node is configured to receive events."""
        return hasattr(self, 'handle_event')

    def flag_preempted(self):
        """Mark the node as pre-empted."""
        RunState.s(self).set_preempted()

    def clear_preempted(self):
        """Revert the node to the non-preempted state."""
        RunState.s(self).unset_preempted()

    def flag_exe(self, runtime):
        """Mark the node as executed."""
        RunState.s(self).set_executed(runtime)

    def clear_exe(self, runtime):
        """Revert the node to the un-executed state."""
        RunState.s(self).unset_executed(runtime)

    def has_executed(self):
        """Return if this node has executed."""
        return RunState.s(self).executed()

    def flag_blocking(self):
        """Mark the node as blocked (for input perhaps)."""
        RunState.s(self).set_blocked()

    def clear_blocking(self):
        """Revert the node to the un-blocked state."""
        RunState.s(self).unset_blocked()

    def execute(self, runtime):
        """
        Execute this node.

        This method can be overridden by sub-classes of Node
        to perform whatever action it is that node is intended to perform.

        The runtime will execute this method, passing an instance of itself.

        A node will be executed after all its source edges have been followed, but
        before any of its target edges are.
        """
        self.flag_exe(runtime)

    def set_target(self, tgt, rule, edge):
        """
        Establish requested connection from this node to the Target.

        The set_target method takes a Node instance and rule name as arguments
        and establishes a connection between the two nodes.

        >>> src = Node()
        >>> tgt = Node()
        >>> src.set_target(tgt, "null", Edge())

        """
        if not hasattr(self, "targetRules"):
            self.targetRules = dict()

        if not hasattr(self, "targetNodes"):
            self.targetNodes = dict()

        if not hasattr(self, '_unfollowedTargets'):
            self._unfollowedTargets = list()
            self._followedTargets = list()

        if rule not in self.targetRules:
            self.targetRules[rule] = list()

        if tgt not in self.targetNodes:
            self.targetNodes[tgt] = dict()

        if rule not in self.targetNodes[tgt]:
            self.targetNodes[tgt][rule] = list()

        if edge not in self.targetNodes[tgt][rule]:
            self.targetNodes[tgt][rule].append(edge)

        if edge not in self.targetRules[rule]:
            self.targetRules[rule].append(edge)

        if edge not in self._unfollowedTargets:
            self._unfollowedTargets.append(edge)

        tgt.set_source(self, rule, edge)

    def rem_target(self, tgt, edge):
        """
        Remove data traces of the passed edge to the target.

        The _[un]followed* lists are operational, and not
        dealt with with this function.
        """
        if hasattr(self, 'targetRules'):
            if edge.rule in self.targetRules and\
               edge in self.targetRules[edge.rule]:
                self.targetRules[edge.rule].remove(edge)

        if hasattr(self, 'targetNodes') and\
           tgt in self.targetNodes and\
           edge.rule in self.targetNodes[tgt] and\
           edge in self.targetNodes[tgt][edge.rule]:
            self.targetNodes[tgt][edge.rule].remove(edge)

        tgt.rem_source(self, edge)

    def set_source(self, src, rule, edge):
        """
        Establish this node's connection back to the source.

        The set_source method takes a Node instance and rule name as arguments
        and establishes a connection back to the source.

        >>> src = Node()
        >>> tgt = Node()
        >>> tgt.set_source(src, "null", Edge())

        """
        if not hasattr(self, "sourceRules"):
            self.sourceRules = dict()

        if not hasattr(self, "sourceNodes"):
            self.sourceNodes = dict()

        if not hasattr(self, '_unfollowedSources'):
            self._unfollowedSources = list()
            self._followedSources = list()

        if rule not in self.sourceRules:
            self.sourceRules[rule] = list()

        if src not in self.sourceNodes:
            self.sourceNodes[src] = dict()

        if rule not in self.sourceNodes[src]:
            self.sourceNodes[src][rule] = list()

        if edge not in self.sourceNodes[src][rule]:
            self.sourceNodes[src][rule].append(edge)

        if edge not in self.sourceRules[rule]:
            self.sourceRules[rule].append(edge)

        if edge not in self._unfollowedSources:
            self._unfollowedSources.append(edge)

    def rem_source(self, src, edge):
        """
        Remove data traces of the passed edge from the target.

        The _[un]followed* lists are operational, and not
        dealt with with this function.
        """
        if hasattr(self, 'sourceRules'):
            if edge.rule in self.sourceRules and\
               edge in self.sourceRules[edge.rule]:
                self.sourceRules[edge.rule].remove(edge)

        if hasattr(self, 'sourceNodes') and\
           src in self.sourceNodes and\
           edge.rule in self.sourceNodes[src] and\
           edge in self.sourceNodes[src][edge.rule]:
            self.sourceNodes[src][edge.rule].remove(edge)

    def get_target_rules(self):
        """
        Return a list of all the rules that this Node applies to its edges out.

        >>> src = Node()
        >>> tgt = Node()
        >>> src.set_target(tgt, "null", Edge())
        >>> src.get_target_rules()
        dict_keys(['null'])
        """
        if hasattr(self, "targetRules"):
            return self.targetRules.keys()
        else:
            return []

    def get_targets(self, rule):
        """
        Get the list of targetRules for the specified Rule.

        >>> src = Node()
        >>> tgt = Node()
        >>> src.set_target(tgt, "null", Edge())
        >>> src.get_targets("null")
        [Edge: [None]  --Edge-->  [None]]
        >>> src.get_targets("other")
        []
        """
        if hasattr(self, "targetRules") and rule in self.targetRules:
            return self.targetRules[rule]
        else:
            return []

    def get_source_rules(self):
        """
        Return a list of all the rules that apply to this Node.

        >>> src = Node()
        >>> tgt = Node()
        >>> src.set_target(tgt, "null", Edge())
        >>> tgt.get_source_rules()
        dict_keys(['null'])
        """
        if hasattr(self, "sourceRules"):
            return self.sourceRules.keys()
        else:
            return []

    def get_sources(self, rule):
        """
        Get the list of sourceRules for the specified Rule.

        >>> src = Node()
        >>> tgt = Node()
        >>> src.set_target(tgt, "null", Edge())
        >>> tgt.get_sources("null") #doctest: +NORMALIZE_WHITESPACE
        [Edge: [None]  --Edge-->  [None]]
        >>> src.get_targets("other")
        []
        """
        if hasattr(self, "sourceRules") and rule in self.sourceRules:
            return self.sourceRules[rule]
        else:
            return []

    def get_unfollowed_sources(self):
        """Return list of any unfollowed sources."""
        if not hasattr(self, '_unfollowedSources'):
            return []
        else:
            return self._unfollowedSources

    def get_unfollowed_targets(self):
        """Return list of any unfollowed targets."""
        if not hasattr(self, '_unfollowedTargets'):
            return []
        else:
            return self._unfollowedTargets

    def connect_allowed(self, other, edge_type, direction, delegate=None):
        """Provide an overridable means to children who may want to prevent connections."""
        return True

    def determine_event_hold_candidacy(self, runtime, node, direction='UP'):
        """
        look up the graph to determine if connected nodes are EHC
        or should be set as such.

        Traverse up the graph until:
           1) an EHC is detected upstream
           2) An un-sourced node is reached
           3) An event expecting node is found.

        Then, traverse down the graph, setting EHC (if upstream conditions dictate) until:
            1) A target-less node is reached
            2)  A node with EHC is detected
            3) An event expecting node is found (not sure what would happen here....)
        """
        if RunState.s(node).nord_ehc_inf_recurse_brake():
            if hasattr(node, '_ehc_enabled'):
                if node._ehc_enabled:
                    return node._event_sources
                else:
                    return node._ehc_enabled
            else:
                return False

        RunState.s(node).set_nord_ehc_inf_recurse_brake()
        if hasattr(node, '_ehc_enabled') and direction == 'UP':
            RunState.s(node).del_nord_ehc_inf_recurse_brake()
            if node._ehc_enabled:
                return node._event_sources
            else:
                return node._ehc_enabled

        if hasattr(node, 'handle_event'):
            node._ehc_enabled = True
            if not hasattr(node, '_event_sources'):
                node._event_sources = []
            node._event_sources += [node.id]
            # return True

        nf_sources = 0
        flow_sources = 0
        eventing_sources = 0
        non_eventing_sources = 0
        evid = None

        # Evaluate all upstream nodes.
        if node.get_source_rules():
            for rule in node.get_source_rules():
                if rule != 'flow':
                    nf_sources += len(node.get_sources(rule))
                if rule == 'flow':
                    flow_sources += 1
                for edge in node.get_sources(rule):
                    evid = self.determine_event_hold_candidacy(runtime, edge.source)
                    if evid:
                        eventing_sources += 1
                        if not hasattr(node, '_event_sources'):
                            node._event_sources = []
                        # Add any new eventing node id's to the list of potential event sources.
                        node._event_sources += list(set(evid) - set(node._event_sources))
                    else:
                        non_eventing_sources += 1

        if not hasattr(node, '_ehc_enabled') or not node._ehc_enabled:
            if eventing_sources:
                node._ehc_enabled = True
            else:
                node._ehc_enabled = False

        # call all the downstream nodes as well.
        if node.get_target_rules():
            for rule in node.get_target_rules():
                for edge in node.get_targets(rule):
                    if not hasattr(node, '_event_sources') or edge.target.id not in node._event_sources:
                        self.determine_event_hold_candidacy(runtime, edge.target, direction="DOWN")
                    else:
                        pass

        if hasattr(node, '_event_sources'):
            evid = node._event_sources

        RunState.s(node).del_nord_ehc_inf_recurse_brake()
        return evid

    def event_hold_candidate(self, runtime):
        """
        Evaluates this node, and and connected nodes if needed,
        to determine if it should not execute because it must wait
        for an upstream event.

        Child classes can affect the behavior of this
        method by explicitly setting their own
        value for:
        _ehc_enabled

        The existance of _ehc_enabled prevents the graph traversal
        process from trying to identify candidacy to hold processing
        while waiting for an upstream event.
        """
        rv = False
        if not hasattr(self, '_ehc_enabled'):
            self.determine_event_hold_candidacy(runtime, self)
        rv = self._ehc_enabled
        return rv

    def set_followed(self, edge, source):
        """
        Mark the passed edge for this node as followed.

        if this Node is the source of the edge
        update the targets lists, if this node is the
        target, updates the sources lists
        """
        if source:
            if not hasattr(self, '_unfollowedTargets'):
                return
            if edge not in self._unfollowedTargets:
                # at one point it was thought this should be an exceptional case...
                pass
            else:
                self._unfollowedTargets.remove(edge)
                self._followedTargets.append(edge)
        else:
            if not hasattr(self, '_unfollowedSources'):
                return
            if edge not in self._unfollowedSources:
                # at one point it was thought this should be an exceptional case...
                pass
            else:
                self._unfollowedSources.remove(edge)
                self._followedSources.append(edge)

    def set_unfollowed(self, edge, source):
        """Mark the passed edge for this node as not followed."""
        if source:
            if not hasattr(self, '_unfollowedTargets'):
                return
            if edge not in self._unfollowedTargets:
                self._unfollowedTargets.append(edge)
            if self._followedTargets and edge in self._followedTargets:
                self._followedTargets.remove(edge)
        else:
            if not hasattr(self, '_unfollowedSources'):
                return
            if edge not in self._unfollowedSources:
                self._unfollowedSources.append(edge)
            if self._followedSources and edge in self._followedSources: 
                self._followedSources.remove(edge)

    def is_ehc_or_runnable(self, runtime):
        if self.event_hold_candidate(runtime):
            if runtime._in_event and \
                    hasattr(self, '_event_sources') and \
                    runtime._in_event in self._event_sources:
                return True
            else:
                return False
        else:
            return True

    def in_event_path(self, runtime):
        if self.event_hold_candidate(runtime):
            if runtime._in_event and \
                    hasattr(self, '_event_sources') and \
                    runtime._in_event in self._event_sources:
                return True
            else:
                return False
        else:
            return False

    def reset_source_edges(self):
        """
        Move all the followed source edges
        to the end of the unfollowed list. But,
        do so in a way that places the followed
        edges after the residual unfollowed edges.

        No idea why this ordering is important...
        """
        if hasattr(self, '_followedSources'):
            for edge in self._followedSources: 
                if edge not in self._unfollowedSources:
                    self._unfollowedSources.append(edge)
                else:
                    self._unfollowedSources.remove(edge)
                    self._unfollowedSources.append(edge)
            self._followedSources.clear()

    def reset_target_edges(self):
        """
        Move all the followed target edges
        to the end of the unfollowed list. But,
        do so in a way that places the followed
        edges after the residual unfollowed edges.

        No idea why this ordering is important...
        """
        if hasattr(self, '_followedTargets'):
            for edge in self._followedTargets:
                if edge not in self._unfollowedTargets:
                    self._unfollowedTargets.append(edge)
                else:
                    self._unfollowedTargets.remove(edge)
                    self._unfollowedTargets.append(edge)
            self._followedTargets.clear() 

    def prep_to_execute(self, runtime, verbose):
        """
        Run the needed preparation steps on this node before it is executed.

        This is called prior to execution. The general case
        is to make sure all edges targeting this node, except
        for flow nodes have been followed, and for flow edges
        which target this node, mark them as followed
        as we are now executing them.

        Also, flag all downstream edges as unfollowed.

        If never reviewed, review all sources to see if this node is downstream of an event listener.
        If it is downstream of an event listener, check to see
        if the triggering event node ID is decorating the prior node????

        Look at all flow sources. If any are event eligible,
        then, check to see if we are being fired by the event?????

        TODO: figure out if the event is the correct
        """

        # Save the runtime so it can be referenced
        # in subsequent calls.
        self._current_runtime = runtime

        if not hasattr(self, '_unfollowedSources'):
            return

        has_unexecuted_sources = False
        unfollowed = self._unfollowedSources.copy()
        event_path_node = []
        to_run_sources = []
        for edge in unfollowed:
            if edge.rule != 'flow':
                """
                For non-flow edges, inject the sources
                which have not yet been executed, to ensure
                that when we follow those edges, the sources
                have been populated.

                IFF all the sources to this node
                have been executed, follow all the un-followed
                edges. Otherwise, let the runtime execute
                the nodes that need to be run first.
                """
                if not edge.source.has_executed():
                    # If we are reacting to an event, ensure that the
                    # upstream being considered is also a candidate
                    # for this event, otherwise, skip it.
                    if not edge.source.is_ehc_or_runnable(runtime):
                        continue
                    elif edge.source.in_event_path(runtime):
                        # set the nodes "in the event path" aside
                        # so they will be executed after any non-event path
                        # nodes.
                        event_path_node.append(edge.source)
                    else:
                        to_run_sources.append(edge.source)
                        # runtime.setNextNode(edge.source, 'UP')
                    RunState.s(self).set_preempted()
                    has_unexecuted_sources = True

        for node in event_path_node:
            # add the event path nodes to the end of the list
            # of nodes to execute. Ideally, there would be only
            # one .... But, I don't think this will get called....
            # unless we pre-empt the eventing node/path
            # such that all it's children and their siblings
            # run first. ... somehow
            # runtime.setNextNode(node, 'UP')
            to_run_sources.append(self)

        # Sort the the inbound nodes so that execution order is deterministic
        # Then add them to the runtime.
        # NOTE: This direction may need to be reversed
        for node in sorted(to_run_sources):  # this relies upon nord.nord.Base
            runtime.setNextNode(node, 'UP')

        if not has_unexecuted_sources:
            for edge in unfollowed:
                if edge.rule != 'flow':
                    if not edge.source.is_ehc_or_runnable(runtime):
                        continue
                    edge.follow(runtime, verbose)

        return

    def get_op_runtime(self):
        """
        Return the runtime set by
        the most recent call to prep_to_execute.
        """
        if hasattr(self, '_current_runtime'):
            return self._current_runtime
        elif hasattr(self, '_run_state'):
            self._current_runtime = RunState.s(self).runtime()
            return self._current_runtime
        else:
            return None

    def next_after_execute(self, runtime, verbose, affect_schedule=True):
        """
        Run the post-execution steps on this node.

        This is called after execution, and potentially after
        deferral, to set up the runtime state for the coming call to
        next
        """
        if hasattr(self, '_unfollowedSources'):
            unfollowed = self._unfollowedSources.copy()
            for edge in unfollowed:
                if edge.rule == 'flow':
                    edge.flag_followed(runtime)

        if not hasattr(self, '_unfollowedTargets'):
            return

        has_unexecuted_targets = False
        unfollowed = self._unfollowedTargets.copy()
        for edge in unfollowed:
            if edge.rule != 'flow':
                """
                For non-flow edges, inject the targets
                which have not yet been executed

                IFF all the targets from this node
                have been executed, follow all the un-followed
                edges. Otherwise, let the runtime execute
                the nodes that need to be run next.
                """
                if not edge.target.has_executed() and affect_schedule:
                    runtime.setNextNode(edge.target, 'DOWN')
                    has_unexecuted_targets = True

        if not has_unexecuted_targets:
            for edge in unfollowed:
                if edge.rule != 'flow':
                    edge.follow(runtime, verbose)

            for edge in unfollowed:
                if edge.rule == 'flow':
                    edge.follow(runtime, verbose)

    def has_out_edges(self):
        """
        Return if this node is the source of any edges.

        This checks to see if there are any edges from this
        node that haven't yet been followed (that should have been)
        """
        if hasattr(self, '_unfollowedTargets') and len(self._unfollowedTargets) > 0:
            return True
        elif hasattr(self, '_unfollowedTargets') and len(self._unfollowedTargets) == 0:
            return False

        # if for some reason _unfollowedTargets does not exist
        # return the truthiness of the check for no target rules
        # instead....
        return len(self.get_target_rules()) != 0

    def get_space(self):
        """Return the imported Base's spacename which should be specific to each extension."""
        return self._spacename

    def to_dict(self):
        """Return a dictionary of all simple datatype members."""
        rv = {}
        for k in (k for k in sorted(self.__dict__.keys()) if not k.startswith("_")):
            v = getattr(self, k)
            if type(v) in (str, int, float, bool):
                rv[k] = v
            """
            The visualization member of the object
            is established by Sigurd, and dump_to_dict method is
            defined in model.py (the parent to all node visualizations)
            """
            if k == "visualization":
                self.visualization.dump_to_dict(rv)  # pragma: no cover # must be covered by sigurd

        if self._spacename != "nord":
            rv["space"] = self._spacename

        return rv

    def __str__(self):
        """Stringify the most basic identifying information about the node."""
        rv = ""
        for k in (k for k in sorted(self.__dict__.keys()) if k in ('id', 'name')):
            if rv == "":
                rv = "{}: {}".format(k, self.__dict__[k])
            else:
                rv = "{}, {}: {}".format(rv, k, self.__dict__[k])
        return "Node({}): <{}>".format(type(self).__name__, rv)

    def __repr__(self):
        """Return the full representation of the node."""
        rv = "{}:".format(type(self).__name__)
        for a in (a for a in sorted(self.__dict__.keys())
                  if a not in ('sourceRules', 'targetRules', 'sourceNodes', 'targetNodes') and
                  not a.startswith('_')):
            rv = "{}\n    .{} = {}".format(rv, a, self.__dict__[a])

        if hasattr(self, "sourceRules"):
            rv = "{}\n  Sources".format(rv)
            for r in self.sourceRules.keys():
                for e in self.sourceRules[r]:
                    orphaned = ""
                    if not e.been_followed():
                        orphaned = "-!!-"
                    rv = "{}\n     - {} <{}- {}"\
                         .format(rv, type(e).__name__, orphaned, str(e.source))

        if hasattr(self, "targetRules"):
            rv = "{}\n  Targets".format(rv)
            for r in self.targetRules.keys():
                for e in self.targetRules[r]:
                    rv = "{}\n     - {} -> {}".format(rv, type(e).__name__, str(e.target))

        return rv


sys.modules[__name__] = Node
