"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

The heart of Nord's execution engine.

The Runtime executes frames. Frames are stacked. A Frame
has a list of nodes to execute. The runtime is steppable.
"""

import nord.nord.Config as config
import nord.nord.RunState as RS
import nord.nord.Frame as Frame
import sys
import asyncio
import time
from collections import deque, Iterable
import nord.pasion.managers.sessionmanager as sm
from nord.nord.utils.Debugger import tb


class Runtime(object):
    """
    Runtimes are the base class which facilitates the execution of the program.

    Runtimes control the execution of a map by following the Flow edges, and
    executing rules on nodes as the edges connecting the nodes expect.

    The Runtime keeps track of which edges it has followed.

    >>> Runtime() #doctest: +ELLIPSIS
    <__main__.Runtime object at 0x...>
    """

    def __init__(self):
        """Initialize the runtime."""
        self.current = None
        self.debug = False
        self.stack = list()
        self.stackPos = 0
        self.input_requests = {}
        self.input_provided = {}
        self.output_to_send = {}
        self.econsession = sm.get_current_session()
        self._exec_count = 0

        """
        The *events queue's are used to
        buffer events requested by front end user's
        and the responses in order until the
        appropriate time to process them.

        in_events are used to drive node execution
        once the current map iteration has completed
        (presumably, the same _exec_count)

        The _in_event item indicates the ID of the node which triggered the current
        execution.
        """
        self.in_events = deque(maxlen=config.max_in_events)
        self.out_events = deque(maxlen=config.max_out_events)
        self._in_event = None

    def add_in_event(self, event):
        """
        append the new event to the queue
        of inbound events.
        """
        self.in_events.append(event)
        self.stopped = False
        self.paused = True

    def next_from_in_events(self):
        """
        Set up this runtime instance to
        next run the requested node in the event.

        In the event of an error, stop the runtime
        by returning True
        """
        if len(self.in_events) == 0:
            return False
        event = self.in_events.pop()
        if 'node_id' not in event:
            self.anomaly = Exception('Runtime Error: malformed event')
            return True
        if event['node_id'] not in self.map.nodes:
            self.anomaly = Exception('Runtime Error: unknown node')
            return True
        node = self.map.nodes[event['node_id']]
        if not node.handles_events():
            self.anomaly = Exception('Runtime Error: event sent to node which cannot process it.')
            return True
        self._in_event = event['node_id']
        return node.handle_event(self, event)

    def get_exec_count(self):
        return self._exec_count

    def incr_exec_count(self):
        self._exec_count += 1

    def set_input(self, nodeid, content):
        """Cache the provided content for the node with provided id."""
        if nodeid in self.map.nodes:
            if nodeid not in self.input_provided:
                self.input_provided[nodeid] = deque()
            self.input_provided[nodeid].append(content)
            if nodeid in self.input_requests:
                del self.input_requests[nodeid]

    def get_input(self, nodeid):
        """Return the earliest input for the requesting node."""
        if nodeid in self.input_provided:
            if len(self.input_provided[nodeid]) > 0:
                return self.input_provided[nodeid].pop()
            else:
                return None
        else:
            self.input_requests[nodeid] = True
            return None

    def has_input(self, nodeid):
        """Check to see if there is input available for the requested node."""
        self.input_requests[nodeid] = True
        return nodeid in self.input_provided and len(self.input_provided[nodeid]) > 0

    def wants_input(self, nodeid):
        """Check to see if the node has requested input."""
        if nodeid in self.input_requests:
            return True
        else:
            return False

    def input_wanters(self):
        """Return the list of nodes waiting for input."""
        return list(self.input_requests.keys())

    def set_output(self, nodeid, content):
        """Cache the provided content for the node with provided id."""
        if nodeid in self.map.nodes:
            if nodeid not in self.output_to_send:
                self.output_to_send[nodeid] = list()
            self.output_to_send[nodeid].insert(0, content)

    def get_output(self, nodeid):
        """Return the earliest input for the requesting node."""
        if nodeid in self.output_to_send and self.output_to_send[nodeid]:
            rv = self.output_to_send[nodeid].pop()
            if len(self.output_to_send[nodeid]) == 0:
                del self.output_to_send[nodeid]
            return rv
        else:
            return None

    def has_output(self, nodeid):
        """Check to see if there is input available for the requested node."""
        return nodeid in self.output_to_send and len(self.output_to_send[nodeid]) > 0

    def get_outputters(self):
        """Return node id's for all nodes with output."""
        return list(self.output_to_send.keys())

    def initialize(self, inmap, verbose=False):
        """The first container node (typically Main) in a Map initializes the runtime."""
        self.map = inmap
        self.paused = False
        self.path = deque(maxlen=config.max_path_length)
        self.finished = False
        self.anomaly = None
        self.stopped = False
        self.verbose = verbose
        self._exec_count = 0
        # self.prep_top_of_stack()

    def setNextNode(self, nxt, direction):
        """Set the passed node to be the next executed by the runtime."""
        if self.current is None:
            self.current = Frame()
        self.current.inject_node(nxt, direction)

    # def pushNextNode(self, nxt):
        # """
        # """
        # self.nextNode.insert(0, nxt)

    def appendNextNode(self, nxt, direction):
        """Add passed node te the end of the current frame's list."""
        if self.current is None:
            self.current = Frame()
            self.current.inject_node(nxt, direction)
        else:
            self.current.append_node(nxt, direction)

    def topframe_append_node(self, nxt, direction):
        """Add passed node to the top frame's list of nodes."""
        self.stack[-1].append_node(nxt, direction)
        if not self.stack[-1].has_started():
            self.stack[-1].set_next()

    def replace_current_node(self, nxt, direction='DOWN'):
        """Replace the current frame pointer node with the provided one."""
        if self.current is None:
            self.current = Frame()
        self.current.replace_current(nxt, direction)

    def setReturn(self, retval):
        """
        Set the return value.

        This is deprecated and need be removed along with it's test code.
        """
        self.current.set_return(retval)

    def setAnomaly(self, exc):
        """TDB."""
        self.anomaly = exc

    def frame_next(self):
        """
        Execute all the non-Flow edge-rules affecting this node (recursively if need be).

        First all the non-flow rules for edges targeting this node are assembled and initialized

        Then all the non-flow rules for edges leaving this node are assembled and initialized

        Then, all the source rules are executed, followed by all the target rules.
        **NOTE** There should probably be a means of ordering which rules are
                 executed in what order for both sources and targets


        """
        try:
            rv = self.current.run_node(self)
        except Exception as e:
            print(f"{self.current.current} threw exception:\n{e}")

            last_type, last_value, exc_traceback = sys.exc_info()
            tb(exc_traceback, last_type, last_value)

            self.anomaly = e
            self.stopped = True
            return False

        # check to see if the stack has grown any
        # If so, process that before continuing the current Node
        if rv and len(self.stack) > self.stackPos:
            self.prep_top_of_stack()
            return rv

        if self.stack and self.current.done():
            # we're still in the stack someplace, but, should be done
            # with this frame, so, on the next iteration, drop back a level
            self.frame_finished()
        elif not self.stack and self.current.done():  # pragma: nocover ... reachable???
            # we're out of stack and out of edges to follow, so, we're done..
            # self.prep_next()
            self.current = None
        # elif self.current.has_out_edges():
        #    self.current = tmp_current

        return rv

    def runstate(self):
        """Helper method to return the runstate interrogator."""
        return RS

############################


# TODO: anomaly and state setting as well
# as how those constructs / concepts interact or are implemented by the runtime


############################

    def pushStack(self):

        """
        Append the current frame'self state, and a new list of next nodes to the stack.

        This prepares the stack to move into a new container such that
        the runtime can get back to the current state when this frame
        is completed.
        """
        self.current.set_deferred()
        nf = Frame()
        # nf.append_node(startNode)
        self.stack.append(nf)

    def prep_top_of_stack(self):
        """Ensure that the state is set up to run the topmost layer of the stack."""
        self.stackPos = len(self.stack) - 1
        self.current = self.stack[-1]

    def frame_finished(self):
        """Upon completion of a frame, deal with the stack as needed."""
        if not self.stack:
            self.current = None
            return
        else:
            last_rv = self.stack[-1].get_return()
            del self.stack[-1]
            if self.stack:
                self.stackPos = len(self.stack) - 1
                self.current = self.stack[-1]
                self.current.set_return(last_rv)
                # Frame return values are redundant with Return Edges

    def begin(self, start=None, debug=False):
        """
        Begin will start the execution of the map.

        If the input node is iterable, inject all
        the nodes in the list.
        """
        self.current = Frame()
        if start is None:
            self.map.detect_start_node()
            node = self.map.start_node
        else:
            node = start
        if isinstance(node, Iterable):
            for n in node:
                self.current.append_node(n, 'DOWN')
        else:
            self.current.inject_node(node, 'DOWN')
        self.stack.append(self.current)
        self.current.set_next()
        self.debug = debug

        while not self.debug and\
                not self.paused and\
                not self.stopped and\
                len(self.input_wanters()) == 0:
            sleeptime = self.next()
            time.sleep(sleeptime)

    async def continue_async(self):
        """Resume execution of the map."""
        while not self.debug and not self.paused and not self.stopped:
            sleeptime = self.next()
            await asyncio.sleep(sleeptime)

    def continue_exec(self):
        """Resume execution of the map."""
        while not self.debug and\
                not self.paused and\
                not self.stopped and\
                len(self.input_wanters()) == 0:
            sleeptime = self.next()
            time.sleep(sleeptime)

    async def begin_async(self, start=None, debug=False):
        """Begin will start the execution of the map."""
        self.current = Frame()
        if start is None:
            self.map.detect_start_node()
            node = self.map.start_node
        else:
            node = start
        if isinstance(node, Iterable):
            for n in node:
                self.current.append_node(n, 'DOWN')
        else:
            self.current.inject_node(node, 'DOWN')
        self.stack.append(self.current)
        self.current.set_next()
        self.debug = debug

        while not self.debug and not self.paused and not self.stopped:
            sleeptime = self.next()
            await asyncio.sleep(sleeptime)

    def next(self):
        """
        Perform the next step in the execution flow.

        Next will evaluate the current condition of the current node, and flow to the next node
        as indicated, or indicate completion of the map, or if an anomaly has occurred set the
        anomaly condition and indicate the map has terminated

        If the runtime is not paused or in debug mode, next will call the next node in the flow
        """
        # if current is not set, get outta here, nothing left to do.
        if self.current is None:
            self.incr_exec_count()
            if not RS.s(self.map).walujapi_once_through() and self.in_events:
                # If this map has not been flagged as executed
                # find any and all possible start nodes that have
                # not been run yet, and queue them for execution.
                startable_nodes, found_any = self.map.get_sourceless_nodes()
                for node in startable_nodes:
                    self.setNextNode(node, 'DOWN')
                self.stopped = False
                RS.s(self.map).set_walujapi_once_through()
                return 0.0

            # That is, unless there are some events to process
            if len(self.in_events) and self._exec_count:
                if self.next_from_in_events():
                    self.stopped = True
            else:
                self.stopped = True

            return 0.0

        # This will bill the user
        # if a session is established, and
        # in the case of insufficient funds, will sleep
        # until funds come available.
        if not sm.consumesw(1.0) or not sm.consumehw(1.0):
            self.setAnomaly(Exception("Insufficient Funds"))
            return 5.0

        if self.current.is_blocked():
            # If run_node returns true, it is still blocked, so return, else, continue
            if self.current.run_node(self):
                return 0.0
            else:
                # the node was blocked, but completed...
                self.current.set_next()

        # add the current node to the list of traversed nodes, and
        # removes oldest elements on the path according to the configuration
        # NOTE: It is not presently known if all Nodes
        # will make in onto this path, see Frame.set_next
        if self.current.current:
            self.path.append(self.current.current)

        # presently the current frame prep_next is a no-op
        self.current.prep_next()
        # frame next will execute the next node in the frame
        deferred = self.frame_next()

        if self.anomaly is not None:
            print("Runtime Anomaly '{}' raised".format(self.anomaly))
            self.stopped = True
            self.current = None

        if not deferred and self.current and not self.is_blocked():
            self.current.set_next()

        return 0.0

    def pause(self):
        """
        Set the paused flag.

        This means calls to next will not proceed through the flow without interaction.
        """
        self.paused = True

    def unpause(self):
        """
        Unset the paused flag.
        """
        self.paused = False

    def resume(self):
        """Unset the paused flag and continue the execution."""
        self.paused = False
        while not self.debug and not self.paused and not self.stopped:
            sleeptime = self.next()
            time.sleep(sleeptime)

    def stop(self):
        """Do nothing ATT."""

    def finish(self):
        """Unpause and turn off debugging such that execution can continue."""
        self.paused = False
        self.debug = False
        self.stopped = True

    def finalize(self):
        """Do nothing ATT."""

    def get_current_node(self):
        """Return the current node (the next to execute)."""
        if self.current is None:
            return None
        return self.current.get_current_node()

    def get_current_direction(self):
        """
        Return the direction by which the current node in the current frame
        was reached.

        'DOWN' means the current node was the Target, or initial node
        'UP' means the node was requested to be run by one of it's targets
        aka a node to which it is the source.
        """
        if self.current is None:
            raise Exception("DEBUGGING: node direction requested, but no Node in frame.")
        else:
            return self.current.get_current_direction()

    def is_blocked(self):
        """Return whether the current frame is currently blocked."""
        if self.current is not None:
            return self.current.is_blocked()
        else:
            return False

    def is_paused(self):
        """Return whether the runtime is paused."""
        return self.paused

    def has_anomaly(self):
        """Indicate if this runtime has an anomaly."""
        return self.anomaly is not None

    def get_anomaly(self):
        """Return the anomaly or None if there is not one."""
        return self.anomaly

    def is_done(self):
        """Return whether the current map is completed."""
        if self.anomaly is not None:
            return True
        if self.stopped:
            return True
        if self.paused:
            return False
        return (self.current is None or self.current.current is None) and not self.stack


def execute_cli_runtime():
    pass
