"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

import nord.nord.Receiver as Receiver
import nord.nord.Property as Property
import sys


class Container(Receiver):
    """
    Container Map elements are used to group related elements together. Containers are somewhat analogous to
    namespaces and blocks from the textual programming paradigm. A container might be subclassed to create
    methods, "classes", "packages", "applications", and so on

    >>> Container() #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Container:

    """
    def __init__(s):
        s._properties = {
            'menuicon': Property(s,
                                                  'menuicon',
                                                  default='unknown',
                                                  description=_(  # noqa: F821
                                                      "The name of the icon to use when "
                                                      "Use'ing this container in another Map"))
        }

    def execute(s, runtime):
        """
            The runtime looks (automatically) for flow rules
            from the container to any of the nodes within the container

            This method will establish flow rules to all contained nodes
            without any other edges targetting them, if there are no flow
            edges to any of the contained nodes

            The nodes will be sorted by name, then ID if there are more
            than one to be kicked off
        """

        # don't bother with this if there are any flow targets
        # contained by this node
        for edge in s.get_targets('flow'):
            for e2 in edge.target.get_sources('contains'):
                if e2.source == s:
                    s.flag_exe(runtime)
                    return

        # add a nev "frame" to the stack for this container
        runtime.pushStack()

        # create an empty list to store any nodes we will need to
        # push into a "new stack" to execute before continuing on with this containers
        # targets
        newFlowReceivers = list()
        # So, first get a list of all the edges from
        # this container to the nodes it contains
        for edge in s.get_targets('contains'):
            topTarget = True
            # Then, eliminate all nodes from candidacy for
            # being the start of a "new stack" which have
            # non-containing edges, or, whose containing
            # edges have already been followed
            for rule in edge.target.get_source_rules():
                if rule != 'contains':
                    # for every edge in this rule type
                    # check to see if any haven't yet been followed.
                    # if none have, then, this is a candidate to
                    # run in a "new stack"
                    alreadyFollowed = True
                    for e4 in edge.target.get_sources(rule):
                        if e4.source == s and not e4.been_followed(runtime):
                            alreadyFollowed = False
                            e4.flag_followed(runtime)
                    if alreadyFollowed:
                        topTarget = False
                else:
                    # check each containing edge, and if any from this container
                    # have not been followed set for candidacy to start a "new stack"
                    """
                    alreadyFollowed = True
                    for e3 in edge.target.get_sources('contains'):
                        if e3.source == s and not e3.been_followed(runtime):
                            alreadyFollowed = False
                    if not alreadyFollowed:
                        e3.flag_followed(runtime)
                        topTarget = True

                        break
                    """

            if topTarget:
                newFlowReceivers.append(edge.target)

        for tgt in sorted(newFlowReceivers):
            # e = Flow()
            # e.set_nodes(s, tgt, 'flow')
            runtime.topframe_append_node(tgt, 'DOWN')

        s.flag_exe(runtime)


sys.modules[__name__] = Container
