"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

import uuid
import nord.nord.InstanceOf as InstanceOf
import sys


class Data(InstanceOf):
    """
    Data nodes are the base class for any "variable" map element.

    >>> Data(name="fred", ntype=str) #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = fred
        .type = <class 'str'>
        .value = None

    >>> Data(name="fred") #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = fred
        .type = None
        .value = None

    >>> Data() #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = ...
        .type = None
        .value = None

    """
    def __init__(s, name=None, ntype=None):
        s.name = name
        if name is None:
            s.name = str(uuid.uuid4())
        s.type = ntype
        s.value = None

    def set(s, value):
        """
        Set the value of this data node
        """
        s.value = value

    def get(s):
        """
        Get the value from this data node
        """
        return s.value

    def suppress_output(self):
        """Supress any output from being generated by this node."""
        self._output_suppressed = True

    def permit_output(self):
        """Remove any output suppression from this node."""
        if hasattr(self, '_output_suppressed'):
            delattr(self, '_output_suppressed')

    def execute(s, runtime):
        """
        Run this node.

        Data nodes, must have data passed to them, or, they will
        prompt for the data if nothing is passed to them.

        Similarly, if they are not passing or assigning their
        value to another node, they will output their value.
        """
        if not len(s.get_sources("pass")) and\
                not len(s.get_sources("assign")) and\
                not len(s.get_all_instance_sources("pass")) and\
                not len(s.get_all_instance_sources("assign")):
            dat = runtime.get_input(s.id)
            if dat is None:
                s.flag_blocking()
            else:
                s.clear_blocking()
                s.set(dat)
                s.flag_exe(runtime)

        else:
            s.flag_exe(runtime)

        if not len(s.get_targets("pass")) and\
                not len(s.get_targets("assign")) and\
                not hasattr(s, '_output_suppressed') and\
                not len(s.get_all_instance_targets("pass")) and\
                not len(s.get_all_instance_targets("assign")):
            runtime.set_output(s.id, s.value)


sys.modules[__name__] = Data
