"""
Copyright 2019 NoRD Software Foundation

This file is part of NoRD.

NoRD is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

NoRD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with NoRD.  If not, see <https://www.gnu.org/licenses/>.

--

"""

import nord.nord.InstanceOf as InstanceOf
import sys
import nord.sigurd.utils.namegen as ng


class Data(InstanceOf):
    """
    Data nodes are the base class for any "variable" map element.

    >>> Data(name="fred", ntype=str) #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = fred
        .type = <class 'str'>
        .value = None

    >>> Data(name="fred") #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = fred
        .type = None
        .value = None

    >>> Data() #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Data:
        .name = ...
        .type = None
        .value = None

    """
    def __init__(s, graph, name=None, ntype=None):
        super().__init__(graph)
        if name is None:
            s.set_untracked_attr('name', ng.get(s))
        else:
            s.set_untracked_attr('name', name)
        if ntype is not None:
            s.set_untracked_attr('type', ntype)
        s.set_untracked_attr('value', None)

    def set(s, value):
        """
        Set the value of this data node
        """
        s.set_untracked_attr('_value_has_been_set', True)
        s.value = value

    def get(s):
        """
        Get the value from this data node
        """
        return s.value

    def suppress_output(self):
        """Supress any output from being generated by this node."""
        self._output_suppressed = True

    def permit_output(self):
        """Remove any output suppression from this node."""
        if hasattr(self, '_output_suppressed'):
            delattr(self, '_output_suppressed')

    def execute(s, runtime):
        """
        Run this node.

        Data nodes, must have data passed, or assigned to them, or,
        they will prompt for the data if nothing has been set yet.

        Similarly, if they are not passing or assigning their
        value to another node, they will output their value.
        """
        if not len(s.get_sources("pass")) and\
                not len(s.get_sources("assign")) and\
                not len(s.get_all_instance_sources("pass")) and\
                not len(s.get_all_instance_sources("assign")) and\
                not hasattr(s, '_value_has_been_set'):
            dat = runtime.get_input(s.id)
            if dat is None:
                s.flag_blocking()
            else:
                s.clear_blocking()
                s.set(dat)
                s.flag_exe(runtime)

        else:
            s.flag_exe(runtime)

        if not len(s.get_targets("pass")) and\
                not len(s.get_targets("assign")) and\
                not hasattr(s, '_output_suppressed') and\
                not len(s.get_all_instance_targets("pass")) and\
                not len(s.get_all_instance_targets("assign")):
            s.set_output_for_fe(runtime)

    def set_output_for_fe(self, runtime):
        runtime.set_output(self.id, self.value)

    def json_serialize(self):
        """
        This should be overridden if the value more complex than
        any atomic data types.
        """
        return self.value


sys.modules[__name__] = Data
